package api

import (
	"errors"
	"fmt"
	"net/http"
	"strconv"
	"sync"

	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"
	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"
	fclog "git.inverse.ca/inverse/fingerbank-collector/log"

	"github.com/gin-gonic/gin"
)

var edc = fccommon.EndpointDataComparator{}

//TODO: add support for filtering and pagination, most likely when FingerbankCollector has its backend finalized
func listEndpointData(c *gin.Context) {
	c.JSON(http.StatusOK, getExpandedEndpoints(false))
}

//TODO: add support for filtering and pagination, most likely when FingerbankCollector has its backend finalized
func listEndpointDataDetails(c *gin.Context) {
	logger := fclog.Logger("")

	const maxApiTries = 3
	const maxApiConcurrency = 5

	eds := getExpandedEndpoints(true)

	var failed bool
	results := make([]*fccommon.QueryResult, len(eds))
	conc := make(chan int, maxApiConcurrency)
	l := &sync.Mutex{}
	w := &sync.WaitGroup{}

	i := 0
	for _, ed := range eds {
		w.Add(1)
		conc <- 1

		if failed {
			w.Done()
			break
		}

		go func(ed *fccommon.ExpandedEndpointData) {
			defer w.Done()
			var result *fccommon.QueryResult
			tries := 0
			// Make sure we don't try more than the max tries
			for result == nil && tries <= maxApiTries {
				tries++

				resultStruct, err := collector.InterrogateAPI(ed.MAC, &ed.EndpointData)

				if err != nil {
					logger.Warn(fmt.Sprintf("Failed to call Fingerbank API for device details: %s", err))

					if tries >= maxApiTries {
						logger.Error("Too many errors attempting to get details for device. Aborting.")
						l.Lock()
						failed = true
						l.Unlock()
						break
					}
				} else {
					result = &resultStruct
					l.Lock()
					results[i] = result
					i++
					l.Unlock()
					<-conc
				}
			}
		}(ed)
	}

	w.Wait()

	if failed {
		c.JSON(http.StatusInternalServerError, gin.H{"error": "Failed to get one or more details around the endpoints. Check server side logs for details."})
	} else {
		switch c.Query("format") {
		case "html":
			c.HTML(http.StatusOK, `listEndpointDataHTML`, gin.H{"Endpoints": results})
		default:
			c.JSON(http.StatusOK, results)
		}
	}
}

func getExpandedEndpoints(localOnly bool) []*fccommon.ExpandedEndpointData {
	var eds []*fccommon.ExpandedEndpointData = make([]*fccommon.ExpandedEndpointData, 0)

	for _, mac := range collector.ListEndpointsIds() {
		ed, _ := collector.FindEndpointDataMAC(mac)
		if !localOnly {
			ed = collector.EndpointDataODA(ed)
		}
		lockid := ed.RLock()
		eds = append(eds, collector.ExpandEndpointData(mac, ed))
		ed.RUnlock(lockid)
	}
	return eds
}

func updateEndpointData(c *gin.Context) {
	logger := fclog.Logger("")

	if ed, err := findEndpointData(c, c.Param("deviceId"), true); err != nil {
		c.JSON(http.StatusNotFound, gin.H{"error": err.Error()})
	} else {
		cp := ed.Copy()
		ed.Update(func(ed *fccommon.EndpointData) bool {
			if err := c.BindJSON(ed); err == nil {
				c.JSON(http.StatusOK, gin.H{"status": "Endpoint successfully updated"})
			} else {
				c.JSON(http.StatusUnprocessableEntity, gin.H{"error": err.Error()})
			}

			changed, _, err := edc.BuildDeltaEndpoint(cp, ed)
			if err != nil {
				logger.Error(fmt.Sprintf("Issue while checking if endpoint was modified: %s", err))
			}

			return changed
		})
	}
}

func deleteEndpointData(c *gin.Context) {
	if ed, err := findEndpointData(c, c.Param("deviceId"), true); err != nil {
		c.JSON(http.StatusNotFound, gin.H{"error": err.Error()})
	} else {
		if collector.DeleteEndpointData(ed) {
			c.JSON(http.StatusOK, gin.H{"status": "Endpoint successfully deleted"})
		} else {
			c.JSON(http.StatusUnprocessableEntity, gin.H{"status": "Couldn't delete endpoint"})
		}
	}
}

func endpointDataIsA(c *gin.Context) {
	deviceId := c.Param("deviceId")
	if ed, err := findEndpointData(c, deviceId, true); err != nil {
		c.JSON(http.StatusNotFound, gin.H{"error": err.Error()})
	} else {
		if otherDeviceId, err := strconv.ParseInt(c.Param("otherDeviceId"), 10, 64); err == nil {
			queryResult, err := collector.InterrogateAPI(deviceId, ed)
			if err != nil {
				c.JSON(http.StatusInternalServerError, gin.H{"status": err.Error()})
				return
			}

			isA, err := fingerbankclient.EndpointIsA(queryResult.CloudAPIResult.Device.Id, int(otherDeviceId))

			if err != nil {
				c.JSON(http.StatusInternalServerError, gin.H{"status": err.Error()})
				return
			}

			c.JSON(http.StatusOK, isA)

		} else {
			c.JSON(http.StatusUnprocessableEntity, gin.H{"status": "Cannot parse the device ID: usage: '/endpoint_data/192.168.1.5/is_a/1234' where 1234 is a device ID in the Fingerbank database"})
		}
	}
}

func getEndpointData(c *gin.Context) {
	if ed, err := findEndpointData(c, c.Param("deviceId"), false); err != nil {
		c.JSON(http.StatusNotFound, gin.H{"error": err.Error()})
	} else {
		lockid := ed.RLock()
		defer ed.RUnlock(lockid)
		c.JSON(http.StatusOK, collector.ExpandEndpointData(ed.GetMac(), ed))
	}
}

func endpointDataDetails(c *gin.Context) {
	if ed, err := findEndpointData(c, c.Param("deviceId"), true); err != nil {
		c.JSON(http.StatusNotFound, gin.H{"error": err.Error()})
	} else {
		result, err := collector.InterrogateAPI(c.Param("deviceId"), ed)
		if err != nil {
			c.JSON(http.StatusInternalServerError, gin.H{"error": err.Error()})
		} else {
			id := ed.RLock()
			defer ed.RUnlock(id)
			c.JSON(http.StatusOK, result)
		}
	}
}

func endpointDataPrintQuery(c *gin.Context) {
	if ed, err := findEndpointData(c, c.Param("deviceId"), false); err != nil {
		c.JSON(http.StatusNotFound, gin.H{"error": err.Error()})
	} else {
		if err != nil {
			c.JSON(http.StatusInternalServerError, gin.H{"error": err.Error()})
		} else {
			c.JSON(http.StatusOK, ed.ToInterrogateQuery(collector))
		}
	}
}

func endpointDataAnalyse(c *gin.Context) {
	if ed, err := findEndpointData(c, c.Param("deviceId"), true); err != nil {
		c.JSON(http.StatusNotFound, gin.H{"error": err.Error()})
	} else {
		events, err := collector.AnalyseEndpoint(ed)
		if err != nil {
			c.JSON(http.StatusInternalServerError, gin.H{"error": err.Error()})
		} else {
			id := ed.RLock()
			defer ed.RUnlock(id)
			c.JSON(http.StatusOK, gin.H{"message": "Analysed the endpoint", "events": events})
		}
	}
}

func findEndpointData(c *gin.Context, deviceId string, localOnly bool) (*fccommon.EndpointData, error) {
	var result *fccommon.EndpointData
	var err error
	if fccommon.ValidMAC(deviceId) {
		result, err = collector.FindEndpointDataMAC(deviceId)
		if err != nil {
			fmt.Printf("Error finding endpoint data: %s", err)
			return nil, errors.New("Cannot find device with this MAC address")
		}
	} else if fccommon.ValidIP(deviceId) {
		result, err = collector.FindEndpointDataIP(deviceId)
		if err != nil {
			fmt.Printf("Error finding endpoint data: %s", err)
			return nil, errors.New("Cannot find device with this IP address")
		}
	} else if fccommon.ValidIPv6(deviceId) {
		result, err = collector.FindEndpointDataIP(deviceId)
		if err != nil {
			fmt.Printf("Error finding endpoint data: %s", err)
			return nil, errors.New("Cannot find device with this IPv6 address")
		}
	} else {
		return nil, errors.New("Not a valid IP or MAC address")
	}
	if !localOnly {
		result = collector.EndpointDataODA(result)
	}
	return result, nil
}
