package main

import (
	"bufio"
	"flag"
	"fmt"
	"net"
	"os"
	"runtime"
	"runtime/pprof"
	"strings"
	"sync"
	"time"

	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"

	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"
	"github.com/google/gopacket/pcap"
	"github.com/inverse-inc/go-utils/sharedutils"

	// Import of the handlers goes here

	fcallnoophandler "git.inverse.ca/inverse/fingerbank-collector/handlers/allnoop"
	fcdhcpv4handler "git.inverse.ca/inverse/fingerbank-collector/handlers/dhcpv4"
	fcdhcpv6handler "git.inverse.ca/inverse/fingerbank-collector/handlers/dhcpv6"
	fcdnshandler "git.inverse.ca/inverse/fingerbank-collector/handlers/dns"
	fchttphandler "git.inverse.ca/inverse/fingerbank-collector/handlers/http"
	fcmdnshandler "git.inverse.ca/inverse/fingerbank-collector/handlers/mdns"
	fcradiushandler "git.inverse.ca/inverse/fingerbank-collector/handlers/radius"
	fctcphandler "git.inverse.ca/inverse/fingerbank-collector/handlers/tcp"
	fctlshandler "git.inverse.ca/inverse/fingerbank-collector/handlers/tls"
	fcupnphandler "git.inverse.ca/inverse/fingerbank-collector/handlers/upnp"
)

type AvailableHandler struct {
	Name        string
	Bpf         string
	Constructor func(fccommon.Collector) fccommon.ProtocolHandler
	Disabled    bool
}

var available_handlers = []AvailableHandler{
	AvailableHandler{
		Name:        fcdnshandler.Name,
		Bpf:         fcdnshandler.DefaultBPF,
		Constructor: fcdnshandler.NewDNSHandler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fcdhcpv4handler.Name,
		Bpf:         fcdhcpv4handler.DefaultBPF,
		Constructor: fcdhcpv4handler.NewDHCPv4Handler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fcdhcpv6handler.Name,
		Bpf:         fcdhcpv6handler.DefaultBPF,
		Constructor: fcdhcpv6handler.NewDHCPv6Handler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fchttphandler.Name,
		Bpf:         fchttphandler.DefaultBPF,
		Constructor: fchttphandler.NewHTTPHandler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fcupnphandler.Name,
		Bpf:         fcupnphandler.DefaultBPF,
		Constructor: fcupnphandler.NewUPnPHandler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fcmdnshandler.Name,
		Bpf:         fcmdnshandler.DefaultBPF,
		Constructor: fcmdnshandler.NewMDNSHandler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fctlshandler.Name,
		Bpf:         fctlshandler.DefaultBPF,
		Constructor: fctlshandler.NewTLSHandler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fcradiushandler.Name,
		Bpf:         fcradiushandler.DefaultBPF,
		Constructor: fcradiushandler.NewRadiusHandler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fctcphandler.Name,
		Bpf:         fctcphandler.DefaultBPF,
		Constructor: fctcphandler.NewTCPHandler,
		Disabled:    false,
	},
	AvailableHandler{
		Name:        fcallnoophandler.Name,
		Bpf:         fcallnoophandler.DefaultBPF,
		Constructor: fcallnoophandler.NewAllnoopHandler,
		Disabled:    true,
	},
}

var interfacesCSV = flag.String(
	"interfaces",
	sharedutils.EnvOrDefault("COLLECTOR_INTERFACES", ""),
	"comma delimited list of interfaces to listen on. Defaults to listening on all ethernet interfaces. (COLLECTOR_INTERFACES environment variable)",
)

var noCapture = flag.Bool(
	"no-capture",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_NO_CAPTURE", "false")),
	"use of this flag disables the packet capture functionality of the collector (COLLECTOR_NO_CAPTURE environment variable)",
)

var onlineConcurrency = flag.Int(
	"online-concurrency",
	sharedutils.EnvOrDefaultInt("COLLECTOR_ONLINE_CONCURRENCY", 10),
	"the amount of concurrent packet analysis when performing online capture. Increase when using the collector on a powerful machine to get better results.",
)

var offlineCapture = flag.String(
	"offline-capture",
	"",
	"path to a pcap file to analyse using the collector. Usage of this flag implies that no online capture will be done. The collector will perform the analysis of the pcap file and then exit.",
)

var offlineConcurrency = flag.Int(
	"offline-concurrency",
	4,
	"the amount of concurrent packet analysis when analysing an offline pcap. Increase when using the collector on a powerful machine to get better results.",
)

var offlineCapturePauseOnCompletion = flag.Bool(
	"offline-capture-pause-on-completion",
	false,
	"whether or not to stop after processing the pcap file so that the API can be queried based on what has been learned via the offline capture",
)

var tcpHandlerPacketThreshold = flag.Int(
	"tcp-handler-packet-threshold",
	sharedutils.EnvOrDefaultInt("COLLECTOR_TCP_HANDLER_PACKET_THRESHOLD", 10),
	"How many TCP SYN + SYN-ACK packets to analyse for an endpoint before ignoring the next packets for the duration defined in tcp-handler-packet-threshold-ignore-duration. A value of 0 disables it.",
)

var tcpHandlerPacketThresholdIgnoreDuration = flag.Duration(
	"tcp-handler-packet-threshold-ignore-duration",
	fccommon.MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_TCP_HANDLER_PACKET_THRESHOLD_IGNORE_DURATION", "5m")),
	"How much time to ignore TCP SYN and SYN-ACK packets after tcp-handler-packet-threshold has been met for an endpoint",
)

// Package initialization
// - Registers the disable flags so that handlers can be enabled/disabled
func init() {
	for i := 0; i < len(available_handlers); i++ {
		upperName := strings.ToUpper(available_handlers[i].Name)

		disabled := false
		disableEnvVar := "COLLECTOR_DISABLE_" + upperName + "_HANDLER"
		if envVal := sharedutils.EnvOrDefault(disableEnvVar, ""); envVal != "" {
			disabled = envVal == "true"
		} else {
			disabled = available_handlers[i].Disabled
		}

		flag.BoolVar(
			&available_handlers[i].Disabled,
			"disable-"+available_handlers[i].Name+"-handler",
			disabled,
			"Disables the "+available_handlers[i].Name+" protocol handler (environment variable: "+disableEnvVar+")",
		)

		filterEnvVar := "COLLECTOR_FILTER_" + upperName + "_HANDLER"
		flag.StringVar(
			&available_handlers[i].Bpf,
			"filter-"+available_handlers[i].Name+"-handler",
			sharedutils.EnvOrDefault(filterEnvVar, available_handlers[i].Bpf),
			"BPF Filter for "+available_handlers[i].Name+" (environment variable: "+filterEnvVar+")",
		)
	}

}

type ListenInterface struct {
	handle *pcap.Handle
	ift    *net.Interface
}

// Open all the ethernet interfaces or the interfaces that are specified via command line
// Returns the net.Interface and a pcap handle in the ListenInterface struct
func openListenInterfaces() map[string]ListenInterface {
	var devices []pcap.Interface
	var err error
	if *interfacesCSV != "" {
		for _, ift := range strings.Split(*interfacesCSV, ",") {
			devices = append(devices, pcap.Interface{Name: ift})
		}
	} else {
		devices, err = pcap.FindAllDevs()
		if err != nil {
			panic(err)
		}
	}

	listenInts := make(map[string]ListenInterface)
	for _, device := range devices {
		handle, err := pcap.OpenLive(device.Name, 1600, true, pcap.BlockForever)

		if err != nil {
			logger.Error(fmt.Sprintf("Can't open %s: %s", device.Name, err))
			continue
		}

		linkType := handle.LinkType()

		if linkType != layers.LinkTypeEthernet || device.Name == "any" {
			logger.Warn(fmt.Sprintf("Ignoring interface %s", device.Name))
			handle.Close()
			continue
		} else {
			// Add the interface to the collector interfaces and populate the handle list
			ift, err := net.InterfaceByName(device.Name)
			if err != nil {
				panic(err)
			}
			collector.Interfaces = append(collector.Interfaces, ift)

			listenInts[device.Name] = ListenInterface{handle: handle, ift: ift}
		}
	}

	// Given we have a weird bug with Golang scheduling when the amount of pcap listen reach the amount of GOMAXPROCS, we set it so every pcap listen doesn't consume a GOMAXPROCS
	runtime.GOMAXPROCS(runtime.GOMAXPROCS(-1) + len(listenInts))

	return listenInts
}

// Setup the collector for a set of interfaces and their pcap handle
func setupCollector(listenInts map[string]ListenInterface) {
	if len(listenInts) == 0 {
		panic("No interface to listen on")
	}

	// The collector needs a single pcap handle so it can create its bpfs
	for _, listenInt := range listenInts {
		collector.SetPcapHandle(listenInt.handle)
		break
	}

	// register all the enabled handlers
	registerHandlers()

	bpf := collector.HandlersBPF()
	logger.Info(fmt.Sprintf("Starting with BPF: %s \n", bpf))

	if bpf == "" {
		panic("Trying to start with an empty BPF")
	}

	for _, listenInt := range listenInts {
		if err := listenInt.handle.SetBPFFilter(bpf); err != nil {
			panic(err)
		}
	}

	collector.PopulateIpAddresses()
}

// Register the enabled handlers in the collector
func registerHandlers() {
	for _, handler := range available_handlers {
		if !handler.Disabled {
			logger.Info(fmt.Sprintf("Registering handler %s \n", handler.Name))
			h := handler.Constructor(collector)
			err := collector.RegisterProtocolHandler(h, handler.Bpf)
			sharedutils.CheckError(err)
			// Special parameters for TCP handler
			if handler.Name == fctcphandler.Name {
				if *tcpHandlerPacketThreshold != 0 {
					h.(*fctcphandler.TCPHandler).InitRateLimit(*tcpHandlerPacketThreshold, *tcpHandlerPacketThresholdIgnoreDuration)
				}
			}
		}

	}

}

// Start the capture process and perform network capture on all the interfaces the collector should be running on
func captureAndProcess() {

	if *noCapture {
		logger.Warn("Configured to skip the packet capture")
		return
	}

	handles := openListenInterfaces()
	setupCollector(handles)

	if *offlineCapture != "" {
		// We need to pick a "fake" interface that we'll give to the collector
		var fakeInt ListenInterface
		for _, i := range handles {
			fakeInt = i
			break
		}

		if handle, err := pcap.OpenOffline(*offlineCapture); err != nil {
			panic("Impossible to open offline pcap file" + err.Error())
		} else {
			bpf := collector.HandlersBPF()
			if err := handle.SetBPFFilter(bpf); err != nil {
				panic(err)
			}

			start := time.Now()
			wg := &sync.WaitGroup{}
			packetSource := gopacket.NewPacketSource(handle, handle.LinkType())
			conc := make(chan int, *offlineConcurrency)
			for packet := range packetSource.Packets() {
				wg.Add(1)
				conc <- 1
				packet := packet
				go func() {
					collector.HandleGoPacket(fakeInt.ift, packet)
					<-conc
					wg.Done()
				}()
			}
			wg.Wait()

			elapsed := time.Now().Sub(start)
			fmt.Fprintf(os.Stderr, "Took %f seconds to process PCAP\n", elapsed.Seconds())
			pprof.StopCPUProfile()

			collector.Destroy()

			if *offlineCapturePauseOnCompletion {
				fmt.Print("Press 'Ctrl-C' to exit...")
				for {
					bufio.NewReader(os.Stdin).ReadBytes('\n')
				}
			}

			os.Exit(0)
		}
	} else {
		conc := make(chan int, *onlineConcurrency)
		for intName, listenInt := range handles {
			go func(intName string, listenInt ListenInterface) {
				wg.Add(1)
				defer func() {
					if r := recover(); r != nil {
						logger.Error(fmt.Sprintf("Can't start pcap on %s. Error: %s", intName, r))
					}

					wg.Done()
				}()

				logger.Info("Listening on " + listenInt.ift.Name)

				packetSource := gopacket.NewPacketSource(listenInt.handle, listenInt.handle.LinkType())
				packetSource.DecodeOptions.Lazy = true
				//packetSource.DecodeOptions.NoCopy = true
				packetsChan := packetSource.Packets()
				for {
					select {
					case packet := <-packetsChan:
						conc <- 1
						go func() {
							defer func() { <-conc }()
							collector.HandleGoPacket(listenInt.ift, packet)
						}()
					}
				}
			}(intName, listenInt)
		}
	}

}
