package fccommon

import (
	"crypto/tls"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"net"
	"net/http"
	"runtime/debug"
	"strconv"
	"strings"
	"sync/atomic"
	"time"

	fcreplication "git.inverse.ca/inverse/fingerbank-collector/common/replication"
	"git.inverse.ca/inverse/fingerbank-collector/dpi"
	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"
	fclog "git.inverse.ca/inverse/fingerbank-collector/log"
	"git.inverse.ca/inverse/fingerbank-collector/timedlock"

	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"
	"github.com/google/gopacket/pcap"
	"github.com/inconshreveable/log15"
	"github.com/inverse-inc/go-utils/sharedutils"
	"github.com/mailru/easyjson/jlexer"
)

var peerHttpClient *http.Client

const (
	PEER_DIAL_TIMEOUT = 1 * time.Second
)

func init() {
	peerDialTimeout := func(network, addr string) (net.Conn, error) {
		return net.DialTimeout(network, addr, PEER_DIAL_TIMEOUT)
	}
	tr := &http.Transport{
		Dial:            peerDialTimeout,
		TLSClientConfig: &tls.Config{InsecureSkipVerify: true},
	}
	peerHttpClient = &http.Client{Transport: tr}
}

var edc = EndpointDataComparator{}

var arpLookup = flag.Bool(
	"arp-lookup",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_ARP_LOOKUP", "false")),
	"Enables ARP lookups when no IP/MAC match is found (COLLECTOR_ARP_LOOKUP envrionment variable)",
)

var localIPLookup = flag.Bool(
	"local-ip-lookup",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_LOCAL_IP_LOOKUP", "true")),
	"Enables IP lookups on the collector OS when no IP/MAC match is found (COLLECTOR_LOCAL_IP_LOOKUP environment variable)",
)

var fakeMac = flag.Bool(
	"fake-mac",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_FAKE_MAC", "false")),
	"Creates a fake IP/MAC association when the MAC address isn't found for an IP. Do not use in production unless you know what you are doing. (COLLECTOR_FAKE_MAC envrionment variable)",
)

var fakeMacPrivateOnly = flag.Bool(
	"fake-mac-private-only",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_FAKE_MAC_PRIVATE_ONLY", "true")),
	"Limits the fake-mac option to only apply for IP addresses that are in a private network. (COLLECTOR_FAKE_MAC_PRIVATE_ONLY environment variable)",
)

var clustered = flag.Bool(
	"clustered",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_CLUSTERED", "false")),
	"Whether or not the collector should be working in cluster mode (sharing information with other collectors that use the same Fingerbank API key). (COLLECTOR_CLUSTERED environment variable)",
)

var clusterInitialSyncWait = flag.Duration(
	"cluster-initial-sync-wait",
	MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_CLUSTER_INITIAL_SYNC_WAIT", "90s")),
	"The amount of time to wait on startup to find the oldest peer to perform the initial sync from. (COLLECTOR_CLUSTER_INITIAL_SYNC_WAIT environment variable)",
)

var clusterResyncInterval = flag.Duration(
	"cluster-resync-interval",
	MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_CLUSTER_RESYNC_INTERVAL", "120s")),
	"The amount of time (in seconds) to wait between resyncs between the different nodes of the cluster (if enabled) to ensure consistency between the peers. A value of 0 disables this. (COLLECTOR_CLUSTER_RESYNC_INTERVAL environment variable)",
)

var resetEndpointDhcpVendorChange = flag.Bool(
	"reset-endpoint-dhcp-vendor-change",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_RESET_ENDPOINT_DHCP_VENDOR_CHANGE", "true")),
	"Whether or not the endpoint data should be reseted after detecting a change in the DHCP vendor. (COLLECTOR_RESET_ENDPOINT_DHCP_VENDOR_CHANGE environment variable)",
)

var contactPeers = flag.String(
	"contact-peers",
	sharedutils.EnvOrDefault("COLLECTOR_CONTACT_PEERS", ""),
	"A comma delimited list of IP addresses to contact to establish clustering communication. By default, the collector will contact all the peers on the same layer 2 network through broadcast packets. (COLLECTOR_CONTACT_PEERS environment variable)",
)

var clusterHelloInterval = flag.Duration(
	"clustering-hello-interval",
	MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_CLUSTERING_HELLO_INTERVAL", fmt.Sprintf("%ds", fcreplication.DEFAULT_HELLO_INTERVAL))),
	`The interval at which HELLO packets are broadcasted on the network to discover new clustering peers. (COLLECTOR_CLUSTERING_HELLO_INTERVAL environment variable)`,
)

var clusterHelloBroadcast = flag.Bool(
	"clustering-hello-broadcast",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_CLUSTERING_HELLO_BROADCAST", "true")),
	`Whether or not the collector should broadcast discovery probes on the network to find new peers. If disabled, make sure you use the 'contact-peers' parameters to allow peer discovery via unicast. (COLLECTOR_CLUSTERING_HELLO_BROADCAST environment variable)`,
)

var odaPeersDiscoveryURL = flag.String(
	"oda-peers-discovery-url",
	sharedutils.EnvOrDefault("COLLECTOR_ODA_PEERS_DISCOVERY_URL", ""),
	`A URL to call to discover ODA peers. Advanced parameter that is used in very particular clustering scenarios. (COLLECTOR_ODA_PEERS_DISCOVERY_URL environment variable)`,
)

var endpointsCacheSize = flag.Int(
	"endpoints-cache-size",
	sharedutils.EnvOrDefaultInt("COLLECTOR_ENDPOINTS_CACHE_SIZE", 500),
	"The size of the endpoints cache. When the collector sees more endpoints than this number it will start to discard some non-critial data points like the full destination hosts and the IP/host cache of the endpoint. (COLLECTOR_ENDPOINTS_CACHE_SIZE environment variable)",
)

var ipHostCacheSize = flag.Int(
	"ip-host-cache-size",
	sharedutils.EnvOrDefaultInt("COLLECTOR_IP_HOST_CACHE_SIZE", 50),
	"The size of the IP/host cache per endpoint. Used to determine the DNS name (host) that is tied to an IP for an endpoint. (COLLECTOR_IP_HOST_CACHE_SIZE environment variable)",
)

var allHostsCacheSize = flag.Int(
	"all-hosts-cache-size",
	sharedutils.EnvOrDefaultInt("COLLECTOR_ALL_HOSTS_CACHE_SIZE", 50),
	"The size of the full hosts cache per endpoint. Will track all destinations of an endpoint up to this size. (COLLECTOR_ALL_HOSTS_CACHE_SIZE environment variable)",
)

var recordAllPortsInAllHostsCache = flag.Bool(
	"record-all-ports-in-all-hosts-cache",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_RECORD_ALL_PORTS_IN_ALL_HOSTS_CACHE", "false")),
	"When enabled, the collector will record all the ports that are seen in the all hosts cache even if no IP->DNS name translation can be made (COLLECTOR_RECORD_ALL_PORTS_IN_ALL_HOSTS_CACHE envrionment variable)",
)

var recordSrcDstPortsInEndpoint = flag.Bool(
	"record-src-dst-ports-in-endpoint",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_RECORD_SRC_DST_PORTS_IN_ENDPOINT", "true")),
	"When enabled, the collector will record all the source and destination ports that are lower than ephemeral-ports-start.",
)

var maxTrackSrcDstPortsInEndpoint = flag.Int(
	"max-track-src-dst-ports-in-endpoint",
	sharedutils.EnvOrDefaultInt("COLLECTOR_MAX_TRACK_SRC_DST_PORTS_IN_ENDPOINT", 5),
	"Maximum of ports to track for src and dst ports tracking. If set to 5, there will be 5 src and 5 dst ports tracked per endpoint.",
)

var ephemeralPortsStart = flag.Int(
	"ephemeral-ports-start",
	sharedutils.EnvOrDefaultInt("COLLECTOR_EPHEMERAL_PORTS_START", 32768),
	"When enabled, the collector will record all the source and destination ports that are lower than ephemeral-ports-start.",
)

var syncAllDestHostsHits = flag.Bool(
	"sync-all-dest-hosts-hits",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_SYNC_ALL_DEST_HOSTS_HITS", "false")),
	`When enabled, it will cluster sync the destination host hit in the LRU cache everytime its seen to keep the LRU buckets in sync between the members. This introduces a lot of sync overhead when seeing a lot of destination hosts traffic (SNI, DNS, etc)`,
)

var syncAllIpHostMappingHits = flag.Bool(
	"sync-all-ip-host-mapping-hits",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_SYNC_ALL_IP_HOST_MAPPING_HITS", "false")),
	`When enabled, it will cluster sync the ip/host hit in the LRU cache everytime its seen to keep the LRU buckets in sync between the members. This introduces a lot of sync overhead when seeing a lot of DNS traffic and TLS SNI traffic`,
)

// Dummy collector for use in testing
type FingerbankCollector struct {
	// Storage related fields
	updateIPMACLock *timedlock.RWLock

	ip2macMap map[string]string
	mac2ipMap map[string]string

	ipv62macMap map[string]string
	mac2ipv6Map map[string]string

	endpointsDataLock *timedlock.RWLock

	EndpointsData MapEndpointsData
	// End of storage related fields

	// PCAP related fields
	pcapHandle *pcap.Handle
	handlers   map[*pcap.BPF]ProtocolHandler

	// The list of interfaces the collector listens on
	Interfaces []*net.Interface
	// The list of IP addresses the collector listens on with their associated MAC
	ipAddresses []ipmacCouple

	// The Fingerbank API key used by this collector
	apiKey string

	// The replication controller
	ReplicationController *fcreplication.Controller

	// The collector's logger
	logger log15.Logger

	// Whether or not the API runs using HTTPS (so that it uses it when talking to other collectors)
	ApiUsesHTTPS bool

	// The endpoint cache that can be distributed accross cluster members (used for IP/host caching currently)
	EndpointsCache *EndpointsCache

	// The list of blacklisted IPs
	BlacklistedIPs []*net.IPNet

	NBAEnabled  bool
	nbaPolicies []*NBAPolicy

	recordedDestinationHosts uint32

	dpiEngine dpi.Engine
}

type ipmacCouple struct {
	ip  net.IP
	mac string
}

// Initialize a new FingerbankCollector
func NewFingerbankCollector(apiKey string) *FingerbankCollector {
	fc := &FingerbankCollector{}

	fc.logger = fclog.Logger("")

	fc.logger.Debug("testing")

	fc.updateIPMACLock = timedlock.NewRWLock()
	fc.updateIPMACLock.Panic = true
	fc.updateIPMACLock.PrintErrors = true

	fc.ip2macMap = make(map[string]string)
	fc.mac2ipMap = make(map[string]string)

	fc.ipv62macMap = make(map[string]string)
	fc.mac2ipv6Map = make(map[string]string)

	fc.endpointsDataLock = timedlock.NewRWLock()
	fc.endpointsDataLock.Panic = true
	fc.endpointsDataLock.PrintErrors = true

	fc.EndpointsData = make(MapEndpointsData)

	fc.handlers = make(map[*pcap.BPF]ProtocolHandler)

	fc.EndpointsCache = NewEndpointsCache(*endpointsCacheSize, *ipHostCacheSize, *allHostsCacheSize)

	fc.apiKey = apiKey

	fc.NBAEnabled = *nbaEnabled

	if fc.NBAEnabled {
		var err error
		fc.nbaPolicies, err = fc.buildNBAPoliciesFromFile(*nbaPoliciesLocation)
		if err != nil {
			fc.logger.Error("Error building NBA policies: " + err.Error())
		}
	}

	if *clustered {
		fc.setupReplication()
	}

	fc.initDPI()

	return fc
}

func (fc *FingerbankCollector) loadIpMaps(buffer io.Reader) error {
	dec := json.NewDecoder(buffer)
	maps := make(map[string]map[string]string)

	err := dec.Decode(&maps)

	if err != nil {
		return err
	}

	lockid := fc.updateIPMACLock.Lock()
	oldTimeout := fc.updateIPMACLock.Timeout
	fc.updateIPMACLock.Timeout = 60 * time.Second
	defer func() {
		fc.updateIPMACLock.Unlock(lockid)
		fc.updateIPMACLock.Timeout = oldTimeout
	}()

	// We replace the whole maps
	fc.ip2macMap = maps["ipv4_macs"]
	fc.ipv62macMap = maps["ipv6_macs"]

	fc.mac2ipMap = make(map[string]string)
	for ip, mac := range fc.ip2macMap {
		fc.mac2ipMap[mac] = ip
	}

	fc.mac2ipv6Map = make(map[string]string)
	for ip, mac := range fc.ipv62macMap {
		fc.mac2ipv6Map[mac] = ip
	}

	return nil
}

// Creates a new EndpointData with the update hook set to the one of this collector
func (fc *FingerbankCollector) NewEndpointData() *EndpointData {
	ed := NewEndpointData(fc.updateEndpointDataHook, fc.analyseEndpointHook, fc.runsInLiveReplication)
	return ed
}

func (fc *FingerbankCollector) runsInLiveReplication() bool {
	return fc.ReplicationController != nil && !fc.ReplicationController.IsInODAMode()
}

func (fc *FingerbankCollector) analyseEndpointHook(ed *EndpointData) {
	go fc.AnalyseEndpoint(ed)
}

// Update hook that is called after an EndpointData is changed/updated
// Lock is held on both endpoints, so this must be quick
func (fc *FingerbankCollector) updateEndpointDataHook(edBefore, edAfter *EndpointData, bsEvents []string) {
	// No need to do anything if there is no replication controller
	if fc.ReplicationController == nil || fc.ReplicationController.IsInODAMode() {
		return
	}

	_, delta, err := edc.BuildDeltaEndpoint(edBefore, edAfter)
	if err != nil {
		fc.logger.Error("Error while building delta of endpoint: " + err.Error())
		return
	}

	mac := edAfter.mac
	deltaJSON, err := json.Marshal(delta)

	go func() {
		if err != nil {
			fc.logger.Error(fmt.Sprintf("Error while encoding delta of endpoint: %s", err))
		}

		fc.ReplicationController.SendContentToOutboundFriends(
			fcreplication.REP_ENDPOINT_DATA_UPDATE,
			fcreplication.REPEndpointDataUpdate{
				Timestamp:             time.Now(),
				DeltaEndpointDataJSON: deltaJSON,
				MacAddress:            mac,
				BootSequenceEvents:    bsEvents,
			},
		)
	}()

}

func (fc *FingerbankCollector) CheckDhcpVendorChange(ed *EndpointData, dhcpVendorBefore string) {
	if *resetEndpointDhcpVendorChange {
		mac := ed.mac

		if dhcpVendorBefore == "" || ed.DhcpVendor == "" {
			return
		}

		if dhcpVendorBefore != ed.DhcpVendor {
			fc.logger.Info(fmt.Sprintf("Detected DHCP vendor change for %s. Will reinititialize endpoint.", mac))
			newEd := fc.NewEndpointData()

			// Keep necessary information in new endpoint
			newEd.UUID = ed.UUID
			newEd.DhcpFingerprint = ed.DhcpFingerprint
			newEd.DhcpVendor = ed.DhcpVendor
			newEd.Hostname = ed.Hostname
			newEd.OUI = ed.OUI
			newEd.mac = ed.mac

			oldEd := *ed
			*ed = *newEd
			ed.lock = oldEd.lock

			fc.EndpointsCache.ResetEndpoint(mac)

			// No need to continue if there is no replication controller
			if fc.ReplicationController == nil {
				return
			}

			edJSON, err := json.Marshal(ed)

			go func() {
				if err != nil {
					fc.logger.Error(fmt.Sprintf("Error while encoding replacement for endpoint: %s", err))
					return
				}
				fc.ReplicationController.SendContentToOutboundFriends(
					fcreplication.REP_ENDPOINT_DATA_REPLACE,
					fcreplication.REPEndpointDataReplace{
						Timestamp:        time.Now(),
						EndpointDataJSON: edJSON,
						MacAddress:       mac,
					},
				)

				fc.ReplicationController.SendContentToOutboundFriends(
					fcreplication.REP_RESET_ENDPOINT_CACHE,
					fcreplication.REPResetEndpointCache{
						Timestamp:  time.Now(),
						EndpointId: mac,
					},
				)
			}()
		}
	}
}

// Lock the read-lock on the endpoints data backend
func (fc *FingerbankCollector) EndpointsDataRLock() uint64 {
	return fc.endpointsDataLock.RLock()
}

// Unlock the read-lock on the endpoints data backend
func (fc *FingerbankCollector) EndpointsDataRUnlock(id uint64) {
	fc.endpointsDataLock.RUnlock(id)
}

func (fc *FingerbankCollector) AllEndpointsData() map[string]*EndpointData {
	return fc.EndpointsData
}

// Deletes an endpoint from the collector, making sure to acquire a RW lock before
func (fc *FingerbankCollector) DeleteEndpointData(ed *EndpointData) bool {
	//TODO: make this more efficient
	lockid := fc.endpointsDataLock.Lock()
	defer fc.endpointsDataLock.Unlock(lockid)

	result := fc.deleteEndpointData(ed)

	if result {
		go fc.syncEndpointDataDelete(ed)
	}

	return result
}

// Deletes an endpoint from the collector
func (fc *FingerbankCollector) deleteEndpointData(ed *EndpointData) bool {
	//TODO: make this more efficient
	var deletedEndpoint *EndpointData
	for key, stored := range fc.EndpointsData {
		if stored.UUID == ed.UUID {
			deletedEndpoint = fc.EndpointsData[key]
			delete(fc.EndpointsData, key)
		}
	}

	return deletedEndpoint != nil
}

// Update the IP/MAC couple in the collector using its in-memory maps
func (fc *FingerbankCollector) updateIPMAC(ip string, mac string) error {
	mac = CleanMAC(mac)
	if !ValidMAC(mac) {
		return errors.New("Invalid MAC address provided")
	}

	var previousMac string
	if ValidIPv6(ip) {
		ip = CleanIPv6(ip)
		lockid := fc.updateIPMACLock.Lock()

		previousMac = fc.ipv62macMap[ip]

		fc.ipv62macMap[ip] = mac
		fc.mac2ipv6Map[mac] = ip
		fc.updateIPMACLock.Unlock(lockid)
	} else if ValidIP(ip) {
		lockid := fc.updateIPMACLock.Lock()

		previousMac = fc.ip2macMap[ip]

		fc.ip2macMap[ip] = mac
		fc.mac2ipMap[mac] = ip
		fc.updateIPMACLock.Unlock(lockid)
	} else {
		return errors.New("Invalid IP address provided")
	}

	// Update the endpoint to expire its cache if the MAC changed
	// If the IP changed, we don't expire since the physical device is the same
	ed, _ := fc.FindEndpointDataMAC(mac)
	ed.Update(func(ed *EndpointData) bool {
		return previousMac != mac
	})

	return nil
}

// Update the IP/MAC couple in the collector using its in-memory maps
// Also take care of synchronizing the peers if needed
func (fc *FingerbankCollector) UpdateIPMAC(ip string, mac string) error {
	mac = CleanMAC(mac)
	if !ValidMAC(mac) {
		return errors.New("Invalid MAC address provided")
	}

	err := fc.updateIPMAC(ip, mac)

	if err != nil {
		return err
	}

	// Asynchronously update the IP/MAC couple on other peers
	if fc.ReplicationController != nil {
		go fc.syncIpMacUpdate(ip, mac)
	}

	return nil
}

// Find an EndpointData by its MAC address, creates and returns a new EndpointData if there is none in the list
func (fc *FingerbankCollector) FindEndpointDataMAC(mac string) (*EndpointData, error) {
	mac = CleanMAC(mac)
	if !ValidMAC(mac) {
		return nil, errors.New("Invalid MAC address provided")
	}

	if fc.EndpointsData == nil {
		fc.logger.Warn("Reinitializing nil EndpointsData map")
		fc.EndpointsData = make(MapEndpointsData)
	}

	lockid := fc.endpointsDataLock.RLock()
	ed, ok := fc.EndpointsData[mac]
	fc.endpointsDataLock.RUnlock(lockid)
	if !ok {
		ed = fc.NewEndpointData()
		// New Endpoint nobody has access to it so we can do what we want
		ed.OUI = mac[0:6]
		ed.mac = mac

		// Aquire lock before writting to the endpoint data map
		lockid := fc.endpointsDataLock.Lock()
		defer fc.endpointsDataLock.Unlock(lockid)
		// Check to make sure nobody updated in the mean time
		ed2, ok := fc.EndpointsData[mac]
		if !ok {
			fc.EndpointsData[mac] = ed
			ip, err := fc.MAC2IP(mac)
			if ip == "" || err != nil {
				ip = "Unknown IP"
			}
			ed.BootSequence.Start()
			fc.logger.Info(fmt.Sprintf("Discovered new endpoint: %s - %s", mac, ip))
		} else {
			// Already there return what is in the map
			ed = ed2
		}
	}
	return ed, nil
}

// Find an EndpointData by its IP address, creates and returns a new EndpointData if there is none in the list
// Will perform an IP2MAC and then call FindEndpointDataMAC to get it from the EndpointsData map
// IP can be an IPv4 or IPv6 address
// TODO: rework this to use a better implementation
func (fc *FingerbankCollector) FindEndpointDataIP(ip string) (*EndpointData, error) {
	if !ValidIP(ip) && !ValidIPv6(ip) {
		return nil, errors.New("Invalid IP address provided")
	}

	if mac, err := fc.IP2MAC(ip); err != nil {
		return nil, err
	} else {
		return fc.FindEndpointDataMAC(mac)
	}
}

// Find a MAC address given an IP address
// Returns an error if its unknown
// Supports using an IPv6 as the IP to search
// TODO: rework this to use a better implementation
func (fc *FingerbankCollector) IP2MAC(ip string) (string, error) {
	var mac string
	var errCollector error
	if mac, errCollector = fc.ip2mac(ip); ip != "" && errCollector == nil {
		return mac, errCollector
	} else if mac, _ = fc.ip2macInterfaces(ip); mac != "" {
		return mac, nil
	} else if mac, _ = fc.ip2macARP(ip); mac != "" {
		return mac, nil
	} else if *fakeMac {
		ipo := net.ParseIP(ip)
		if !*fakeMacPrivateOnly || ipo.IsPrivate() {
			var mac string
			var err error
			if ValidIP(ip) {
				mac, err = FakeMAC(ipo)
			} else {
				mac, err = RandomMAC()
			}
			if err == nil {
				fc.UpdateIPMAC(ip, mac)
				return mac, nil
			} else {
				return mac, err
			}
		}
	}
	// Default to returning the error we got from the collector
	return "", errCollector
}

// Find a MAC address given an IP address
// Returns an error if its unknown
func (dc *FingerbankCollector) ip2mac(ip string) (string, error) {
	if ValidIP(ip) {
		lockid := dc.updateIPMACLock.RLock()
		mac, ok := dc.ip2macMap[ip]
		dc.updateIPMACLock.RUnlock(lockid)
		if ok {
			return mac, nil
		} else {
			return "", errors.New("Cannot find IP in ip2macMap")
		}
	} else if ValidIPv6(ip) {
		lockid := dc.updateIPMACLock.RLock()
		mac, ok := dc.ipv62macMap[ip]
		dc.updateIPMACLock.RUnlock(lockid)
		if ok {
			return mac, nil
		} else {
			return "", errors.New("Cannot find IPv6 in ipv62macMap")
		}
	} else {
		return "", errors.New("Invalid IP address provided")
	}
}

func (fc *FingerbankCollector) PopulateIpAddresses() {
	fc.ipAddresses = make([]ipmacCouple, 0)

	if !*localIPLookup {
		return
	}

	fc.logger.Info("Populating IP addresses of where the collector is running")
	for _, i := range fc.Interfaces {
		addrs, err := i.Addrs()
		if err != nil {
			panic(err)
		}

		for _, addr := range addrs {
			var ip net.IP
			switch v := addr.(type) {
			case *net.IPNet:
				ip = v.IP
			case *net.IPAddr:
				ip = v.IP
			}
			mac := CleanMAC(i.HardwareAddr.String())
			if !ip.IsLoopback() && mac != "" {
				fc.ipAddresses = append(fc.ipAddresses, ipmacCouple{ip: ip, mac: mac})
			}
		}
	}
}

// Perform an IP2MAC by looking at all the interfaces running on this endpoint since the IP may belong to the collector itself
func (fc *FingerbankCollector) ip2macInterfaces(ip string) (string, error) {
	if !*localIPLookup {
		return "", nil
	}

	//fc.logger.Debug(fmt.Sprintf("Attempting to resolve IP %s using local interfaces", ip))

	netip := net.ParseIP(ip)

	for _, info := range fc.ipAddresses {
		if netip.Equal(info.ip) {
			fc.logger.Info(fmt.Sprintf("Found IP/MAC (ip:%s|mac:%s) on the local interface", info.ip, info.mac))
			err := fc.UpdateIPMAC(info.ip.String(), info.mac)
			if err != nil {
				fc.logger.Info(fmt.Sprintf("Failed to update IP/MAC from local interface result. Error: %s", err))
			}
			return info.mac, nil
		}
	}

	return "", errors.New("Cannot find IP on local interfaces")
}

// Perform an IP2MAC using ARP on all the interfaces the collector is listening on
func (fc *FingerbankCollector) ip2macARP(ip string) (string, error) {
	cacheKey := "arp-lookup-" + ip

	if !*arpLookup {
		return "", nil
	}

	if x, found := Cache.Get(cacheKey); found {
		return x.(string), nil
	}

	//fc.logger.Debug(fmt.Sprintf("Attempting to resolve IP %s using ARP query", ip))
	for _, device := range fc.Interfaces {
		if mac, _ := ARPFindMACNotRouter(device, ip); mac != "" {
			fc.logger.Info(fmt.Sprintf("Found an ARP match using the ARP query (ip:%s|mac:%s)", ip, mac))
			err := fc.UpdateIPMAC(ip, mac)
			if err != nil {
				fc.logger.Error(fmt.Sprintf("Failed to update IP/MAC from ARP result. Error: %s", err))
			}
			Cache.Set(cacheKey, mac, 5*time.Minute)
			return mac, nil
		}
	}
	Cache.Set(cacheKey, "", 10*time.Minute)
	return "", errors.New("Cannot find IP using ARP")
}

// Find an IP address given an MAC address
// Returns an error if its unknown
func (fc *FingerbankCollector) MAC2IP(mac string) (string, error) {
	mac = CleanMAC(mac)
	if !ValidMAC(mac) {
		return "", errors.New("Invalid MAC address provided")
	}

	lockid := fc.updateIPMACLock.RLock()
	ip, ok := fc.mac2ipMap[mac]
	fc.updateIPMACLock.RUnlock(lockid)
	if ok {
		return ip, nil
	} else {
		return "", errors.New("Cannot find IP in mac2ipMap")
	}
}

// Find an IPv6 address given an MAC address
// Returns an error if its unknown
// Only returns the IPv6 IP address
func (fc *FingerbankCollector) MAC2IPv6(mac string) (string, error) {
	mac = CleanMAC(mac)
	if !ValidMAC(mac) {
		return "", errors.New("Invalid MAC address provided")
	}

	lockid := fc.updateIPMACLock.RLock()
	ip, ok := fc.mac2ipv6Map[mac]
	fc.updateIPMACLock.RUnlock(lockid)
	if ok {
		return ip, nil
	} else {
		return "", errors.New("Cannot find IP in mac2ipv6Map")
	}
}

func (fc *FingerbankCollector) addIpHostMapping(endpointId string, ip string, host string) bool {
	return fc.EndpointsCache.AddIpHostToEndpoint(endpointId, ip, host)
}
func (fc *FingerbankCollector) AddIpHostMapping(endpointId string, ip string, host string) {
	exists := fc.addIpHostMapping(endpointId, ip, host)
	// Asynchronously update the IP/host couple on other peers
	if fc.ReplicationController != nil {
		if *syncAllIpHostMappingHits || !exists {
			go fc.syncAddIpHostMapping(endpointId, ip, host)
		}
	}
}

// Register a protocol handler in this collector
func (fc *FingerbankCollector) RegisterProtocolHandler(ph ProtocolHandler, bpfString string) error {
	bpf, err := fc.pcapHandle.NewBPF(bpfString)

	ph.SetCollector(fc)

	if err != nil {
		return err
	}

	fc.handlers[bpf] = ph

	return nil
}

// Get the BPF that represents all the protocol handlers registered BPFs
func (fc *FingerbankCollector) HandlersBPF() string {
	var allBpf []string
	for bpf, _ := range fc.handlers {
		allBpf = append(allBpf, "("+bpf.String()+")")
	}
	return strings.Join(allBpf, " or ")
}

// Set the PCAP handle for this collector
// Used to match BPFs against packets for proper dispatching of the packets to the right collectors
func (fc *FingerbankCollector) SetPcapHandle(handle *pcap.Handle) error {
	if fc.pcapHandle != nil {
		return errors.New("PCAP handle is already registered. Create a new collector to change the PCAP handle")
	} else {
		fc.pcapHandle = handle
		return nil
	}
}

// Get the PacketInfo for a gopacket.Packet
// Will populate the IPv4 or v6 source and destination address if its an IP packet
// Will populate the source + destination MAC address if its an Ethernet packet
// Will populate the source + destination port if its a TCP or UDP packet
func (fc *FingerbankCollector) getPacketInfo(gp gopacket.Packet) PacketInfo {
	pi := PacketInfo{}

	if layer := gp.Layer(layers.LayerTypeEthernet); layer != nil {
		eth := layer.(*layers.Ethernet)
		pi.SrcMAC = eth.SrcMAC
		pi.DstMAC = eth.DstMAC
	}

	if layer := gp.Layer(layers.LayerTypeIPv4); layer != nil {
		ip := layer.(*layers.IPv4)
		pi.SrcIP = ip.SrcIP.String()
		pi.NetSrcIP = ip.SrcIP
		pi.DstIP = ip.DstIP.String()
		pi.NetDstIP = ip.DstIP
	}

	if layer := gp.Layer(layers.LayerTypeIPv6); layer != nil {
		ip := layer.(*layers.IPv6)
		pi.SrcIP = CleanIPv6(ip.SrcIP.String())
		pi.DstIP = CleanIPv6(ip.DstIP.String())
	}

	if layer := gp.Layer(layers.LayerTypeTCP); layer != nil {
		tcp, _ := layer.(*layers.TCP)
		pi.SrcPort = int(tcp.SrcPort)
		pi.DstPort = int(tcp.DstPort)
		pi.Proto = "tcp"
	}

	if layer := gp.Layer(layers.LayerTypeUDP); layer != nil {
		udp, _ := layer.(*layers.UDP)
		pi.SrcPort = int(udp.SrcPort)
		pi.DstPort = int(udp.DstPort)
		pi.Proto = "udp"
	}

	return pi
}

func (fc *FingerbankCollector) HandleIPPacket(proto string, srcIP, dstIP net.IP, srcPort, dstPort int) {
	srcEndpoint, srcEndpointErr := fc.FindEndpointDataIP(srcIP.String())

	if *recordSrcDstPortsInEndpoint {
		if srcEndpointErr == nil {
			srcEndpoint.RecordPorts(proto, srcPort, dstPort)
		}
		dstEndpoint, dstEndpointErr := fc.FindEndpointDataIP(dstIP.String())
		if dstEndpointErr == nil {
			dstEndpoint.RecordPorts(proto, dstPort, srcPort)
		}
	}

	if srcEndpointErr != nil {
		return
	}

	host, err := fc.EndpointIpToHost(srcEndpoint.GetMac(), dstIP.String())
	if err == nil {
		fc.AddDestinationHost(srcEndpoint, host, dstPort, proto)
	} else if *recordAllPortsInAllHostsCache {
		fc.AddDestinationHost(srcEndpoint, "*", dstPort, proto)
	}

	if fc.NBAEnabled {
		fc.MatchAgainstBlacklist(srcIP, dstIP, srcPort, dstPort)
	}
}

// Handle a gopacket.Packet and send it to the registered handlers
func (fc *FingerbankCollector) HandleGoPacket(ift *net.Interface, gp gopacket.Packet) error {
	captureInfo := gp.Metadata().CaptureInfo
	rawPacket := gp.Data()

	// The BPF filters do not work against tagged packets for a reason we still ignore
	// For this reason, we reset the ethernet type to ipv4 and remove the 802.1q bytes
	var bpfMatchPacket []byte
	if rawPacket[12] == 0x81 && rawPacket[13] == 0x00 {
		toTake := 46
		bpfMatchPacket = make([]byte, toTake)

		for i := 0; i < toTake; i++ {
			bpfMatchPacket[i] = rawPacket[i]
		}

		bpfMatchPacket = append(bpfMatchPacket[:14], bpfMatchPacket[14+4:]...)
		bpfMatchPacket[12] = 0x08
	} else {
		bpfMatchPacket = rawPacket
	}

	l4Payload := []byte("")
	appLayer := gp.TransportLayer()
	if appLayer != nil {
		l4Payload = gp.TransportLayer().LayerPayload()
	} else {
		fc.logger.Debug("Cannot get payload from packet")
	}

	packetInfo := fc.getPacketInfo(gp)

	fc.HandleIPPacket(packetInfo.Proto, packetInfo.NetSrcIP, packetInfo.NetDstIP, packetInfo.SrcPort, packetInfo.DstPort)

	if fc.dpiEngine != nil {
		fc.dpiEngine.ProcessPkt(&DPIMetadata{
			SrcPort: uint16(packetInfo.SrcPort),
			SrcIP:   packetInfo.NetSrcIP,
			DstPort: uint16(packetInfo.DstPort),
			DstIP:   packetInfo.NetDstIP,
			Proto:   packetInfo.Proto,
		}, gp, bpfMatchPacket, captureInfo, int(layers.LinkTypeEthernet))
	}

	for bpf, handler := range fc.handlers {
		if bpf.Matches(captureInfo, bpfMatchPacket) {
			func() {
				defer func() {
					if r := recover(); r != nil {
						fc.logger.Error("=================================================")
						fc.logger.Error(fmt.Sprintf("Recovered panic for handler %s (%s)", handler, r))
						fc.logger.Error(string(debug.Stack()))
						fc.logger.Error("=================================================")
					}
				}()

				handler.Handle(packetInfo, l4Payload, gp.Data())
			}()
		}
	}
	return nil
}

func (fc *FingerbankCollector) CleanupInactiveEndpointData(d time.Duration) {
	lockid := fc.endpointsDataLock.Lock()

	ids := []string{}
	diff, _ := time.ParseDuration("-" + d.String())
	expiryTime := time.Now().Add(diff)
	for id, ed := range fc.EndpointsData {
		if ed.LastUpdated.Before(expiryTime) {
			fc.logger.Debug(fmt.Sprintf("Adding endpoint %s to be deleted since it hasn't been seen for %s.", id, d))
			ids = append(ids, id)
		}
	}

	fc.endpointsDataLock.Unlock(lockid)

	for _, id := range ids {
		fc.logger.Info(fmt.Sprintf("Endpoint %s hasn't been seen for %s. Deleting it", id, d))
		ed, err := fc.FindEndpointDataMAC(id)
		if err != nil {
			fc.logger.Error(fmt.Sprintf("Cannot delete %s because endpoint cannot be found (%s)", id, err))
			continue
		}
		fc.DeleteEndpointData(ed)
	}
}

func (fc *FingerbankCollector) EndpointIpToHost(endpointId string, ip string) (string, error) {
	if ihc, found := fc.EndpointsCache.GetEndpointIpHostCache(endpointId); found {
		if host, found := ihc.Get(ip); found {
			return host, nil
		} else {
			return "", errors.New("Unable to find IP in endpoint's IpHostCache")
		}
	} else {
		return "", errors.New("Unable to find IpHostCache of endpoint")
	}

}

func (fc *FingerbankCollector) ExpandEndpointData(mac string, ed *EndpointData) *ExpandedEndpointData {
	ip, _ := fc.MAC2IP(mac)
	ipv6, _ := fc.MAC2IPv6(mac)
	eed := &ExpandedEndpointData{EndpointData: *ed, MAC: mac, IP: ip, IPv6: ipv6, DiscoveredAt: ed.GetCreatedAt(), SecondsOnNetwork: ed.TimeOnNetwork().Seconds()}
	if ihc, found := fc.EndpointsCache.GetEndpointIpHostCache(mac); found {
		eed.IpHostCache = ihc.ToMap()
	} else {
		eed.IpHostCache = map[string]string{}
	}
	eed.AllHostsCache = ed.AllHostsCacheMap(fc)
	return eed
}

var queryCacheTime = flag.Duration(
	"query-cache-time",
	MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_QUERY_CACHE_TIME", "5m")),
	"Time during which the Fingerbank API query result is cache if the endpoint data hasn't changed (COLLECTOR_QUERY_CACHE_TIME environment variable)",
)

type QueryResult struct {
	EndpointData   *ExpandedEndpointData                 `json:"endpoint_data"`
	CloudAPIResult *fingerbankclient.InterrogateResponse `json:"cloud_api_result"`
}

func (fc *FingerbankCollector) addDestinationHost(endpointId, host string, port int, proto string) bool {
	return fc.EndpointsCache.AddHostToEndpoint(endpointId, host, fmt.Sprintf("%s:%d", proto, port))
}

func (fc *FingerbankCollector) AddDestinationHost(ed *EndpointData, host string, port int, proto string) {
	atomic.AddUint32(&fc.recordedDestinationHosts, 1)
	ed.AddDestinationHost(host, port)
	seq := strconv.FormatInt(time.Now().Unix()*1000, 10) + fmt.Sprintf(":all_destination_hosts:%s:%d", host, port)
	ed.AddBootSequence(seq)
	if *allHostsCacheSize != 0 {
		exists := fc.addDestinationHost(ed.mac, host, port, proto)

		// Asynchronously send the destination host for the AllHostsCache
		if fc.ReplicationController != nil && !fc.ReplicationController.IsInODAMode() {
			if *syncAllDestHostsHits || !exists {
				go fc.syncAddDestinationHost(ed.mac, host, port, proto)
			}
		}
	}
}

func (fc *FingerbankCollector) InterrogateAPI(deviceId string, ed *EndpointData) (QueryResult, error) {
	var err error
	var result QueryResult
	var mac string
	var cloudAPIResult *fingerbankclient.InterrogateResponse

	ed = fc.EndpointDataODA(ed)

	if o, found := Cache.Get("fingerbank-api-query-" + ed.UUID); found {
		result = o.(QueryResult)

		if !ed.GetLastUpdated().After(result.EndpointData.GetLastUpdated()) {
			fc.logger.Debug("Returning interrogate result from cache")
			cloudAPIResult = result.CloudAPIResult
		}
	}

	if ValidMAC(deviceId) {
		mac = deviceId
	} else {
		mac, _ = fc.IP2MAC(deviceId)
	}

	ed2 := ed.Copy()

	query := ed2.ToInterrogateQuery(fc)
	d, _ := json.Marshal(query)
	fc.logger.Debug(fmt.Sprintf("Performing the query using the following parameters: %s", string(d)))

	if cloudAPIResult == nil {
		cloudAPIResult, err = fingerbankclient.Interrogate(query)
	}

	if err != nil {
		return QueryResult{}, errors.New(fmt.Sprintf("Cannot communicate with Fingerbank API (%s)", err))
	}

	id := ed.RLock()
	result = QueryResult{EndpointData: fc.ExpandEndpointData(mac, ed), CloudAPIResult: cloudAPIResult}
	ed.RUnlock(id)

	Cache.Set("fingerbank-api-query-"+ed.UUID, result, *queryCacheTime)

	return result, nil
}

func (fc *FingerbankCollector) DeviceIsA(deviceId int, wantsDeviceId int) (bool, error) {

	cacheKey := fmt.Sprintf("fingerbank-api-device-is-a-%d-%d", deviceId, wantsDeviceId)
	if o, found := Cache.Get(cacheKey); found {
		result := o.(bool)
		fc.logger.Debug("Returning 'device is a' result from cache")
		return result, nil
	}

	cloudAPIResult, err := fingerbankclient.EndpointIsA(deviceId, wantsDeviceId)
	result := cloudAPIResult.Result

	if err != nil {
		return cloudAPIResult.Result, errors.New(fmt.Sprintf("Cannot communicate with Fingerbank API (%s)", err))
	}

	Cache.Set(cacheKey, result, *queryCacheTime)

	return result, nil
}

func (fc *FingerbankCollector) loadEndpointsData(buffer io.Reader) error {
	oldTimeout := fc.endpointsDataLock.Timeout
	oldRTimeout := fc.endpointsDataLock.RTimeout
	fc.endpointsDataLock.Panic = true
	fc.endpointsDataLock.Timeout = 60 * time.Second
	fc.endpointsDataLock.RTimeout = 60 * time.Second
	lockid := fc.endpointsDataLock.Lock()
	defer func() {
		fc.endpointsDataLock.Timeout = oldTimeout
		fc.endpointsDataLock.RTimeout = oldRTimeout
		fc.endpointsDataLock.Unlock(lockid)
	}()

	data, err := ioutil.ReadAll(buffer)
	fc.EndpointsData.UnmarshalEasyJSON(&jlexer.Lexer{Data: data})

	// Some of the internals like the gob encoder/decoder + the update hook need to be setup
	// Also the MAC needs to be set as its unexported
	for mac, ed := range fc.EndpointsData {
		ed.init(fc.updateEndpointDataHook, fc.analyseEndpointHook, fc.runsInLiveReplication)
		ed.Update(func(ed *EndpointData) bool {
			ed.mac = mac
			// We don't want to have the updates hooks be called or anything special
			return false
		})
	}

	if err != nil {
		msg := fmt.Sprintf("Error while decoding the endpoints data: %s", err)
		return errors.New(msg)
	} else {
		return nil
	}
}

func (fc *FingerbankCollector) ListEndpointsIds() []string {
	lockid := fc.EndpointsDataRLock()
	defer fc.EndpointsDataRUnlock(lockid)

	keys := make([]string, len(fc.EndpointsData))
	i := 0
	for k, _ := range fc.EndpointsData {
		keys[i] = k
		i++
	}

	return keys
}

func (fc *FingerbankCollector) IpMaps() map[string]map[string]string {
	endpoints := fc.ListEndpointsIds()

	ipv4Macs := make(map[string]string)
	ipv6Macs := make(map[string]string)
	for _, mac := range endpoints {
		ip, err := fc.MAC2IP(mac)
		if err == nil {
			ipv4Macs[ip] = mac
		}

		ip, err = fc.MAC2IPv6(mac)
		if err == nil {
			ipv6Macs[ip] = mac
		}
	}

	return map[string]map[string]string{"ipv4_macs": ipv4Macs, "ipv6_macs": ipv6Macs}
}

func (fc *FingerbankCollector) loadEndpointsCache(buffer io.Reader) error {
	dec := json.NewDecoder(buffer)
	return dec.Decode(fc.EndpointsCache)
}

func (fc *FingerbankCollector) Stats() map[string]interface{} {
	return map[string]interface{}{
		"endpoints_count":            len(fc.EndpointsData),
		"ip2mac_count":               len(fc.ip2macMap),
		"mac2ipv6_count":             len(fc.mac2ipv6Map),
		"ipv62mac_count":             len(fc.ipv62macMap),
		"mac2ip_count":               len(fc.mac2ipMap),
		"recorded_destination_hosts": atomic.LoadUint32(&fc.recordedDestinationHosts),
	}
}

func (fc *FingerbankCollector) Destroy() {
	if fc.dpiEngine != nil {
		fc.dpiEngine.Destroy()
	}
}
