package fccommon

import (
	fcreplication "git.inverse.ca/inverse/fingerbank-collector/common/replication"

	"encoding/json"
	"fmt"
	"time"
)

func (fc *FingerbankCollector) HandleREPEndpointDataUpdate(request interface{}) fcreplication.ReplicationMessage {
	updateRequest := request.(fcreplication.REPEndpointDataUpdate)
	fc.logger.Debug(fmt.Sprintf("Handling endpoint update for device: %s", updateRequest.MacAddress))

	endpoint, err := fc.FindEndpointDataMAC(updateRequest.MacAddress)
	if err != nil {
		fc.logger.Warn(fmt.Sprintf("Unable to find endpoint to update. MAC address: %s", updateRequest.MacAddress))
		return fcreplication.ReplicationMessage{}
	}

	lockid := endpoint.lock.Lock()
	defer endpoint.lock.Unlock(lockid)
	err = json.Unmarshal(updateRequest.DeltaEndpointDataJSON, endpoint)
	if err != nil {
		fc.logger.Error(fmt.Sprintf("Error while updating endpoint. Error: %s", err))
	}
	endpoint.BootSequence.AlwaysAdd(updateRequest.BootSequenceEvents)
	endpoint.LastUpdated = time.Now()

	return fcreplication.ReplicationMessage{}
}

func (fc *FingerbankCollector) HandleREPEndpointDataReplace(request interface{}) fcreplication.ReplicationMessage {
	updateRequest := request.(fcreplication.REPEndpointDataReplace)
	fc.logger.Debug(fmt.Sprintf("Handling endpoint replace for device: %s", updateRequest.MacAddress))

	endpoint, err := fc.FindEndpointDataMAC(updateRequest.MacAddress)
	if err != nil {
		fc.logger.Warn(fmt.Sprintf("Unable to find endpoint to update. MAC address: %s", updateRequest.MacAddress))
		return fcreplication.ReplicationMessage{}
	}

	lockid := endpoint.lock.Lock()
	defer endpoint.lock.Unlock(lockid)

	newEd := fc.NewEndpointData()

	err = json.Unmarshal(updateRequest.EndpointDataJSON, newEd)
	if err != nil {
		fc.logger.Error(fmt.Sprintf("Error while decoding endpoint for replacement. Error: %s", err))
	}
	endpoint.LastUpdated = time.Now()
	newEd.lock = endpoint.lock
	*endpoint = *newEd

	return fcreplication.ReplicationMessage{}
}

func (fc *FingerbankCollector) HandleREPEndpointDataDelete(request interface{}) fcreplication.ReplicationMessage {
	deleteRequest := request.(fcreplication.REPEndpointDataDelete)
	fc.logger.Debug(fmt.Sprintf("Handling endpoint delete for device: %s", deleteRequest.MacAddress))

	endpoint, err := fc.FindEndpointDataMAC(deleteRequest.MacAddress)
	if err != nil {
		fc.logger.Warn(fmt.Sprintf("Unable to find endpoint to delete. MAC address: %s", deleteRequest.MacAddress))
		return fcreplication.ReplicationMessage{}
	}

	lockid := fc.endpointsDataLock.Lock()
	defer fc.endpointsDataLock.Unlock(lockid)
	res := fc.deleteEndpointData(endpoint)
	if !res {
		fc.logger.Warn(fmt.Sprintf("Unable to delete endpoint %s", deleteRequest.MacAddress))
	}

	return fcreplication.ReplicationMessage{}
}

func (fc *FingerbankCollector) HandleREPIpMacUpdate(request interface{}) fcreplication.ReplicationMessage {
	updateRequest := request.(fcreplication.REPIpMacUpdate)
	ip := updateRequest.IP
	mac := updateRequest.MAC
	fc.logger.Debug(fmt.Sprintf("Handling IP/MAC (ip:%s|mac:%s) update", ip, mac))

	fc.updateIPMAC(ip, mac)

	return fcreplication.ReplicationMessage{}
}

func (fc *FingerbankCollector) HandleREPAddIpHostMapping(request interface{}) fcreplication.ReplicationMessage {
	updateRequest := request.(fcreplication.REPAddIpHostMapping)
	endpointId := updateRequest.EndpointId
	ip := updateRequest.IP
	host := updateRequest.Host
	fc.logger.Debug(fmt.Sprintf("Handling AddIpHostMapping (endpointId:%s|ip:%s|host:%s) update", endpointId, ip, host))

	fc.addIpHostMapping(endpointId, ip, host)

	return fcreplication.ReplicationMessage{}
}

func (fc *FingerbankCollector) HandleREPAddDestinationHost(request interface{}) fcreplication.ReplicationMessage {
	updateRequest := request.(fcreplication.REPAddDestinationHost)
	endpointId := updateRequest.EndpointId
	host := updateRequest.Host
	port := updateRequest.Port
	proto := updateRequest.Proto
	fc.logger.Debug(fmt.Sprintf("Handling AddDestinationHost (endpointId:%s|host:%s|port:%d|proto:%s) update", endpointId, host, port, proto))

	fc.addDestinationHost(endpointId, host, port, proto)

	return fcreplication.ReplicationMessage{}
}

func (fc *FingerbankCollector) HandleREPResetEndpointCache(request interface{}) fcreplication.ReplicationMessage {
	resetRequest := request.(fcreplication.REPResetEndpointCache)
	endpointId := resetRequest.EndpointId
	fc.logger.Debug(fmt.Sprintf("Handling ResetEndpointCache (endpointId:%s)", endpointId))

	fc.EndpointsCache.ResetEndpoint(endpointId)

	return fcreplication.ReplicationMessage{}
}
