package fccommon

import (
	"errors"
	"reflect"
	"time"
)

var INCOMPARABLE_FIELDS = map[string]bool{
	"UUID": true,
}

type EndpointDataComparator struct{}

// Compares two EndpointData and return a third structure representing the delta difference between the two EndpointData
// The attributes of the delta which are maps will only contain the keys that are different between the two EndpointData
// The second endpoint is assumed to be the most recent, so the values of the delta will be the ones of the second EndpointData
func (edc *EndpointDataComparator) BuildDeltaEndpoint(ed1, ed2 *EndpointData) (bool, *DeltaEndpointData, error) {
	red1 := reflect.ValueOf(ed1).Elem()
	red2 := reflect.ValueOf(ed2).Elem()

	delta := &DeltaEndpointData{}
	rdelta := reflect.ValueOf(delta).Elem()

	changed := false

	for i := 0; i < rdelta.NumField(); i++ {
		fieldName := rdelta.Type().Field(i).Name
		exported := rdelta.Type().Field(i).PkgPath == ""

		if !exported {
			continue
		}
		if _, ignored := INCOMPARABLE_FIELDS[fieldName]; ignored {
			continue
		}

		f1 := red1.FieldByName(fieldName)
		f2 := red2.FieldByName(fieldName)
		deltaf := rdelta.Field(i)
		var err error
		var localChanged bool
		switch f1.Kind() {
		default:
			localChanged, err = edc.deltaValueField(f1, f2, deltaf)
		case reflect.Map:
			localChanged, err = edc.deltaMapField(f1, f2, deltaf)
		}

		// Bail out on any field error
		if err != nil {
			return false, nil, err
		}

		if localChanged {
			changed = true
		}
	}

	return changed, delta, nil
}

// Compare two fields, then if they are different, set the deltaField value to the value of the 2nd field (assumed to be the most recent)
// The returned boolean defines whether or not the value was set in the deltaField
func (edc *EndpointDataComparator) deltaValueField(f1, f2, deltaField reflect.Value) (bool, error) {
	v1 := f1.Interface()
	v2 := f2.Interface()

	same, err := edc.compareValues(v1, v2)

	if err != nil {
		return false, err
	} else if !same {
		deltaField.Set(f2)
		return true, nil
	} else {
		return false, nil
	}
}

// Compare two maps, then if they are different, set the deltafield with a map that contains all the different keys+values between the maps
// The returned boolean defines whether or not the value was set in the deltaField
func (edc *EndpointDataComparator) deltaMapField(f1, f2, deltaField reflect.Value) (bool, error) {
	v1 := f1.Interface()
	v2 := f2.Interface()

	var deltaVal reflect.Value
	var changed bool

	switch tv1 := v1.(type) {
	default:
		return false, errors.New("Impossible to find type of map " + reflect.TypeOf(v1).String())
	case map[string]map[string]int:
		tv2 := v2.(map[string]map[string]int)
		deltaMap := make(map[string]map[string]int)

		diffKeys := []string{}
		// Assuming those maps can only grow, so looking at the dst endpoint data
		for k, _ := range tv2 {
			if diffKeysStrMap, err := edc.compareMapStringInt(tv1[k], tv2[k]); err != nil {
				return false, err
			} else if len(diffKeysStrMap) > 0 {
				diffKeys = append(diffKeys, k)
			}
		}

		for _, k := range diffKeys {
			changed = true
			deltaMap[k] = tv2[k]
		}
		deltaVal = reflect.ValueOf(deltaMap)

	case map[string]map[string]string:
		tv2 := v2.(map[string]map[string]string)
		deltaMap := make(map[string]map[string]string)

		diffKeys := []string{}
		// Assuming those maps can only grow, so looking at the dst endpoint data
		for k, _ := range tv2 {
			if diffKeysStrMap, err := edc.compareMapStringString(tv1[k], tv2[k]); err != nil {
				return false, err
			} else if len(diffKeysStrMap) > 0 {
				diffKeys = append(diffKeys, k)
			}
		}

		for _, k := range diffKeys {
			changed = true
			deltaMap[k] = tv2[k]
		}
		deltaVal = reflect.ValueOf(deltaMap)

	case map[int]int:
		tv2 := v2.(map[int]int)
		deltaMap := make(map[int]int)
		diffKeys, err := edc.compareMapIntInt(tv1, tv2)

		if err != nil {
			return false, err
		}

		for _, k := range diffKeys {
			changed = true
			deltaMap[k] = tv2[k]
		}
		deltaVal = reflect.ValueOf(deltaMap)

	case map[string]int:
		tv2 := v2.(map[string]int)
		deltaMap := make(map[string]int)
		diffKeys, err := edc.compareMapStringInt(tv1, tv2)

		if err != nil {
			return false, err
		}

		for _, k := range diffKeys {
			changed = true
			deltaMap[k] = tv2[k]
		}
		deltaVal = reflect.ValueOf(deltaMap)

	case map[string]map[int]int:
		tv2 := v2.(map[string]map[int]int)
		deltaMap := make(map[string]map[int]int)

		diffKeys := []string{}
		// Assuming those maps can only grow, so looking at the dst endpoint data
		for k, _ := range tv2 {
			if diffKeysIntMap, err := edc.compareMapIntInt(tv1[k], tv2[k]); err != nil {
				return false, err
			} else if len(diffKeysIntMap) > 0 {
				diffKeys = append(diffKeys, k)
			}
		}

		for _, k := range diffKeys {
			changed = true
			deltaMap[k] = tv2[k]
		}
		deltaVal = reflect.ValueOf(deltaMap)

	case map[string]map[int]TrackableTime:
		tv2 := v2.(map[string]map[int]TrackableTime)
		deltaMap := make(map[string]map[int]TrackableTime)

		diffKeys := []string{}
		// Assuming those maps can only grow, so looking at the dst endpoint data
		for k, _ := range tv2 {
			if diffKeysIntMap, err := edc.compareMapIntTime(tv1[k], tv2[k]); err != nil {
				return false, err
			} else if len(diffKeysIntMap) > 0 {
				diffKeys = append(diffKeys, k)
			}
		}

		for _, k := range diffKeys {
			changed = true
			deltaMap[k] = tv2[k]
		}
		deltaVal = reflect.ValueOf(deltaMap)

	case map[string]bool:
		tv2 := v2.(map[string]bool)
		deltaMap := make(map[string]bool)
		diffKeys, err := edc.compareMapStringBool(tv1, tv2)

		if err != nil {
			return false, err
		}

		for _, k := range diffKeys {
			changed = true
			deltaMap[k] = tv2[k]
		}
		deltaVal = reflect.ValueOf(deltaMap)
	}

	if deltaVal.IsValid() {
		deltaField.Set(deltaVal)
		return changed, nil
	} else {
		return false, nil
	}
}

// Compares two strings together, doesn't get any simpler than that
func (edc *EndpointDataComparator) compareString(s1, s2 string) (bool, error) {
	return s1 == s2, nil
}

// Compares two ints together, doesn't get any simpler than that
func (edc *EndpointDataComparator) compareInt(i1, i2 int) (bool, error) {
	return i1 == i2, nil
}

// Compares two bools together, doesn't get any simpler than that
func (edc *EndpointDataComparator) compareBool(b1, b2 bool) (bool, error) {
	return b1 == b2, nil
}

// Compares two values together. If they have the same type, their value is compared, otherwise it is assumed that their value is different.
func (edc *EndpointDataComparator) compareValues(v1, v2 interface{}) (bool, error) {
	if reflect.TypeOf(v1) != reflect.TypeOf(v2) {
		return false, nil
	}

	switch tv1 := v1.(type) {
	default:
		return false, errors.New("Can't find type of value")
	case time.Time:
		// Times are always considered equal for EndpointData because they track event timestamps which are (for now) specific to each EndpointData pointer
		return true, nil
	case TrackableTime:
		// Times are always considered equal for EndpointData because they track event timestamps which are (for now) specific to each EndpointData pointer
		return tv1.UnixNano() == v2.(TrackableTime).UnixNano(), nil
	case string:
		return edc.compareString(tv1, v2.(string))
	case int:
		return edc.compareInt(tv1, v2.(int))
	case bool:
		return edc.compareBool(tv1, v2.(bool))
	}
}

// Compare two map[string]bool and return the keys that have different values
func (edc *EndpointDataComparator) compareMapStringBool(m1, m2 map[string]bool) ([]string, error) {
	nm1 := make(map[string]interface{})
	for k, v := range m1 {
		nm1[k] = v
	}

	nm2 := make(map[string]interface{})
	for k, v := range m2 {
		nm2[k] = v
	}

	return edc.compareMapString(nm1, nm2)
}

// Compare two map[string]int and return the keys that have different values
func (edc *EndpointDataComparator) compareMapStringString(m1, m2 map[string]string) ([]string, error) {
	nm1 := make(map[string]interface{})
	for k, v := range m1 {
		nm1[k] = v
	}

	nm2 := make(map[string]interface{})
	for k, v := range m2 {
		nm2[k] = v
	}

	return edc.compareMapString(nm1, nm2)
}

// Compare two map[string]int and return the keys that have different values
func (edc *EndpointDataComparator) compareMapStringInt(m1, m2 map[string]int) ([]string, error) {
	nm1 := make(map[string]interface{})
	for k, v := range m1 {
		nm1[k] = v
	}

	nm2 := make(map[string]interface{})
	for k, v := range m2 {
		nm2[k] = v
	}

	return edc.compareMapString(nm1, nm2)
}

// Compare two map[string]interface{} and return the keys that have different values
func (edc *EndpointDataComparator) compareMapString(m1, m2 map[string]interface{}) ([]string, error) {
	keys := make(map[string]bool)
	for k, _ := range m1 {
		keys[k] = true
	}
	for k, _ := range m2 {
		keys[k] = true
	}

	differentKeys := make(map[string]bool)
	for k, _ := range keys {
		v1 := m1[k]
		v2 := m2[k]
		same, err := edc.compareValues(v1, v2)

		if err != nil {
			return []string{}, err
		}

		if !same {
			differentKeys[k] = true
		}
	}

	differentKeysSlice := make([]string, 0, len(differentKeys))
	for k, _ := range differentKeys {
		differentKeysSlice = append(differentKeysSlice, k)
	}
	return differentKeysSlice, nil
}

// Compare two map[int]int and return the keys that have different values
func (edc *EndpointDataComparator) compareMapIntTime(m1, m2 map[int]TrackableTime) ([]int, error) {
	nm1 := make(map[int]interface{})
	for k, v := range m1 {
		nm1[k] = v
	}

	nm2 := make(map[int]interface{})
	for k, v := range m2 {
		nm2[k] = v
	}

	return edc.compareMapInt(nm1, nm2)
}

// Compare two map[int]int and return the keys that have different values
func (edc *EndpointDataComparator) compareMapIntInt(m1, m2 map[int]int) ([]int, error) {
	nm1 := make(map[int]interface{})
	for k, v := range m1 {
		nm1[k] = v
	}

	nm2 := make(map[int]interface{})
	for k, v := range m2 {
		nm2[k] = v
	}

	return edc.compareMapInt(nm1, nm2)
}

// Compare two map[int]interface{} and return the keys that have different values
func (edc *EndpointDataComparator) compareMapInt(m1, m2 map[int]interface{}) ([]int, error) {
	keys := make(map[int]bool)
	for k, _ := range m1 {
		keys[k] = true
	}
	for k, _ := range m2 {
		keys[k] = true
	}

	differentKeys := make(map[int]bool)
	for k, _ := range keys {
		v1 := m1[k]
		v2 := m2[k]
		same, err := edc.compareValues(v1, v2)

		if err != nil {
			return []int{}, err
		}

		if !same {
			differentKeys[k] = true
		}
	}

	differentKeysSlice := make([]int, 0, len(differentKeys))
	for k, _ := range differentKeys {
		differentKeysSlice = append(differentKeysSlice, k)
	}
	return differentKeysSlice, nil
}
