package fccommon

import (
	"bytes"
	"encoding/gob"
	"encoding/json"
	"fmt"
	"math/rand"
	"sync"
	"testing"
	"time"

	"github.com/davecgh/go-spew/spew"
	"github.com/inverse-inc/go-utils/sharedutils"
)

func TestEndpointDataUpdateFromJSON(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	ed.DhcpFingerprint = "192,168,5,3"
	ed.DhcpVendor = "yes hello"
	ed.AddUserAgent("Mozilla/5.0 testing")

	ed.UpdateFromJSON([]byte(`{"dhcp_vendor":"majew", "user_agents":{"Mozilla/5.0 ski":1}}`))

	if ed.DhcpFingerprint != "192,168,5,3" {
		t.Error("Update of the endpoint deleted the DHCP fingerprint although it wasn't part of the update")
	}

	if ed.UserAgents["Mozilla/5.0 testing"] != 1 {
		t.Error("Update of the endpoint changed a User-Agent although it wasn't part of the update")
	}

	if ed.DhcpVendor != "majew" {
		t.Error("Update of the endpoint didn't overwritte a field that was in the update request")
	}

	if ed.UserAgents["Mozilla/5.0 ski"] != 1 {
		t.Error("Update of the endpoint didn't add a User-Agent that was part of the update")
	}

}

func TestEndpointDataUpdateConcurrency(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed.lock.Timeout = 1 * time.Second
	ed.lock.RTimeout = 1 * time.Second
	ed.lock.PrintErrors = true

	c := make(chan int, 5)
	w := &sync.WaitGroup{}
	go func() {
		for {
			c <- 1
			go func() {
				defer func() {
					//if r := recover(); r != nil {
					//	fmt.Println(r)
					//}
					<-c
					w.Done()
				}()
				w.Add(1)
				s := fmt.Sprint(rand.Intn(100))
				ed.AddTCPSynSignature(s)
				idl := ed.lock.Lock()
				if idl != 0 {
					ed.lock.Unlock(idl)
				}

				id := ed.RLock()
				if id != 0 {
					spew.Sdump(ed)
					ed.RUnlock(id)
				}
			}()
		}
	}()

	// let this run for TestEndpointDataUpdateConcurrency_SECONDS or 5 seconds by default
	runFor := sharedutils.EnvOrDefaultInt("TestEndpointDataUpdateConcurrency_SECONDS", 5)
	fmt.Println("Will run for", runFor)
	time.Sleep(time.Duration(runFor) * time.Second)

	fmt.Println("Done")
}

func TestEndpointDataCopy(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	ed.DhcpFingerprint = "vidange"

	ed2 := ed.Copy()

	if &ed == &ed2 {
		t.Error("The endpoint data copy points to the same memory space as the source")
	}

	if ed.DhcpFingerprint != "vidange" {
		t.Error("The initial endpoint was altered during the copy")
	}

	if ed.DhcpFingerprint != ed2.DhcpFingerprint {
		t.Error("The endpoint data copy doesn't contain the same data as the source")
	}

	ed.DhcpFingerprint = "test"

	if ed2.DhcpFingerprint != "vidange" {
		t.Error("Altering the initial endpoint altered the copy")
	}
}

func TestAddLLDPCapability(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	ed.AddLLDPCapability("toto")
	updated1 := ed.LastUpdated

	if ed.LLDPCapabilities["toto"] != true {
		t.Error("Adding LLDP capability didn't add it")
	}

	// Add it another time and make sure last updated wasn't touched
	ed.AddLLDPCapability("toto")
	updated2 := ed.LastUpdated

	if updated1 != updated2 {
		t.Error("Endpoint LastUpdated was updated although nothing has changed")
	}
}

func TestAddCDPCapability(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	ed.AddCDPCapability("toto")
	updated1 := ed.LastUpdated

	if ed.CDPCapabilities["toto"] != true {
		t.Error("Adding CDP capability didn't add it")
	}

	// Add it another time and make sure last updated wasn't touched
	ed.AddCDPCapability("toto")
	updated2 := ed.LastUpdated

	if updated1 != updated2 {
		t.Error("Endpoint LastUpdated was updated although nothing has changed")
	}
}

func TestAddJA3SigCouple(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	ed.AddJA3SigCouple("acme.com", 443, "toto", "toto_srv")
	updated1 := ed.LastUpdated

	if ed.JA3Data["acme.com:443"]["toto;toto_srv"] != 1 {
		t.Error("Adding JA3 client signature capability didn't add it")
	}

	// Add it another time and make sure last updated wasn't touched
	ed.AddJA3SigCouple("acme.com", 443, "toto", "toto_srv")
	updated2 := ed.LastUpdated

	if ed.JA3Data["acme.com:443"]["toto;toto_srv"] != 2 {
		t.Error("Adding JA3 client signature capability didn't add it")
	}

	if updated1 != updated2 {
		t.Error("Endpoint LastUpdated was updated although nothing has changed")
	}

	ed.AddJA3SigCouple("acme.com", 443, "tata", "tata_srv")
	updated3 := ed.LastUpdated
	if ed.JA3Data["acme.com:443"]["tata;tata_srv"] != 1 {
		t.Error("Adding JA3 client signature capability didn't add it")
	}

	if updated1 == updated3 {
		t.Error("Endpoint LastUpdated wasn't updated although something has changed")
	}

	for i := 0; i < *maxTrackJA3Domains*2; i++ {
		ed.AddJA3SigCouple(fmt.Sprintf("%d.acme.com", i), 443, "tata", "tata_srv")
	}

	if len(ed.JA3Data) != *maxTrackJA3Domains {
		t.Error("Wrong amount of domains recorded in JA3 data")
	}

	if ed.JA3Data["acme.com:443"]["toto;toto_srv"] != 2 {
		t.Error("JA3 domain with the most data got purged")
	}
	if ed.JA3Data["acme.com:443"]["tata;tata_srv"] != 1 {
		t.Error("JA3 domain with the most data got purged")
	}

}

func TestEndpointDataToInterrogateQuery(t *testing.T) {
	fc := NewFingerbankCollector("")
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	for i := 0; i < *maxTrackJA3Domains; i++ {
		ed.AddJA3SigCouple(fmt.Sprintf("%d.acme.com", i), 443, "tata", "tata_srv")
	}

	data := ed.ToInterrogateQuery(fc)

	for i := 0; i < *maxTrackJA3Domains; i++ {
		domainPort := fmt.Sprintf("%d.acme.com:443", i)
		if data.JA3Data[domainPort][0] != "tata;tata_srv" {
			t.Error("Failed to add client sig to interrogate data")
		}
	}

	// Tata will be there 4 times, toto 3 times, titi 2 times, tutu 1 time
	ed.AddJA3SigCouple("0.acme.com", 443, "tata", "tata_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "tata", "tata_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "tata", "tata_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "toto", "tata_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "toto", "toto_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "toto", "toto_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "titi", "titi_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "titi", "titi_srv")
	ed.AddJA3SigCouple("0.acme.com", 443, "tutu", "tutu_srv")

	data = ed.ToInterrogateQuery(fc)
	if data.JA3Data["0.acme.com:443"][0] != "tata;tata_srv" {
		t.Error("Failed to add client sig to interrogate data")
	}
	if data.JA3Data["0.acme.com:443"][1] != "toto;toto_srv" {
		t.Error("Failed to add client sig to interrogate data")
	}
	if data.JA3Data["0.acme.com:443"][2] != "titi;titi_srv" {
		t.Error("Failed to add client sig to interrogate data")
	}
	if len(data.JA3Data["0.acme.com:443"]) != 3 {
		t.Error("Wrong amount of sigs added to the domain")
	}
}

func BenchmarkEndpointDataEncodingGob(b *testing.B) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed.DhcpFingerprint = "192,168,5,3"
	var buf bytes.Buffer
	var err error

	enc := gob.NewEncoder(&buf)
	dec := gob.NewDecoder(&buf)

	for i := 0; i < b.N; i++ {
		err = enc.Encode(ed)
		sharedutils.CheckError(err)
		ned := EndpointData{}
		err = dec.Decode(&ned)
		sharedutils.CheckError(err)
	}
}

func BenchmarkEndpointDataEncodingJSON(b *testing.B) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed.DhcpFingerprint = "192,168,5,3"
	var buf bytes.Buffer
	var err error

	enc := json.NewEncoder(&buf)
	dec := json.NewDecoder(&buf)

	for i := 0; i < b.N; i++ {
		err = enc.Encode(ed)
		sharedutils.CheckError(err)
		ned := EndpointData{}
		err = dec.Decode(&ned)
		sharedutils.CheckError(err)
	}
}

func BenchmarkEndpointDataCopy(b *testing.B) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed.DhcpFingerprint = "192,168,5,3"

	for i := 0; i < b.N; i++ {
		ed.Copy()
	}
}

//func TestEndpointDataBootSequence(t *testing.T) {
//	fc := NewFingerbankCollector("")
//	ed := fc.NewEndpointData()
//	ed.BootSequence.Start()
//	ed.BootSequence.Add("test")
//
//	if len(ed.BootSequence.Events) != 2 {
//		t.Error("Unexpected length for boot sequence events", len(ed.BootSequence.Events))
//	}
//
//	if ed.BootSequence.Events[1] != "test" {
//		t.Error("Unexpected value in boot sequence event", ed.BootSequence.Events[1])
//	}
//
//	ed.AddUserAgent("banana")
//	ed.AddUserAgent("banana")
//
//	if len(ed.BootSequence.Events) != 5 {
//		t.Error("Unexpected length for boot sequence events", len(ed.BootSequence.Events))
//	}
//
//	if !regexp.MustCompile(`.+:user_agents:banana`).MatchString(ed.BootSequence.Events[2]) {
//		t.Error("Unexpected value in boot sequence event", ed.BootSequence.Events[2])
//	}
//
//	if !regexp.MustCompile(`.+:most_accurate_user_agent:banana`).MatchString(ed.BootSequence.Events[3]) {
//		t.Error("Unexpected value in boot sequence event", ed.BootSequence.Events[3])
//	}
//
//	if !regexp.MustCompile(`.+:user_agents:banana`).MatchString(ed.BootSequence.Events[4]) {
//		t.Error("Unexpected value in boot sequence event", ed.BootSequence.Events[4])
//	}
//
//}
