package fccommon

import (
	"net"
	"testing"
	"time"

	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"
)

func TestNBAPolicyWatchedDeviceAttributes(t *testing.T) {
	p := &NBAPolicy{
		WatchedDeviceAttributes: map[string]bool{
			DHCP_FINGERPRINT:  true,
			HOSTNAME:          true,
			DESTINATION_HOSTS: true,
		},
	}

	if !p.WatchDhcpFingerprint() {
		t.Error("Invalid watch value")
	}
	if !p.WatchHostname() {
		t.Error("Invalid watch value")
	}
	if !p.WatchDestinationHosts() {
		t.Error("Invalid watch value")
	}
	if p.WatchTcpSynSignatures() {
		t.Error("Invalid watch value")
	}
}

func TestNetworkProfileEngineDiffWithEndpointData(t *testing.T) {
	collector := NewFingerbankCollector("")

	prof := fingerbankclient.NetworkProfile{
		DhcpFingerprints: []string{"1,2,3,4,5"},
		Hostnames:        []string{"zammits-pc"},
		UserAgents:       []string{`Mozilla/5.0 testing`},
	}
	pol := &NBAPolicy{
		WatchedDeviceAttributes: map[string]bool{
			DHCP_FINGERPRINT: true,
			HOSTNAME:         false,
			USER_AGENTS:      true,
		},
	}

	e := NewNetworkProfileEngine(prof, pol)

	// Empty endpoint data should yield no difference
	matched, unmatched := e.DiffWithEndpointData(collector, &EndpointData{})

	if len(matched) != 0 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 0 {
		t.Error("Length of result is not good")
	}

	// Test a non-array attribute that is the same
	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{DhcpFingerprint: "1,2,3,4,5"})

	if len(unmatched) != 0 {
		t.Error("Length of result is not good")
	}

	// Test a non-array attribute that is the same and array attribute that is the same
	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{DhcpFingerprint: "1,2,3,4,5", UserAgents: map[string]int{"Mozilla/5.0 testing": 1}})

	if len(matched) != 2 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 0 {
		t.Error("Length of result is not good")
	}

	// Test a valid and invalid attribute
	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{DhcpFingerprint: "1,2,3,4,5", UserAgents: map[string]int{"Mozilla/5.0 testinge": 1}})

	if len(matched) != 1 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 1 {
		t.Error("Length of result is not good")
	}

	if unmatched[0] != USER_AGENTS {
		t.Error("Content of array isn't good")
	}

	// Test an invalid attribute but that isn't wanted by the profile
	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{Hostname: "not-zammits-pc"})

	if len(matched) != 0 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 0 {
		t.Error("Length of result is not good")
	}

	prof = fingerbankclient.NetworkProfile{
		DestinationHosts:    []string{"1"},
		DhcpFingerprints:    []string{"1"},
		DhcpVendors:         []string{"1"},
		Hostnames:           []string{"1"},
		MdnsServices:        []string{"1"},
		OUIs:                []string{"1"},
		TcpSynSignatures:    []string{"1"},
		TcpSynAckSignatures: []string{"1"},
		UpnpServerStrings:   []string{"1"},
		UpnpUserAgents:      []string{"1"},
		UserAgents:          []string{"1"},
	}
	pol = &NBAPolicy{
		WatchedDeviceAttributes: map[string]bool{
			DHCP_FINGERPRINT:       true,
			DHCP_VENDOR:            true,
			HOSTNAME:               true,
			OUI:                    true,
			DESTINATION_HOSTS:      true,
			MDNS_SERVICES:          true,
			TCP_SYN_SIGNATURES:     true,
			TCP_SYN_ACK_SIGNATURES: true,
			UPNP_SERVER_STRINGS:    true,
			UPNP_USER_AGENTS:       true,
			USER_AGENTS:            true,
		},
	}

	e = NewNetworkProfileEngine(prof, pol)

	// Test all attributes are correct
	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{
		DhcpFingerprint:     "1",
		DhcpVendor:          "1",
		Hostname:            "1",
		OUI:                 "1",
		DestinationHosts:    map[string]map[string]int{"1": map[string]int{"1": 1}},
		MDNSServices:        map[string]int{"1": 1},
		TCPSynSignatures:    map[string]int{"1": 1},
		TCPSynAckSignatures: map[string]int{"1": 1},
		UPnPServerStrings:   map[string]int{"1": 1},
		UPnPUserAgents:      map[string]int{"1": 1},
		UserAgents:          map[string]int{"1": 1},
	})

	if len(matched) != 11 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 0 {
		t.Error("Length of result is not good")
	}

	// Test all attributes are correct with the array having an incorrect attribute
	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{
		DhcpFingerprint:     "1",
		DhcpVendor:          "1",
		Hostname:            "1",
		OUI:                 "1",
		DestinationHosts:    map[string]map[string]int{"1": map[string]int{"1": 1}, "2": map[string]int{"1": 1}},
		MDNSServices:        map[string]int{"1": 1},
		TCPSynSignatures:    map[string]int{"1": 1},
		TCPSynAckSignatures: map[string]int{"1": 1},
		UPnPServerStrings:   map[string]int{"1": 1},
		UPnPUserAgents:      map[string]int{"1": 1},
		UserAgents:          map[string]int{"1": 1},
	})

	if len(matched) != 11 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 1 {
		t.Error("Length of result is not good")
	}

	// Test all attributes are incorrect
	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{
		DhcpFingerprint:     "2",
		DhcpVendor:          "2",
		Hostname:            "2",
		OUI:                 "2",
		DestinationHosts:    map[string]map[string]int{"2": map[string]int{"1": 1}},
		MDNSServices:        map[string]int{"2": 1},
		TCPSynSignatures:    map[string]int{"2": 1},
		TCPSynAckSignatures: map[string]int{"2": 1},
		UPnPServerStrings:   map[string]int{"2": 1},
		UPnPUserAgents:      map[string]int{"2": 1},
		UserAgents:          map[string]int{"2": 1},
	})

	if len(matched) != 0 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 11 {
		t.Error("Length of result is not good")
	}

	// Test respect of watch flag
	pol = &NBAPolicy{
		WatchedDeviceAttributes: map[string]bool{
			DHCP_FINGERPRINT:       false,
			DHCP_VENDOR:            false,
			HOSTNAME:               false,
			OUI:                    false,
			DESTINATION_HOSTS:      false,
			MDNS_SERVICES:          false,
			TCP_SYN_SIGNATURES:     false,
			TCP_SYN_ACK_SIGNATURES: false,
			UPNP_SERVER_STRINGS:    false,
			UPNP_USER_AGENTS:       false,
			USER_AGENTS:            false,
		},
	}

	e = NewNetworkProfileEngine(prof, pol)

	matched, unmatched = e.DiffWithEndpointData(collector, &EndpointData{
		DhcpFingerprint:     "2",
		DhcpVendor:          "2",
		Hostname:            "2",
		OUI:                 "2",
		DestinationHosts:    map[string]map[string]int{"2": map[string]int{"1": 1}},
		MDNSServices:        map[string]int{"2": 1},
		TCPSynSignatures:    map[string]int{"2": 1},
		TCPSynAckSignatures: map[string]int{"2": 1},
		UPnPServerStrings:   map[string]int{"2": 1},
		UPnPUserAgents:      map[string]int{"2": 1},
		UserAgents:          map[string]int{"2": 1},
	})

	if len(matched) != 0 {
		t.Error("Length of result is not good")
	}

	if len(unmatched) != 0 {
		t.Error("Length of result is not good")
	}

}

func TestNetworkProfileEngineDiffThreshold(t *testing.T) {
	collector := NewFingerbankCollector("")

	prof := fingerbankclient.NetworkProfile{
		DhcpFingerprints: []string{"1,2,3,4,5"},
		UserAgents:       []string{`Mozilla/5.0 testing`},
	}
	pol := &NBAPolicy{
		WatchedDeviceAttributes: map[string]bool{
			DHCP_FINGERPRINT: true,
			USER_AGENTS:      true,
		},
	}

	e := NewNetworkProfileEngine(prof, pol)

	// Matches one
	diffScore := e.DiffScore(collector, &EndpointData{DhcpFingerprint: "1,2,3,4,5"})

	if diffScore != 10 {
		t.Error("Unexpected diffScore")
	}

	// Matches one and not another
	diffScore = e.DiffScore(collector, &EndpointData{DhcpFingerprint: "1,2,3,4,5", UserAgents: map[string]int{"test/test": 1}})

	if diffScore != 5 {
		t.Error("Unexpected diffScore")
	}

	// Matches none
	diffScore = e.DiffScore(collector, &EndpointData{DhcpFingerprint: "2,3,4,5", UserAgents: map[string]int{"test/test": 1}})

	if diffScore != -15 {
		t.Error("Unexpected diffScore")
	}

	// Put an override on the DHCP fingerprint and ensure its honored
	e.Policy.DeviceAttributesDiffThresholdOverrides = map[string]int{
		DHCP_FINGERPRINT: 15,
	}

	// Ensure its honored for negative points
	diffScore = e.DiffScore(collector, &EndpointData{DhcpFingerprint: "2,3,4,5", UserAgents: map[string]int{"test/test": 1}})

	if diffScore != -20 {
		t.Error("Unexpected diffScore")
	}

	// Ensure its honored for positive points
	diffScore = e.DiffScore(collector, &EndpointData{DhcpFingerprint: "1,2,3,4,5", UserAgents: map[string]int{"test/test": 1}})

	if diffScore != 10 {
		t.Error("Unexpected diffScore")
	}

}

func TestNetworkProfileEngineAnalyseEndpoint(t *testing.T) {
	collector := NewFingerbankCollector("")

	prof := fingerbankclient.NetworkProfile{
		DhcpFingerprints: []string{"1,2,3,4,5"},
		UserAgents:       []string{`Mozilla/5.0 testing`},
	}
	pol := &NBAPolicy{
		DevicesIncluded:             []int{1},
		DevicesExcluded:             []int{-1},
		WatchBlacklistedIPs:         "enabled",
		BlacklistedIpHostsThreshold: 1,
		BlacklistedIpHostsWindow:    1 * time.Hour,
		WatchedDeviceAttributes: map[string]bool{
			DHCP_FINGERPRINT: true,
			USER_AGENTS:      true,
		},
		DeviceAttributesDiffScore: 4,
	}

	e := NewNetworkProfileEngine(prof, pol)

	// Test perfect profile
	events := e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 0 {
		t.Error("Unexpected events length")
	}

	// Test diff score lower than acceptable
	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5,6",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 1 {
		t.Error("Unexpected events length")
	}

	if events[0] != DIFF_SCORE_TOO_LOW_EVENT {
		t.Error("Unexpected event value")
	}

	// Test diff score just a bit above acceptable
	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing dinde dinde": 1},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 0 {
		t.Error("Unexpected events length")
	}

	// Test perfect profile but with recent blacklisted IP
	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
				BlacklistedCommunications: map[string]map[int]TrackableTime{
					"1.2.3.4": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now()},
					},
				},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 1 {
		t.Error("Unexpected events length")
	}

	if events[0] != BLACKLISTED_IPS_THRESHOLD_TOO_HIGH {
		t.Error("Unexpected event value")
	}

	// Test perfect profile but with old blacklisted IP
	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
				BlacklistedCommunications: map[string]map[int]TrackableTime{
					"1.2.3.4": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now().Add(-2 * time.Hour)},
					},
				},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 0 {
		t.Error("Unexpected events length")
	}

	// Test perfect profile but with mix of old and new blacklisted IPs
	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
				BlacklistedCommunications: map[string]map[int]TrackableTime{
					"1.2.3.4": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now().Add(-2 * time.Hour)},
					},
					"1.2.3.5": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now()},
					},
				},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 1 {
		t.Error("Unexpected events length")
	}

	if events[0] != BLACKLISTED_IPS_THRESHOLD_TOO_HIGH {
		t.Error("Unexpected event value")
	}

	// Test needing more than 1 blacklisted IP
	e.Policy.BlacklistedIpHostsThreshold = 2

	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
				BlacklistedCommunications: map[string]map[int]TrackableTime{
					"1.2.3.4": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now().Add(-2 * time.Hour)},
					},
					"1.2.3.5": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now()},
					},
				},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 0 {
		t.Error("Unexpected events length")
	}

	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
				BlacklistedCommunications: map[string]map[int]TrackableTime{
					"1.2.3.4": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now()},
					},
					"1.2.3.5": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now()},
					},
				},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 1 {
		t.Error("Unexpected events length")
	}

	if events[0] != BLACKLISTED_IPS_THRESHOLD_TOO_HIGH {
		t.Error("Unexpected event value")
	}

	// Test whitelisting of blacklisted IPs
	e.Policy.WhitelistedIPs = []*net.IPNet{
		&net.IPNet{
			IP:   net.ParseIP("1.2.3.4"),
			Mask: net.CIDRMask(32, 32),
		},
	}

	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5",
				UserAgents:      map[string]int{"Mozilla/5.0 testing": 1},
				BlacklistedCommunications: map[string]map[int]TrackableTime{
					"1.2.3.4": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now()},
					},
					"1.2.3.5": map[int]TrackableTime{
						0: TrackableTime{Time: time.Now()},
					},
				},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 0 {
		t.Error("Unexpected events length")
	}

	// Test ignoring the invalid profile
	e.Policy.WatchedDeviceAttributes = map[string]bool{}

	events = e.AnalyseEndpoint(collector, &QueryResult{
		EndpointData: &ExpandedEndpointData{
			EndpointData: EndpointData{
				DhcpFingerprint: "1,2,3,4,5,6",
				UserAgents:      map[string]int{"Mozilla/5.0 testing invalid": 1},
			},
		},
		CloudAPIResult: &fingerbankclient.InterrogateResponse{
			Device: fingerbankclient.Device{Id: 1},
		},
	})

	if len(events) != 0 {
		t.Error("Unexpected events length")
	}

}
