package fcreplication

import (
	"bytes"
	"compress/gzip"
	"encoding/json"
	"errors"
	"io"
	"sync"
	"time"

	"github.com/inconshreveable/log15"
)

var ErrorBatcherMessageTooLarge = errors.New("The message is too large for the batcher")

const FLUSH_AT_LEAST = 1 * time.Second

const MAX_MESSAGE_SIZE = DEFAULT_MAX_BATCHER_BYTE_SIZE

const DEFAULT_MAX_BATCHER_BYTE_SIZE = 825

const BATCHER_MTU = 1000

const GZIP_OVERHEAD_BYTES = 50

const GZIP_HEADER_SIZE = 15

type BatcherEntry struct {
	empty       bool
	storage     []byte
	buffer      *bytes.Buffer
	gzipw       *gzip.Writer
	lastFlushed time.Time
}

type Batcher struct {
	sync.Mutex

	// The packets that are being batched for a peer
	toBatch map[string]*BatcherEntry

	MaxByteSize int

	// The function to call to send the message
	f func(id string, packet []byte)

	logger log15.Logger

	bufferPool sync.Pool
}

func NewBatcher(logger log15.Logger) *Batcher {
	b := &Batcher{
		toBatch:     map[string]*BatcherEntry{},
		MaxByteSize: DEFAULT_MAX_BATCHER_BYTE_SIZE,
		logger:      logger,
	}
	b.bufferPool = sync.Pool{
		New: func() interface{} {
			return make([]byte, MAX_MESSAGE_SIZE+GZIP_OVERHEAD_BYTES)
		},
	}
	return b
}

func (b *Batcher) Start() {
	go func() {
		for {
			time.Sleep(1 * time.Second)
			b.flushAll()
		}
	}()
}

func (b *Batcher) Add(id string, payload []byte) error {
	b.Lock()
	defer b.Unlock()

	if len(payload) > MAX_MESSAGE_SIZE {
		b.logger.Error("Received a payload that is larger than the MAX_MESSAGE_SIZE", len(payload))
		return ErrorBatcherMessageTooLarge
	}

	if b.toBatch[id] == nil {
		storage := b.bufferPool.Get().([]byte)[:]
		buffer := bytes.NewBuffer(storage)
		buffer.Reset()
		w := gzip.NewWriter(buffer)
		b.toBatch[id] = &BatcherEntry{storage: storage, buffer: buffer, gzipw: w, empty: true}
		w.Flush()
	}
	e := b.toBatch[id]

	// If we have a chance or exceeding the max size with this payload (assuming 0 compression is made), then we flush now
	if e.buffer.Len()+len(payload) > b.MaxByteSize {
		b.flush(id, e)
	}

	if e.empty {
		e.empty = false
		e.gzipw.Write([]byte(`[`))
	} else {
		e.gzipw.Write([]byte(`,`))
	}
	e.gzipw.Write(payload)

	if e.buffer.Len() >= b.MaxByteSize {
		b.flush(id, e)
	} else {
		e.gzipw.Flush()
	}
	return nil
}

func (b *Batcher) flushAll() {
	b.Lock()
	defer b.Unlock()
	// Ensure all batched payloads are flushed at least every second
	for id, e := range b.toBatch {
		if time.Since(e.lastFlushed) > FLUSH_AT_LEAST {
			if e.buffer.Len() > 0 {
				b.flush(id, e)
			}
		}
	}
}

func (b *Batcher) flushAllNow() {
	b.Lock()
	defer b.Unlock()
	// Ensure all batched payloads are flushed at least every second
	for id, e := range b.toBatch {
		if e.buffer.Len() > 0 {
			b.flush(id, e)
		}
	}
}

func (b *Batcher) safeFlush(id string, e *BatcherEntry) {
	b.Lock()
	defer b.Unlock()
	b.flush(id, e)
}

func (b *Batcher) flush(id string, e *BatcherEntry) {
	e.gzipw.Write([]byte(`]`))
	//ensure everything is flushed
	e.gzipw.Close()

	if e.buffer.Len() > BATCHER_MTU {
		b.logger.Error("The batched packet to send has exceeded the MTU")
	} else {
		b.f(id, e.buffer.Bytes())
	}

	e.storage = e.storage[:]
	e.buffer.Reset()
	e.gzipw.Reset(e.buffer)
	e.lastFlushed = time.Now()
	e.empty = true
}

func (b *Batcher) DecodeMessage(id string, data []byte) ([]UnencryptedReplicationMessage, error) {
	buf := bytes.NewBuffer(data)
	gzipr, err := gzip.NewReader(buf)
	if err != nil {
		return []UnencryptedReplicationMessage{}, err
	}

	err = nil
	jsonMsgs := b.bufferPool.Get().([]byte)[0:0]
	defer b.bufferPool.Put(jsonMsgs)
	for !errors.Is(err, io.EOF) {
		buf := b.bufferPool.Get().([]byte)
		defer b.bufferPool.Put(buf)
		var n int
		n, err = gzipr.Read(buf)
		if err != nil && !errors.Is(err, io.EOF) {
			return []UnencryptedReplicationMessage{}, err
		}
		if n != 0 {
			jsonMsgs = append(jsonMsgs, buf[0:n]...)
		}
	}

	msgs := []UnencryptedReplicationMessage{}
	err = json.Unmarshal(jsonMsgs, &msgs)
	if err != nil {
		return msgs, err
	}

	return msgs, nil
}
