package fcreplication

import (
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"net"
	"sync"
	"time"

	fclog "git.inverse.ca/inverse/fingerbank-collector/log"

	"github.com/google/uuid"
	"github.com/inconshreveable/log15"
	"github.com/inverse-inc/go-utils/sharedutils"
	cache "github.com/patrickmn/go-cache"
)

var batched = flag.Bool(
	"batched-replication",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_BATCHED_REPLICATION", "true")),
	"Whether or not to use batched replication when available (COLLECTOR_BATCHED_REPLICATION environment variable)",
)

var odaReplication = flag.Bool(
	"oda-replication",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_ODA_REPLICATION", "true")),
	"This flag enables on-demand-aggregation of the data which is lighter and scales better but causes data loss when a collector peers is unavailable. (COLLECTOR_ODA_REPLICATION environment variable)",
)

const (
	REPLICATION_PORT       = 1192
	CONNECTION_SECRET_SIZE = 32

	// Interval at which the HELLO messages are sent
	DEFAULT_HELLO_INTERVAL = 5
	// Interval in seconds at which the peers maintenance is done
	PEERS_MAINTENANCE_INTERVAL = 5
	// Timeout in seconds after which a peer is forgotten
	PEER_TIMEOUT = 30
)

var REPLICATION_PORT_STR = fmt.Sprintf("%d", REPLICATION_PORT)

const (
	// Types 30 and below will not require any ACK to be sent
	// Types 10 and below are reserved for the handshake
	REP_HANDSHAKE_PKT   = 0
	REP_HELLO           = REP_HANDSHAKE_PKT + 1
	REP_HELLO_REPLY     = REP_HANDSHAKE_PKT + 2
	REP_CHALLENGE       = REP_HANDSHAKE_PKT + 3
	REP_CHALLENGE_REPLY = REP_HANDSHAKE_PKT + 4
	REP_CHALLENGE_ACK   = REP_HANDSHAKE_PKT + 5

	REP_INTERNAL_PKT = 10
	REP_ACK          = REP_INTERNAL_PKT + 1

	REP_PKT_REQUIRES_ACK      = 30
	REP_PING                  = REP_PKT_REQUIRES_ACK + 1
	REP_ENDPOINT_DATA_UPDATE  = REP_PKT_REQUIRES_ACK + 2
	REP_IP_MAC_UPDATE         = REP_PKT_REQUIRES_ACK + 3
	REP_ENDPOINT_DATA_REPLACE = REP_PKT_REQUIRES_ACK + 4
	REP_ENDPOINT_DATA_DELETE  = REP_PKT_REQUIRES_ACK + 5
	REP_ADD_IP_HOST_MAPPING   = REP_PKT_REQUIRES_ACK + 6
	REP_ADD_DESTINATION_HOST  = REP_PKT_REQUIRES_ACK + 7
	REP_RESET_ENDPOINT_CACHE  = REP_PKT_REQUIRES_ACK + 8
	REP_BATCHED_MESSAGES      = REP_PKT_REQUIRES_ACK + 9
)

// This map tracks the messages that should be sent when the collector is running in On-Demand-Aggregation mode
// This mode aims to reduce the amount of replication messages in large scale environment where the collectors will share their data on demand
// More details here: https://wiki.inverse.ca/focus/fingerbank/daemonforembedded/DataReplicationRework#Option_1:_Correlation_of_data_in_peers
var odaModeMessages = map[uint8]bool{}

func init() {
	//everything smaller than a ping must go in because it's part of the handshakes or the internal messages
	for i := uint8(0); i <= REP_PING; i++ {
		odaModeMessages[i] = true
	}

	// Add the other messages that need to be sent in ODA mode
	for _, i := range []uint8{
		REP_IP_MAC_UPDATE,
		REP_ENDPOINT_DATA_REPLACE,
		REP_ENDPOINT_DATA_DELETE,
		REP_ADD_IP_HOST_MAPPING,
		REP_RESET_ENDPOINT_CACHE,
		REP_BATCHED_MESSAGES,
	} {
		odaModeMessages[i] = true
	}
}

type REPACK struct {
	// The time at which the ACK is being sent
	Timestamp time.Time
	Version   uint8
}

func NewREPACK() REPACK {
	return REPACK{
		Timestamp: time.Now(),
		Version:   REPLICATION_VERSION,
	}
}

type ActionHandler func(request interface{}) ReplicationMessage

type Controller struct {
	// The master shared secret which is used in discovery probes
	masterSecret []byte

	// The list of known peers
	inboundPeers  map[string]*FingerbankCollectorPeer
	outboundPeers map[string]*FingerbankCollectorPeer

	// The list of peers used purely for ODA replication (no support for traditional/legacy replication)
	odaPeers map[string]*ODAPeer

	// The list of peers it should attempt to connect to explicitely (unicast packets)
	wantedPeers []net.IP

	// When the collector is booting, it needs to figure out who to grab a full sync from, this flag controls this behavior
	booting bool
	// This collector's UUID
	UUID uuid.UUID
	// Time at which the controller was created
	startedAt time.Time
	// Global lock for the controller, since it must only handle a single packet at a time
	lock *sync.Mutex
	// The action map, where the key is a message type, and the value, a func to handle the message. Handshake, internal and ping packets don't use this
	// The request will need to be type asserted
	ActionMap map[uint8]ActionHandler
	// The last time a peer was forgotten (so a peer has failed)
	lastPeerChange time.Time

	// The logger of the controller
	logger log15.Logger

	// The batching module
	batcher *Batcher

	// The interval at which the HELLO packets must be sent
	HelloInterval time.Duration
}

func NewController(uuid uuid.UUID, masterSecret []byte, wantedPeers []net.IP, helloBroadcast bool) *Controller {
	c := &Controller{
		UUID:          uuid,
		masterSecret:  masterSecret,
		inboundPeers:  make(map[string]*FingerbankCollectorPeer),
		outboundPeers: make(map[string]*FingerbankCollectorPeer),
		wantedPeers:   wantedPeers,
		booting:       true,
		startedAt:     time.Now(),
		lock:          &sync.Mutex{},
		ActionMap:     make(map[uint8]ActionHandler),
		logger:        fclog.Logger(""),
		HelloInterval: DEFAULT_HELLO_INTERVAL,
	}

	c.batcher = NewBatcher(c.logger)
	c.batcher.f = c.sendBatchedContentToPeer
	c.batcher.Start()

	if helloBroadcast {
		// Add L2 broadcast to the wanted peers to discover all the collectors on the same L2 network
		c.wantedPeers = append(c.wantedPeers, net.ParseIP("255.255.255.255"))
	}

	c.logger.Info(fmt.Sprintf("Creating replication controller with UUID %s", uuid.String()))
	return c
}

// Starts a few goroutines
// One that sends discovery probes in broadcast + multicast on regular intervals
// Another one that receives discovery probes from peers
// Another one that performs maintenance on the peers to delete inactive ones
func (ctrl *Controller) StartReplication() {
	go ctrl.peersMaintenance()
	go ctrl.discoveryIn()
	go ctrl.discoveryOut()
}

func (ctrl *Controller) GetLastPeerChange() time.Time {
	return ctrl.lastPeerChange
}

func (ctrl *Controller) Uptime() time.Duration {
	return time.Duration(time.Since(ctrl.startedAt).Seconds())
}

// Returns a copy of the current oda peers
func (ctrl *Controller) GetODAPeers() map[string]ODAPeer {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()
	m := make(map[string]ODAPeer)
	for id, peer := range ctrl.odaPeers {
		m[id] = *peer
	}
	return m
}

// Returns a copy of the current inbound peers
func (ctrl *Controller) GetInboundPeers() map[string]FingerbankCollectorPeer {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()
	m := make(map[string]FingerbankCollectorPeer)
	for id, peer := range ctrl.inboundPeers {
		m[id] = *peer
	}
	return m
}

// Returns a copy of the current outbound peers
func (ctrl *Controller) GetOutboundPeers() map[string]FingerbankCollectorPeer {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()
	m := make(map[string]FingerbankCollectorPeer)
	for id, peer := range ctrl.outboundPeers {
		m[id] = *peer
	}
	return m
}

func (ctrl *Controller) IsInODAMode() bool {
	return *odaReplication
}

func (ctrl *Controller) discoveryOut() {
	for {
		// send discovery probes, sleep
		uptime := uint64(ctrl.Uptime())
		rawMsg := NewREPHello(ctrl.UUID, uptime)
		content, err := ctrl.encryptMessage(ctrl.masterSecret, rawMsg)
		sharedutils.CheckError(err)

		msg := ctrl.NewReplicationMessage(REP_HELLO, content)
		for _, wanted := range ctrl.wantedPeers {
			ctrl.logger.Debug("Sending discovery probe to " + wanted.String())
			ctrl.sendReplicationMessage("", wanted, msg)
		}
		// Send probes to outbound peers to invite them as inbound peers in case the broadcast probe doesn't make it there
		for _, peer := range ctrl.outboundPeers {
			ctrl.logger.Debug("Sending discovery probe to " + peer.IP.String())
			ctrl.sendReplicationMessage("", peer.IP, msg)
		}

		time.Sleep(ctrl.HelloInterval)
	}
}

func (ctrl *Controller) discoveryIn() {
	udpAddr, err := net.ResolveUDPAddr("udp", ":"+REPLICATION_PORT_STR)
	if err != nil {
		panic("Cannot create UDP address " + err.Error())
	}

	udpConn, err := net.ListenUDP("udp", udpAddr)
	if err != nil {
		panic("Cannot bind to replication port " + err.Error())
	}

	for {
		// receive broadcast + multicast probes (probably need to bind to a multicast addr)
		// Send the packets to a handlePacketIn goroutine
		var buf [2048]byte
		n, clientAddr, err := udpConn.ReadFromUDP(buf[0:])
		if err != nil {
			ctrl.logger.Error(fmt.Sprintf("Error Reading from the replication port: %s", err))
		} else {
			buf2 := buf[0:n]
			go func() {
				ctrl.lock.Lock()
				defer ctrl.lock.Unlock()

				done := make(chan bool, 1)
				go func() {
					err := ctrl.handlePacketIn(clientAddr.IP, buf2)
					if err != nil {
						ctrl.logger.Error(fmt.Sprintf("Error while handling packet: %s", err))
					}
					done <- true
				}()

				select {
				case <-time.After(5 * time.Second):
					ctrl.logger.Error("Took too long to handle inbound replication packet")
				case <-done:
					ctrl.logger.Debug("Handled inbound replication packet")
				}
			}()
		}
	}
}

func (ctrl *Controller) handlePacketIn(peerIP net.IP, packet []byte) error {
	// figure out message type by parsing packet into a ReplicationMessage
	msg, err := ctrl.decodeReplicationMessage(packet)
	if err != nil {
		return err
	}

	// Associate the IP address of the peer in the internal DB
	if msg.CollectorUUID != "" {
		if peer, err := ctrl.FindOutboundPeer(msg.CollectorUUID); err == nil {
			peer.IP = peerIP
		}
		if peer, err := ctrl.FindInboundPeer(msg.CollectorUUID); err == nil {
			peer.IP = peerIP
		}
	}

	resp, err := ctrl.handlePacketInByType(msg, peerIP)
	if err != nil {
		return err
	}

	// We got a response to send back
	if resp.Type > 0 {
		err = ctrl.sendReplicationMessage(msg.CollectorUUID, peerIP, resp)
		if err != nil {
			return err
		}
	}

	// That request requires an ACK
	if msg.Type > REP_PKT_REQUIRES_ACK {
		ctrl.sendACK(msg)
	}

	return nil
}

func (ctrl *Controller) handlePacketInByType(msg ReplicationMessage, peerIP net.IP) (ReplicationMessage, error) {
	var resp ReplicationMessage
	var err error
	switch msg.Type {
	case REP_HELLO:
		var p REPHello
		err = ctrl.decodePacketContent(msg, &p)
		if err != nil {
			return resp, err
		}
		resp, err = ctrl.handlePacketHello(peerIP, msg, p)

	case REP_HELLO_REPLY:
		var p REPHelloReply
		err = ctrl.decodePacketContent(msg, &p)
		if err != nil {
			return resp, err
		}
		resp, err = ctrl.handlePacketHelloReply(peerIP, msg, p)

	case REP_CHALLENGE:
		var p REPChallenge
		err = ctrl.decodePacketContent(msg, &p)
		if err != nil {
			return resp, err
		}
		resp, err = ctrl.handlePacketChallenge(peerIP, msg, p)

	case REP_CHALLENGE_REPLY:
		var p REPChallenge
		err = ctrl.decodePacketContent(msg, &p)
		if err != nil {
			return resp, err
		}
		err = ctrl.handlePacketChallengeReply(peerIP, msg, p)

	case REP_ACK:
		var p REPACK
		err = ctrl.decodePacketContent(msg, &p)
		if err != nil {
			return resp, err
		}

		if peer, err := ctrl.FindOutboundPeer(msg.CollectorUUID); err != nil {
			ctrl.logger.Warn(fmt.Sprintf("Unable to find peer %s from a received ACK ID", msg.CollectorUUID))
		} else {
			ctrl.logger.Debug(fmt.Sprintf("Received ACK from peer %s for message ID", msg.CollectorUUID))
			peer.RecordSeen()
			peer.version = p.Version
		}

		if peer, err := ctrl.FindInboundPeer(msg.CollectorUUID); err == nil {
			peer.version = p.Version
		}
	case REP_BATCHED_MESSAGES:
		err = ctrl.handlePacketBatchedMessages(msg, peerIP)
		return resp, err

	case REP_PING:
		// Nothing special to do for this, just let the ACK be sent back
		return resp, nil

	default:
		var found bool
		found, resp, err = ctrl.customHandleEncryptedPacket(msg, peerIP)
		if found {
			return resp, err
		} else {
			msg := fmt.Sprintf("Cannot handle packet of type %d", msg.Type)
			ctrl.logger.Warn(msg)
			return resp, errors.New(msg)
		}
	}

	return resp, nil
}

func (ctrl *Controller) dispatchActionMap(msgType uint8, p interface{}) ReplicationMessage {
	if action := ctrl.ActionMap[msgType]; action != nil {
		return action(p)
	} else {
		ctrl.logger.Warn(fmt.Sprintf("No handler for action type %d", msgType))
		return ReplicationMessage{}
	}
}

func (ctrl *Controller) sendPacket(peerIP net.IP, payload []byte) error {
	conn, err := net.Dial("udp", peerIP.String()+":"+REPLICATION_PORT_STR)
	if err != nil {
		return err
	}
	defer conn.Close()
	conn.Write(payload)

	return nil
}

// Same as sendContentToPeer but acquires the lock
func (ctrl *Controller) SendContentToPeer(peerUUID string, peer *FingerbankCollectorPeer, msgType uint8, content interface{}) error {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()
	return ctrl.sendContentToPeer(peerUUID, peer, msgType, content)
}

func (ctrl *Controller) addBatchedContentToPeer(peerUUID string, peer *FingerbankCollectorPeer, msgType uint8, content interface{}) error {
	contentBytes, err := json.Marshal(content)

	jsonBytes, err := json.Marshal(UnencryptedReplicationMessage{Type: msgType, Content: contentBytes})
	if err != nil {
		return err
	}

	ctrl.batcher.Add(peerUUID, jsonBytes)

	return nil
}

func (ctrl *Controller) sendBatchedContentToPeer(peerUUID string, msg []byte) {
	peer, err := ctrl.FindOutboundPeer(peerUUID)
	if err != nil {
		ctrl.logger.Error(fmt.Sprintf("Error while sending batched content to outbound friend %s. Error is: %s", peerUUID, err))
		return
	}

	err = ctrl.sendRawContentToPeer(peerUUID, peer, REP_BATCHED_MESSAGES, msg)
	if err != nil {
		ctrl.logger.Error(fmt.Sprintf("Error while sending batched content to outbound friend %s. Error is: %s", peerUUID, err))
		return
	}
}

func (ctrl *Controller) sendRawContentToPeer(peerUUID string, peer *FingerbankCollectorPeer, msgType uint8, content []byte) error {
	encryptedBytes, err := ctrl.encryptMessage(peer.Secret, content)
	if err != nil {
		return err
	}

	return ctrl.sendReplicationMessage(
		peerUUID,
		peer.IP,
		ctrl.NewReplicationMessage(msgType, encryptedBytes),
	)

}

// Takes any struct, JSON encodes it. Then encrypts the content with the shared peer secret and sends it in a ReplicationMessage
func (ctrl *Controller) sendContentToPeer(peerUUID string, peer *FingerbankCollectorPeer, msgType uint8, content interface{}) error {
	jsonBytes, err := json.Marshal(content)
	if err != nil {
		return err
	}

	return ctrl.sendRawContentToPeer(peerUUID, peer, msgType, jsonBytes)
}

func (ctrl *Controller) SendContentToOutboundFriends(msgType uint8, content interface{}) error {
	if ctrl.IsInODAMode() && !odaModeMessages[msgType] {
		ctrl.logger.Debug(fmt.Sprintf("Message type %d will not be sent because the collector is running in ODA mode", msgType))
		return nil
	}

	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()

	for peerUUID, peer := range ctrl.outboundPeers {
		if peer.friend {
			var err error
			if *batched && peer.version >= 2 {
				err = ctrl.addBatchedContentToPeer(peerUUID, peer, msgType, content)
			} else {
				err = ctrl.sendContentToPeer(peerUUID, peer, msgType, content)
			}
			if err != nil {
				ctrl.logger.Error(fmt.Sprintf("Error while sending content to outbound friend %s. Error is: %s", peerUUID, err))
				return err
			}
		}
	}
	return nil
}

func (ctrl *Controller) generateConnectionSecret() []byte {
	return sharedutils.RandomBytes(CONNECTION_SECRET_SIZE)
}

var ackDebounceCache = cache.New(5*time.Second, 1*time.Minute)

func (ctrl *Controller) sendACK(msg ReplicationMessage) error {
	peerUUID := msg.CollectorUUID
	if _, found := ackDebounceCache.Get(peerUUID); found {
		ctrl.logger.Debug(fmt.Sprintf("Debouncing ACK for to peer %s", peerUUID))
		return nil
	} else {
		ackDebounceCache.Set(peerUUID, true, cache.DefaultExpiration)
		if peer, err := ctrl.FindInboundPeer(peerUUID); err == nil {
			ctrl.logger.Debug(fmt.Sprintf("Sending ACK to peer %s", peerUUID))
			return ctrl.sendContentToPeer(peerUUID, peer, REP_ACK, NewREPACK())
		} else {
			msg := fmt.Sprintf("Unable to find inbound peer %s to send ACK", peerUUID)
			ctrl.logger.Warn(msg)
			return errors.New(msg)
		}
	}
}

func (ctrl *Controller) handlePacketBatchedMessages(msg ReplicationMessage, peerIP net.IP) error {
	peer, err := ctrl.FindInboundPeer(msg.CollectorUUID)
	if err != nil {
		return err
	}

	decrypted, err := ctrl.decryptMessage(peer.Secret, msg.EncryptedContent)
	messages, err := ctrl.batcher.DecodeMessage("", decrypted)
	if err != nil {
		return errors.New(fmt.Sprintf("Error while decoding replication message from %s: %s", peerIP, err))
	}

	for _, message := range messages {
		_, _, err = ctrl.customHandleUnencryptedPacket(message, peerIP)
		if err != nil {
			msg := fmt.Sprintf("Failed to handle message of type %d: %s", message.Type, err)
			ctrl.logger.Error(msg)
			return errors.New(msg)
		}
	}
	return nil
}
