package fcreplication

import (
	"encoding/json"
	"fmt"
	"net"
	"testing"
	"time"

	"github.com/google/uuid"
	"github.com/inverse-inc/go-utils/sharedutils"
)

func TestControllerDecodeReplicationMessage(t *testing.T) {
	u := uuid.New()
	ctrl := NewController(u, []byte("1234567890123456"), []net.IP{}, true)
	msg := ctrl.NewReplicationMessage(REP_ACK, []byte("yes-hello"))

	js, err := json.Marshal(msg)
	sharedutils.CheckError(err)

	msgDecoded, err := ctrl.decodeReplicationMessage(js)
	sharedutils.CheckError(err)

	if string(msgDecoded.EncryptedContent) != "yes-hello" {
		t.Error("Content wasn't decoded properly from JSON encoded ReplicationMessage")
	}

	if string(msgDecoded.CollectorUUID) != u.String() {
		t.Error("UUID wasn't decoded properly from JSON encoded ReplicationMessage")
	}
}

func TestControllerDecodePacketContent(t *testing.T) {
	testUUIDObject := uuid.New()
	testUUID := testUUIDObject.String()

	fmt.Println("Using test UUID", testUUID)

	ctrlUUID := uuid.New()
	ctrl := NewController(ctrlUUID, []byte("1234567890123456"), []net.IP{}, true)

	// REP_HELLO
	{
		rawMsg := NewREPHello(testUUIDObject, 20)
		content, err := ctrl.encryptMessage(ctrl.masterSecret, rawMsg)
		sharedutils.CheckError(err)
		msg := ReplicationMessage{
			Type:             REP_HELLO,
			EncryptedContent: content,
		}

		var hello REPHello

		err = ctrl.decodePacketContent(msg, &hello)
		sharedutils.CheckError(err)

		if hello.UUID() != testUUID {
			t.Error("Error while decoding REPHello")
		}

		if hello.Uptime() != 20 {
			t.Error("Error while decoding REPHello")
		}
	}

	// REP_HELLO_REPLY
	{
		content, err := ctrl.encryptMessage(ctrl.masterSecret, []byte("0987654321123456"))
		sharedutils.CheckError(err)
		msg := ReplicationMessage{
			CollectorUUID:    testUUID,
			Type:             REP_HELLO_REPLY,
			EncryptedContent: content,
		}

		var helloReply REPHelloReply

		err = ctrl.decodePacketContent(msg, &helloReply)
		sharedutils.CheckError(err)

		if string(helloReply.ConnectionSecret()) != "0987654321123456" {
			t.Error("Error while decoding REPHelloReply")
		}
	}

	// REP_CHALLENGE
	{
		connSecret := sharedutils.RandomBytes(16)
		ctrl.outboundPeers[testUUID] = &FingerbankCollectorPeer{Secret: connSecret}
		challengeWord := "dinde"

		content, err := ctrl.encryptMessage(connSecret, []byte(challengeWord))
		sharedutils.CheckError(err)
		msg := ReplicationMessage{
			CollectorUUID:    testUUID,
			Type:             REP_CHALLENGE,
			EncryptedContent: content,
		}

		var challenge REPChallenge

		err = ctrl.decodePacketContent(msg, &challenge)
		sharedutils.CheckError(err)

		if string(challenge.ChallengeWord()) != challengeWord {
			t.Error("Error while decoding REPChallenge")
		}
	}

	// REP_CHALLENGE_REPLY
	{
		connSecret := sharedutils.RandomBytes(16)
		ctrl.inboundPeers[testUUID] = &FingerbankCollectorPeer{Secret: connSecret}
		challengeWord := "dinde"

		// First encrypt with the master secret, then the connection secret
		content, err := ctrl.encryptMessage(ctrl.masterSecret, []byte(challengeWord))
		sharedutils.CheckError(err)
		content, err = ctrl.encryptMessage(connSecret, content)
		sharedutils.CheckError(err)

		msg := ReplicationMessage{
			CollectorUUID:    testUUID,
			Type:             REP_CHALLENGE_REPLY,
			EncryptedContent: content,
		}

		var challengeReply REPChallenge

		err = ctrl.decodePacketContent(msg, &challengeReply)
		sharedutils.CheckError(err)

		if string(challengeReply.ChallengeWord()) != challengeWord {
			t.Error("Error while decoding REPChallenge")
		}
	}

	// REP_ENDPOINT_DATA_UPDATE
	{
		connSecret := sharedutils.RandomBytes(16)
		ctrl.inboundPeers[testUUID] = &FingerbankCollectorPeer{Secret: connSecret}

		deltaJSON := []byte(`{"dhcp_fingerprint":"192,168,5,3"}`)
		update, err := json.Marshal(REPEndpointDataUpdate{
			Timestamp:             time.Now(),
			DeltaEndpointDataJSON: deltaJSON,
		})
		sharedutils.CheckError(err)

		content, err := ctrl.encryptMessage(connSecret, update)
		sharedutils.CheckError(err)

		msg := ReplicationMessage{
			CollectorUUID:    testUUID,
			Type:             REP_ENDPOINT_DATA_UPDATE,
			EncryptedContent: content,
		}

		var endpointDataUpdate REPEndpointDataUpdate

		err = ctrl.decodePacketContent(msg, &endpointDataUpdate)
		sharedutils.CheckError(err)

		if string(endpointDataUpdate.DeltaEndpointDataJSON) != string(deltaJSON) {
			t.Error("Error while decoding REPEndpointDataUpdate")
		}
	}

	// REP_ACK
	{
		connSecret := sharedutils.RandomBytes(16)
		ctrl.outboundPeers[testUUID] = &FingerbankCollectorPeer{Secret: connSecret}

		ack, err := json.Marshal(REPACK{
			Timestamp: time.Now(),
		})
		sharedutils.CheckError(err)

		content, err := ctrl.encryptMessage(connSecret, ack)
		sharedutils.CheckError(err)

		msg := ReplicationMessage{
			CollectorUUID:    testUUID,
			Type:             REP_ACK,
			EncryptedContent: content,
		}

		var ack2 REPACK

		err = ctrl.decodePacketContent(msg, &ack2)
		sharedutils.CheckError(err)
	}

	// Invalid decryption
	{
		connSecret := sharedutils.RandomBytes(16)
		ctrl.outboundPeers[testUUID] = &FingerbankCollectorPeer{Secret: connSecret}

		ack, err := json.Marshal(REPACK{
			Timestamp: time.Now(),
		})
		sharedutils.CheckError(err)

		// Using a different secret than the one that is expected
		content, err := ctrl.encryptMessage(ctrl.masterSecret, ack)
		sharedutils.CheckError(err)

		msg := ReplicationMessage{
			CollectorUUID:    testUUID,
			Type:             REP_ACK,
			EncryptedContent: content,
		}

		var ack2 REPACK

		err = ctrl.decodePacketContent(msg, &ack2)

		if err == nil {
			t.Error("Invalid decoding + decryption didn't provide an error")
		}
	}

}
