package fcreplication

import (
	"encoding/binary"
	"errors"
	"fmt"
	mathRand "math/rand"
	"net"
	"time"

	"github.com/google/uuid"
	"github.com/inverse-inc/go-utils/sharedutils"
)

const (
	CHALLENGE_WORD_SIZE = 100

	REPLICATION_VERSION = uint8(2)
)

// A HELLO message only contains the other collector UUID
type REPHello []byte

// Build a new HELLO packet for a specific UUID and uptime
// Will put two series of random bytes (one at the beginning, one at the end) and put the length of the first random bytes in the first byte of the array. Refer to the internal wiki for details on the packet
func NewREPHello(uuid uuid.UUID, uptime uint64) REPHello {
	random1Length := uint64(mathRand.Intn(100) + 100)
	random1 := sharedutils.RandomBytes(random1Length)

	random2Length := uint64(mathRand.Intn(100) + 100)
	random2 := sharedutils.RandomBytes(random2Length)

	// We'll be putting 10 "random" bytes
	rawMsg := []byte{uint8(random1Length)}
	rawMsg = append(rawMsg, random1...)

	// Encoding the UUID
	uuidBytes, _ := uuid.MarshalBinary()
	rawMsg = append(rawMsg, uuidBytes...)

	// Encoding the uptime
	rawUptime := make([]byte, binary.MaxVarintLen64)
	binary.PutUvarint(rawUptime, uptime)
	rawMsg = append(rawMsg, rawUptime...)

	// Random bytes again
	rawMsg = append(rawMsg, random2...)

	return REPHello(rawMsg)
}

// Get the size of the first series of random bytes which is a uint8 at the very start of the bytes array
func (reph REPHello) firstRandomSize() uint8 {
	return uint8(reph[0])
}

// Get the starting index of the UUID in the bytes array
// Its located after the first series of random bytes
func (reph REPHello) uuidStartIndex() uint64 {
	// 1 for the random size, then the value of the random size
	return 1 + uint64(reph.firstRandomSize())
}

// Get the UUID of the peer in the HELLO
func (reph REPHello) UUID() string {
	u, err := uuid.FromBytes(reph[reph.uuidStartIndex() : reph.uuidStartIndex()+16])
	if err != nil {
		panic(err)
	} else {
		return u.String()
	}
}

// Get the index of where the uptime in minutes is located
// Its right after the UUID in the bytes array
func (reph REPHello) uptimeStartIndex() uint64 {
	return reph.uuidStartIndex() + 16
}

// Get the uptime in minutes of the peer from the HELLO
func (reph REPHello) Uptime() uint64 {
	uptimeBytes := reph[reph.uptimeStartIndex() : reph.uptimeStartIndex()+binary.MaxVarintLen64]
	uptime, _ := binary.Uvarint(uptimeBytes)

	return uptime
}

func (reph REPHello) IsValid() bool {
	// Those packets have to be at least 200 bytes since there is at least that in random
	if len(reph) < 200 {
		return false
	}

	// We need a size at least big like the first random, 16 bytes UUID, and a uint64
	// Should be bigger since there should be at least 100 random bytes after that sequence
	if len(reph) < int(reph.firstRandomSize())+16+binary.MaxVarintLen64 {
		return false
	}

	return true
}

// Handle a HELLO packet by recording the UUID of the peer from the HELLO, and its uptime
func (ctrl *Controller) handlePacketHello(peerIP net.IP, msg ReplicationMessage, hello REPHello) (ReplicationMessage, error) {

	if !hello.IsValid() {
		return ReplicationMessage{}, errors.New("HELLO packet is invalid")
	}

	// In HELLO packets the UUID isn't part of the outer headers
	peerUUID := hello.UUID()

	if peerUUID == ctrl.UUID.String() {
		ctrl.logger.Debug("Ignoring HELLO packet since it was sent by this collector")
		return ReplicationMessage{}, nil
	}

	if peer, err := ctrl.FindOutboundPeer(peerUUID); err != nil {
		ctrl.logger.Info(fmt.Sprintf("Discovered new peer %s. Initiating handshake.", peerUUID))

		connectionSecret := ctrl.generateConnectionSecret()

		peer = ctrl.AddOutboundPeer(peerUUID, peerIP, connectionSecret)

		if peer == nil {
			return ReplicationMessage{}, errors.New("Unable to add outbound peer in replication controller")
		}

		peer.upSince = time.Now().Add(-(time.Duration(hello.Uptime()) * time.Second))

		content, err := ctrl.encryptMessage(ctrl.masterSecret, connectionSecret)
		sharedutils.CheckError(err)

		peer.RecordSeen()

		return ctrl.NewReplicationMessage(REP_HELLO_REPLY, content), nil
	} else {
		// Peer is already known, no need to do anything with those packets
		ctrl.logger.Debug(fmt.Sprintf("Ignoring HELLO packet from %s, peer is already known \n", peerUUID))
		return ReplicationMessage{}, nil
	}
}

// Decode a HELLO which is a bunch of bytes with various infos encrypted using the master key
func (ctrl *Controller) decodePacketHello(msg ReplicationMessage, into interface{}) error {
	// For HELLO packets, the master secret is used
	content, err := ctrl.decryptMessage(ctrl.masterSecret, msg.EncryptedContent)
	if err != nil {
		// For HELLO packets, we won't return an error when failing to decrypt it in case other collectors with different keys are there on the network
		ctrl.logger.Debug(
			fmt.Sprintf(
				"Error while decoding HELLO packet (%s). This likely indicates other Fingerbank collectors running with a different key are running on this network or that your collectors aren't using the same key.",
				err,
			),
		)
		return nil
	}

	s := into.(*REPHello)
	*s = REPHello(content)

	if !(*s).IsValid() {
		return errors.New("HELLO packet is not valid")
	}

	return nil
}

type REPHelloReply []byte

func (rephr REPHelloReply) ConnectionSecret() []byte {
	return []byte(rephr)
}

// Handle an HELLO-REPLY by recording the peer as a new inbound peer, then sending him a random word though a CHALLENGE twhich is encrypted using the connection secret
func (ctrl *Controller) handlePacketHelloReply(peerIP net.IP, msg ReplicationMessage, helloReply REPHelloReply) (ReplicationMessage, error) {
	peerUUID := msg.CollectorUUID

	ctrl.logger.Info(fmt.Sprintf("Peer %s has replied to our HELLO probe. Initiating challenge \n", peerUUID))
	peer := ctrl.AddInboundPeer(peerUUID, peerIP, helloReply.ConnectionSecret())
	peer.challengeWord = sharedutils.RandomBytes(CHALLENGE_WORD_SIZE)
	content, err := ctrl.encryptMessage(peer.Secret, peer.challengeWord)
	sharedutils.CheckError(err)

	peer.RecordSeen()

	return ctrl.NewReplicationMessage(REP_CHALLENGE, content), nil

}

// Decode a HELLO-REPLY which is the connection secret encoded using the master secret
func (ctrl *Controller) decodePacketHelloReply(msg ReplicationMessage, into interface{}) error {
	// For HELLO-REPLY packets, the master secret is used

	content, err := ctrl.decryptMessage(ctrl.masterSecret, msg.EncryptedContent)
	if err != nil {
		return err
	}

	s := into.(*REPHelloReply)
	*s = REPHelloReply(content)
	return nil
}

type REPChallenge []byte

func (repc REPChallenge) ChallengeWord() []byte {
	return []byte(repc)
}

// Handle a CHALLENGE by encoding the word it contains using the master secret then the connection secret
// After this stage, the other collector will be a friend for the outbound traffic
func (ctrl *Controller) handlePacketChallenge(peerIP net.IP, msg ReplicationMessage, challenge REPChallenge) (ReplicationMessage, error) {
	peerUUID := msg.CollectorUUID

	if peer, err := ctrl.FindOutboundPeer(peerUUID); err == nil {
		ctrl.logger.Info(fmt.Sprintf("Peer %s has sent us a challenge", peerUUID))
		peer.challengeWord = challenge.ChallengeWord()

		// Peer is now a friend, since when it will accept the challenge, it will start accepting anything from us
		ctrl.logger.Info(fmt.Sprintf("Now sending outbound packets to %s", peerUUID))
		peer.friend = true
		peer.RecordSeen()

		// First encrypt with the master secret, then the connection secret
		content, err := ctrl.encryptMessage(ctrl.masterSecret, peer.challengeWord)
		sharedutils.CheckError(err)
		content, err = ctrl.encryptMessage(peer.Secret, content)
		sharedutils.CheckError(err)

		return ctrl.NewReplicationMessage(REP_CHALLENGE_REPLY, content), nil
	} else {
		msg := "Received CHALLENGE for unknown peer " + peerUUID
		ctrl.logger.Warn(msg)
		return ReplicationMessage{}, errors.New(msg)
	}

}

// Decode a CHALLENGE which is a word encrypted using the peer secret
func (ctrl *Controller) decodePacketChallenge(msg ReplicationMessage, into interface{}) error {

	peer, err := ctrl.FindOutboundPeer(msg.CollectorUUID)
	if err != nil {
		return err
	}

	content, err := ctrl.decryptMessage(peer.Secret, msg.EncryptedContent)
	if err != nil {
		return err
	}

	s := into.(*REPChallenge)
	*s = REPChallenge(content)
	return nil
}

// Handle the CHALLENGE-REPLY which validates the decoded word matches the one sent in the CHALLENGE
// After this stage, the other collector will be a friend for the inbound traffic
func (ctrl *Controller) handlePacketChallengeReply(peerIP net.IP, msg ReplicationMessage, challengeReply REPChallenge) error {
	peerUUID := msg.CollectorUUID

	if peer, err := ctrl.FindInboundPeer(peerUUID); err == nil {
		ctrl.logger.Info(fmt.Sprintf("Peer %s has sent us the reply to our challenge", peerUUID))
		if !ctrl.validateChallengeReply(peer, challengeReply) {
			msg := "Invalid challenge word received from peer " + peerUUID
			ctrl.logger.Warn(msg)
			return errors.New(msg)
		} else {
			ctrl.logger.Info(fmt.Sprintf("Now accepting inbound packets from %s \n", peerUUID))
			peer.friend = true
			peer.RecordSeen()
			return nil
		}
	} else {
		msg := "Received CHALLENGE-REPLY for unknown peer " + peerUUID
		ctrl.logger.Warn(msg)
		return errors.New(msg)
	}
}

// Validate a challenge reply word for a peer
func (ctrl *Controller) validateChallengeReply(peer *FingerbankCollectorPeer, challengeReply REPChallenge) bool {
	request := peer.challengeWord
	reply := challengeReply.ChallengeWord()
	if len(request) != len(reply) {
		ctrl.logger.Warn("Challenge request and reply lengths aren't the same")
		return false
	}

	for i := 0; i < len(request); i++ {
		if request[i] != reply[i] {
			ctrl.logger.Warn(fmt.Sprintf("Byte %d isn't the same on for both challenges", i))
			return false
		}
	}

	return true
}

func (ctrl *Controller) decodePacketChallengeReply(msg ReplicationMessage, into interface{}) error {
	// The CHALLENGE is encrypted using the peer secret and the master key
	// First we need to decode using the connection secret, then using the master key

	peer, err := ctrl.FindInboundPeer(msg.CollectorUUID)
	if err != nil {
		return err
	}

	content, err := ctrl.decryptMessage(peer.Secret, msg.EncryptedContent)
	if err != nil {
		return err
	}
	content, err = ctrl.decryptMessage(ctrl.masterSecret, content)
	if err != nil {
		return err
	}

	s := into.(*REPChallenge)
	*s = REPChallenge(content)
	return nil
}
