package fcreplication

import (
	"encoding/json"
	"errors"
	"fmt"
	"net"
	"net/http"
	"time"

	"github.com/inverse-inc/go-utils/sharedutils"
)

const (
	TRACKABLE_PEER_DEAD_ERROR_THRESHOLD = 3
	TRACKABLE_PEER_DEAD_DURATION        = 1 * time.Minute

	ODA_PEERS_DISCOVERY_INTERVAL = 15 * time.Second
)

type ApiSyncCapablePeer interface {
	GetURL(defaultProtocol string) string

	ShouldQuery() bool
	RecordError()
	RecordSuccess()
}

type ODAPeer struct {
	TrackablePeer

	peerURL string
}

func (p *ODAPeer) GetURL(string) string {
	return p.peerURL
}

func (p *ODAPeer) String() string {
	return fmt.Sprintf("ODAPeer %s", p.peerURL)
}

type FingerbankCollectorPeer struct {
	TrackablePeer

	// The IP address of the peer
	IP net.IP
	// The shared secret specific to this peer
	Secret []byte
	// Last time the collector was seen. Used for expiring
	lastSeen time.Time
	// The uptime of the peer, used for determining the best sync option
	upSince time.Time
	// The word used during the challenge
	challengeWord []byte
	// This flag is set to true when the challenge is completed
	friend bool
	// The connection to send replication packets to the peer (outbound peers only)
	replicationConn net.Conn
	// The replication version of the peer
	version uint8
}

func (fcp *FingerbankCollectorPeer) String() string {
	return fmt.Sprintf("FingerbankCollectorPeer %s", fcp.IP)
}

func (fcp *FingerbankCollectorPeer) GetURL(protocol string) string {
	port := sharedutils.EnvOrDefault("PORT", "8080")

	return fmt.Sprintf("%s://%s:%s", protocol, fcp.IP, port)
}

// Record that we've seen this collector peer
func (fcp *FingerbankCollectorPeer) RecordSeen() {
	fcp.lastSeen = time.Now()
}

func (fcp *FingerbankCollectorPeer) Uptime() time.Duration {
	return time.Duration(time.Since(fcp.upSince).Seconds())
}

func (fcp FingerbankCollectorPeer) MarshalJSON() ([]byte, error) {
	return json.Marshal(struct {
		IP       net.IP    `json:"ip"`
		LastSeen time.Time `json:"last_seen"`
		UpSince  time.Time `json:"up_since"`
		Friend   bool      `json:"friend"`
		Version  uint8     `json:"version"`
	}{
		IP:       fcp.IP,
		LastSeen: fcp.lastSeen,
		UpSince:  fcp.upSince,
		Friend:   fcp.friend,
		Version:  fcp.version,
	})
}

func (fcp *FingerbankCollectorPeer) SendOutboundMsg(payload []byte) error {
	_, err := fcp.replicationConn.Write(payload)
	return err
}

func (ctrl *Controller) peersMaintenance() {
	for {
		func() {
			ctrl.lock.Lock()
			defer ctrl.lock.Unlock()
			ctrl.expirePeers()
		}()

		time.Sleep(PEERS_MAINTENANCE_INTERVAL * time.Second)
	}
}

// Send a REP_PING to outbound peers to see if they are alive and at the same time expire the ones that haven't been seen for PEER_TIMEOUT
func (ctrl *Controller) expirePeers() {
	for peerUUID, peer := range ctrl.outboundPeers {
		if peer.friend {
			ctrl.logger.Debug(fmt.Sprintf("Sending REP_PING to %s", peerUUID))
			ctrl.sendContentToPeer(peerUUID, peer, REP_PING, nil)
		}

		if peer.lastSeen.Before(time.Now().Add(-PEER_TIMEOUT * time.Second)) {
			ctrl.logger.Warn(fmt.Sprintf("Forgetting outbound peer %s that hasn't been seen for %d seconds", peerUUID, int(time.Since(peer.lastSeen).Seconds())))
			delete(ctrl.outboundPeers, peerUUID)
			ctrl.lastPeerChange = time.Now()
		}
	}
}

// Adds an outbound peer to the controller, if the peer UUID already exists, it replaces it
func (ctrl *Controller) AddOutboundPeer(uuid string, peerIP net.IP, secret []byte) *FingerbankCollectorPeer {
	ctrl.RemovePeerByIP(ctrl.outboundPeers, peerIP)

	peerReplicationAddress := peerIP.String() + ":" + REPLICATION_PORT_STR
	conn, err := net.Dial("udp", peerReplicationAddress)
	if err != nil {
		ctrl.logger.Error("Unable to dial connection to " + peerReplicationAddress)
		return nil
	}

	ctrl.outboundPeers[uuid] = &FingerbankCollectorPeer{
		Secret:          secret,
		IP:              peerIP,
		replicationConn: conn,
	}
	ctrl.outboundPeers[uuid].InitTrackablePeer(TRACKABLE_PEER_DEAD_ERROR_THRESHOLD, TRACKABLE_PEER_DEAD_DURATION)
	ctrl.lastPeerChange = time.Now()
	return ctrl.outboundPeers[uuid]
}

// Find an outbound peer by its UUID
func (ctrl *Controller) FindOutboundPeer(uuid string) (*FingerbankCollectorPeer, error) {
	if peer, ok := ctrl.outboundPeers[uuid]; ok {
		return peer, nil
	} else {
		return nil, errors.New("Couldn't find outbound peer with UUID " + uuid)
	}
}

// Adds an inbound peer to the controller, if the peer UUID already exists, it replaces it
func (ctrl *Controller) AddInboundPeer(uuid string, peerIP net.IP, secret []byte) *FingerbankCollectorPeer {
	ctrl.RemovePeerByIP(ctrl.inboundPeers, peerIP)
	ctrl.inboundPeers[uuid] = &FingerbankCollectorPeer{
		Secret: secret,
		IP:     peerIP,
	}
	ctrl.lastPeerChange = time.Now()
	return ctrl.inboundPeers[uuid]
}

// Find an inbound peer by its UUID
func (ctrl *Controller) FindInboundPeer(uuid string) (*FingerbankCollectorPeer, error) {
	if peer, ok := ctrl.inboundPeers[uuid]; ok {
		return peer, nil
	} else {
		return nil, errors.New("Couldn't find peer with UUID " + uuid)
	}
}

// Removes a peer by its IP address
func (ctrl *Controller) RemovePeerByIP(peerMap map[string]*FingerbankCollectorPeer, peerIP net.IP) bool {
	for id, peer := range peerMap {
		if peer.IP.Equal(peerIP) {
			ctrl.lastPeerChange = time.Now()
			delete(peerMap, id)
			return true
		}
	}
	return false
}

func (ctrl *Controller) FindOldestPeerForInitialSync(initialWait time.Duration) (string, *FingerbankCollectorPeer) {
	time.Sleep(initialWait)

	return ctrl.FindOldestPeer()
}

func (ctrl *Controller) FindOldestPeer() (string, *FingerbankCollectorPeer) {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()

	ctrl.logger.Debug("Attempting to find the best peer to sync from")

	bestUpSincePeerUUID := ""
	var bestUpSincePeer *FingerbankCollectorPeer
	bestUpSince := ctrl.startedAt
	// Its the outbound peers that send us HELLO packets that contain the uptime
	for peerUUID, peer := range ctrl.outboundPeers {
		if peer.friend && peer.upSince.Before(bestUpSince) {
			bestUpSince = peer.upSince
			bestUpSincePeer = peer
			bestUpSincePeerUUID = peerUUID
		}
	}

	if bestUpSincePeerUUID == "" {
		ctrl.logger.Warn("Couldn't find a peer that has more uptime than this collector. Will not sync from anybody.")
		return "", nil
	} else {
		ctrl.logger.Info(fmt.Sprintf("Found peer %s that is up since %s. Will sync from him.", bestUpSincePeerUUID, bestUpSince))
		return bestUpSincePeerUUID, bestUpSincePeer
	}
}

func (ctrl *Controller) ListOutboundPeers() (p []string) {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()

	for peerUUID, _ := range ctrl.outboundPeers {
		p = append(p, peerUUID)
	}

	return p
}

func (ctrl *Controller) ListInboundPeers() (p []string) {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()

	for peerUUID, _ := range ctrl.inboundPeers {
		p = append(p, peerUUID)
	}

	return p
}

func (ctrl *Controller) ApiSyncCapablePeers() (peers []ApiSyncCapablePeer) {
	ctrl.lock.Lock()
	defer ctrl.lock.Unlock()

	for _, peer := range ctrl.outboundPeers {
		peers = append(peers, peer)
	}
	for _, peer := range ctrl.odaPeers {
		peers = append(peers, peer)
	}

	return peers
}

func (ctrl *Controller) DiscoverOdaPeers(u string) {
	for {
		ctrl.logger.Info(fmt.Sprintf("Refreshing list of ODA peer from %s", u))
		if resp, err := http.Get(u); err == nil {
			data := struct {
				Servers []string
			}{}
			if err := json.NewDecoder(resp.Body).Decode(&data); err == nil {
				func() {
					ctrl.lock.Lock()
					defer ctrl.lock.Unlock()
					ctrl.odaPeers = map[string]*ODAPeer{}
					for _, server := range data.Servers {
						ctrl.AddOdaPeer(server)
					}
				}()
			} else {
				ctrl.logger.Error(fmt.Sprintf("Unable to decode ODA peers from %s: %s", u, err))
			}
		} else {
			ctrl.logger.Error(fmt.Sprintf("Failed to discover ODA peers on %s: %s", u, err))
		}

		time.Sleep(ODA_PEERS_DISCOVERY_INTERVAL)
	}
}

func (ctrl *Controller) AddOdaPeer(peerURL string) *ODAPeer {
	uid := peerURL
	ctrl.odaPeers[uid] = &ODAPeer{peerURL: peerURL}
	ctrl.odaPeers[uid].InitTrackablePeer(TRACKABLE_PEER_DEAD_ERROR_THRESHOLD, TRACKABLE_PEER_DEAD_DURATION)
	ctrl.lastPeerChange = time.Now()
	return ctrl.odaPeers[uid]
}
