// +build dpi_godpi

package dpi

import (
	"errors"
	"strings"

	fclog "git.inverse.ca/inverse/fingerbank-collector/log"
	"github.com/google/gopacket"
	"github.com/inconshreveable/log15"
	godpi "github.com/mushorg/go-dpi"
	"github.com/mushorg/go-dpi/modules/wrappers"
)

const (
	LPIPrefix  = "lpi:"
	NDPIPrefix = "ndpi:"
)

type GodpiConfig struct {
	WithLPI  bool
	WithNDPI bool
	Logger   log15.Logger
}

type GodpiEngine struct {
	callback func(metadata Metadata, signatures []string)
	lpi      *wrappers.LPIWrapper
	ndpi     *wrappers.NDPIWrapper
	logger   log15.Logger
}

func CreateGodpiEngine(cfg GodpiConfig) (*GodpiEngine, error) {
	e := &GodpiEngine{}
	if cfg.WithLPI {
		e.lpi = wrappers.NewLPIWrapper()
	}
	if cfg.WithNDPI {
		e.ndpi = wrappers.NewNDPIWrapper()
	}
	if e.lpi == nil && e.ndpi == nil {
		return nil, errors.New("No DPI engine was enabled in the configuration")
	}
	godpi.Initialize()

	if cfg.Logger == nil {
		e.logger = fclog.Logger("")
	} else {
		e.logger = cfg.Logger
	}

	return e, nil
}

func (e *GodpiEngine) Destroy() {
	if e.lpi != nil {
		e.lpi.DestroyWrapper()
	}
	if e.ndpi != nil {
		e.ndpi.DestroyWrapper()
	}
}

func (e *GodpiEngine) SetCallback(callback func(metadata Metadata, signatures []string)) {
	e.callback = callback
}

func (e *GodpiEngine) ProcessPkt(metadata Metadata, gp gopacket.Packet, data []byte, info gopacket.CaptureInfo, linkType int) {
	flow, _ := godpi.GetPacketFlow(gp)

	sigs := []string{}

	if e.lpi != nil {
		proto, cat, err := e.lpi.ClassifyFlowFull(flow)
		if err != nil {
			// TODO: have a real logger
			e.logger.Error("ERROR in LPI: " + err.Error())
		} else {
			sigs = append(sigs, LPIPrefix+strings.ToLower(string(proto)), LPIPrefix+strings.ToLower(cat))
		}
	}

	if e.ndpi != nil {
		proto, cat, err := e.ndpi.ClassifyFlowFull(flow)
		if err != nil {
			// TODO: have a real logger
			e.logger.Error("ERROR in nDPI: " + err.Error())
		} else {
			sigs = append(sigs, NDPIPrefix+strings.ToLower(string(proto)), NDPIPrefix+strings.ToLower(cat))
		}
	}

	e.callback(metadata, sigs)
}
