package fingerbankclient

import (
	"bufio"
	"bytes"
	"crypto/sha256"
	"crypto/tls"
	"crypto/x509"
	"encoding/csv"
	"encoding/hex"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"regexp"
	"strconv"
	"strings"
	"time"

	"git.inverse.ca/inverse/fingerbank-api-golang/filter_engine"
	"git.inverse.ca/inverse/fingerbank-collector/host_analyser"
	fclog "git.inverse.ca/inverse/fingerbank-collector/log"
	"github.com/inverse-inc/go-utils/sharedutils"
)

var FingerbankKey string

var FingerbankProtocol = "https"
var FingerbankHost = "api-ss.fingerbank.org"
var FingerbankPort = 443
var FingerbankKeyFingerprint []byte

var httpclient *http.Client

func init() {
	var err error
	FingerbankKeyFingerprint, err = hex.DecodeString("c3f56255fb9f7583bc4633fdea21c36c340f9bb63fd85837e0582fc3c92043a1")
	if err != nil {
		panic(err)
	}

	tr := &http.Transport{
		MaxIdleConnsPerHost: 30,
		TLSHandshakeTimeout: 0 * time.Second,
	}
	tr.Proxy = http.ProxyFromEnvironment

	tr.TLSClientConfig = &tls.Config{
		InsecureSkipVerify:    true,
		VerifyPeerCertificate: fingerbankTLSValidator(),
	}

	httpclient = &http.Client{Transport: tr}
}

func fingerbankURI(path string) string {
	path = FingerbankProtocol + "://" + FingerbankHost + ":" + strconv.FormatInt(int64(FingerbankPort), 10) + path

	key := "key=" + FingerbankKey
	if strings.Contains(path, "?") {
		path += "&" + key
	} else {
		path += "?" + key
	}

	return path
}

func fingerbankTLSValidator() func(rawCerts [][]byte, verifiedChains [][]*x509.Certificate) error {
	return func(rawCerts [][]byte, verifiedChains [][]*x509.Certificate) error {
		keyPinValid := false
		for _, rawPeerCert := range rawCerts {
			crts, err := x509.ParseCertificates(rawPeerCert)
			sharedutils.CheckError(err)
			peercert := crts[0]

			der, err := x509.MarshalPKIXPublicKey(peercert.PublicKey)
			hash := sha256.Sum256(der)
			if err != nil {
				panic(err)
			}
			if bytes.Compare(hash[0:], FingerbankKeyFingerprint) == 0 {
				keyPinValid = true
			}
		}
		if keyPinValid == false {
			return errors.New("Invalid key fingerprint. Dropping connection to API")
		} else {
			return nil
		}
	}
}

func doQuery(req *http.Request, validStatus []int) (*http.Response, error) {
	req.Header.Set("User-Agent", "Fingerbank-Collector/1.0 Build/PacketFence")

	resp, err := httpclient.Do(req)
	if err != nil {
		return resp, err
	}

	valid := false
	for _, status := range validStatus {
		if resp.StatusCode == status {
			valid = true
		}
	}

	if !valid {
		defer resp.Body.Close()
		body, _ := ioutil.ReadAll(resp.Body)

		return resp, errors.New(fmt.Sprintf("Error while querying API. Got response: %s", body))
	}

	return resp, nil
}

func jsonQuery(method string, url string, validStatus []int, query interface{}, responseContainer interface{}) error {
	payload, err := json.Marshal(query)
	if err != nil {
		return err
	}

	buf := bytes.NewBuffer(payload)

	req, err := http.NewRequest(
		method,
		fingerbankURI(url),
		buf,
	)
	if err != nil {
		return errors.New(fmt.Sprintf("Error while creating request: %s", err))
	}

	req.Header["Content-Type"] = []string{"application/json"}

	resp, err := doQuery(req, validStatus)

	if err != nil {
		return err
	}

	defer resp.Body.Close()
	body, err := ioutil.ReadAll(resp.Body)

	if err != nil {
		return errors.New(fmt.Sprintf("Error while reading response: %s", err))
	}

	err = json.Unmarshal(body, responseContainer)

	if err != nil {
		return errors.New(fmt.Sprintf("Error while decoding JSON response: %s", err))
	}

	return nil
}

func Interrogate(query InterrogateQuery) (*InterrogateResponse, error) {
	var res InterrogateResponse
	for _, ua := range query.UserAgents {
		if ua == "HAPROXY-load-balancing-check" {
			msg := "Skipping Fingerbank API interrogate because this query is for the PacketFence server itself"
			fclog.Logger("").Info(msg)
			return &res, nil
		}
	}
	err := jsonQuery("GET", "/api/v2/combinations/interrogate", []int{200, 404}, query, &res)
	return &res, err
}

func EndpointIsA(endpointDeviceId, wantsDeviceId int) (EndpointIsAResponse, error) {
	var res EndpointIsAResponse
	err := jsonQuery("GET", fmt.Sprintf("/api/v2/devices/%d/is_a/%d", endpointDeviceId, wantsDeviceId), []int{200}, map[string]string{}, &res)
	return res, err
}

func GetNetworkProfile(deviceId int) (NetworkProfile, error) {
	var res NetworkProfile
	err := jsonQuery("GET", fmt.Sprintf("/api/v2/devices/%d/profiling_rules", deviceId), []int{200}, map[string]string{}, &res)
	return res, err
}

func DeviceChildsIds(deviceId int) ([]int, error) {
	var childsIds []int
	err := jsonQuery("GET", fmt.Sprintf("/api/v2/devices/%d/childs_ids", deviceId), []int{200}, map[string]string{}, &childsIds)
	return childsIds, err
}

func DeviceVirtualChildsIds(deviceId int) ([]int, error) {
	var childsIds []int
	err := jsonQuery("GET", fmt.Sprintf("/api/v2/devices/%d/virtual_childs_ids", deviceId), []int{200}, map[string]string{}, &childsIds)
	return childsIds, err
}

func FetchTextFile(path string) ([]string, error) {
	req, err := http.NewRequest(
		"GET",
		fingerbankURI(path),
		nil,
	)

	resp, err := doQuery(req, []int{200})

	if err != nil {
		return nil, err
	}

	defer resp.Body.Close()
	scanner := bufio.NewScanner(resp.Body)

	hosts := make([]string, 0)
	for scanner.Scan() {
		line := scanner.Text()
		hosts = append(hosts, line)
	}

	return hosts, nil
}

func DestinationHostsAnalysisMap() ([]string, []*regexp.Regexp, *host_analyser.HostAnalyser, error) {
	lines, err := FetchTextFile("/api/v2/download/destination-hosts-analysis")

	logger := fclog.Logger("")

	if err != nil {
		return nil, nil, nil, err
	}

	strictWhitelistMap := make(map[int][]filter_engine.Condition)

	var stringHosts []string
	var regexpHosts []*regexp.Regexp

	for _, line := range lines {
		r := csv.NewReader(strings.NewReader(line))
		info, err := r.Read()

		if err == io.EOF {
			logger.Error(fmt.Sprintf("Ignoring line '%s' in destination hosts analysis because it seems to be empty", line))
			continue
		}
		if err != nil {
			logger.Error(fmt.Sprintf("Ignoring line '%s' in destination hosts analysis because it generated a CSV error %s", line, err))
			continue
		}
		if len(info) != 3 {
			logger.Error(fmt.Sprintf("Ignoring line '%s' in destination hosts analysis because it doesn't have the right amount of info", line))
			continue
		}

		analysisType := strings.Trim(info[0], " ")

		matcher := strings.Trim(info[1], " ")

		deviceStr := strings.Trim(info[2], " ")
		deviceId64, err := strconv.ParseInt(deviceStr, 10, 64)
		sharedutils.CheckError(err)
		deviceId := int(deviceId64)

		var c filter_engine.Condition
		if matches := regexp.MustCompile(`^/(.*)/$`).FindStringSubmatch(matcher); len(matches) == 2 {
			if re, err := regexp.Compile(matches[1]); err != nil {
				logger.Error(fmt.Sprintf("Ignoring line '%s' in the destination hosts analysis because its regex is invalid", line))
				continue
			} else {
				regexpHosts = append(regexpHosts, re)
				c = filter_engine.NewCaseInsensitiveRegexpCondition(matches[1])
			}
		} else {
			stringHosts = append(stringHosts, matcher)
			c = filter_engine.NewEqualCondition(matcher)
		}

		switch analysisType {
		case "strict_whitelist":
			if _, found := strictWhitelistMap[deviceId]; !found {
				strictWhitelistMap[deviceId] = make([]filter_engine.Condition, 0)
			}
			strictWhitelistMap[deviceId] = append(strictWhitelistMap[deviceId], c)
		default:
			logger.Error(fmt.Sprintf("Ignoring line '%s' in the destination hosts analysis because it doesn't have a valid analysis type", line))
			continue
		}
	}

	return stringHosts, regexpHosts, host_analyser.NewHostAnalyser(strictWhitelistMap), nil
}

func DestinationHosts() ([]string, []*regexp.Regexp, error) {
	logger := fclog.Logger("")

	allHosts, err := FetchTextFile("/api/v2/download/destination-hosts")

	if err != nil {
		return nil, nil, err
	}

	var stringHosts []string
	var regexpHosts []*regexp.Regexp

	for _, s := range allHosts {
		if matches := regexp.MustCompile(`^/(.*)/$`).FindStringSubmatch(s); len(matches) == 2 {
			if re, err := regexp.Compile(matches[1]); err != nil {
				logger.Error(fmt.Sprintf("Ignoring line '%s' in the destination hosts map because its regex is invalid", s))
			} else {
				regexpHosts = append(regexpHosts, re)
			}
		} else {
			stringHosts = append(stringHosts, s)
		}
	}

	return stringHosts, regexpHosts, nil
}
