package fcdhcpv4handler

import (
	"encoding/binary"
	"errors"
	"flag"
	"net"
	"strconv"
	"strings"
	"time"

	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"
	"github.com/patrickmn/go-cache"
)

const (
	Name       = "dhcpv4"
	DefaultBPF = "udp and (port 67 or port 68 or port 767)"
)

var updateOnRequest = flag.Bool("dhcpv4-update-mac-on-request", true, "Update the MAC address associated to an IP on the DHCP request instead of the DHCP ACK. Should only be set if the collector won't be seeing the DHCP ACK packets.")

const DHCPMagic uint32 = 0x63825363

const (
	BOOTREQUEST = 1
	BOOTREPLY   = 2
)

const BootSequenceDiscoverDelay = time.Second * 5

const (
	DHCPDISCOVER = 1
	DHCPOFFER    = 2
	DHCPREQUEST  = 3
	DHCPDECLINE  = 4
	DHCPACK      = 5
	DHCPNAK      = 6
	DHCPRELEASE  = 7
	DHCPINFORM   = 8
)

const (
	NetMask            = 1
	RequestedIPAddress = 50
	MessageType        = 53
	ParameterRequest   = 55
	VendorClass        = 60
	HostName           = 12
)

// DHCPv4Option rerpresents a DHCP option.
type DHCPv4Option struct {
	Type   uint8
	Length uint8
	Data   []byte
}

type DHCPv4 struct {
	Op              uint8
	Htype           uint8
	Hlen            uint8
	Hops            uint8
	Xid             uint32
	Secs            uint16
	Flags           uint16
	ClientIp        net.IP
	YourIp          net.IP
	ServerIp        net.IP
	GatewayIp       net.IP
	HardwareAddress []byte
	ServerName      []byte
	File            []byte
	Options         []DHCPv4Option
}

// Create a new handler
func NewDHCPv4Handler(collector fccommon.Collector) fccommon.ProtocolHandler {
	h := &DHCPv4Handler{
		GenericHandler:    fccommon.GenericHandler{Collector: collector},
		DHCPDiscoverCache: cache.New(5*time.Second, 2*time.Minute),
	}
	return h
}

// Returns the BPF of the packets the protocol handler is interested in
func (h *DHCPv4Handler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h *DHCPv4Handler) String() string {
	return "DHCPv4 Handler"
}

// Name that represents this protocol handler
func (h *DHCPv4Handler) Name() string {
	return Name
}

func DHCPv4OptionFromBytes(option *DHCPv4Option, data []byte) error {
	if len(data) < 1 {
		return errors.New("Invalid length for option")
	}
	option.Type = data[0]
	if option.Type == 255 || option.Type == 0 {
		option.Data = nil
	} else {
		if len(data) == 1 {
			return errors.New("Invalid length for option")
		}
		option.Length = data[1]
		if len(data) < int(option.Length) {
			return errors.New("Invalid length for option")
		}
		option.Data = data[2 : option.Length+2]
	}
	return nil
}

func DHCPv4FromBytes(dhcp *DHCPv4, data []byte) error {
	length := len(data)
	if length < 240 {
		return errors.New("Bad DHCP length")
	}
	magic := binary.BigEndian.Uint32(data[236:240])
	if magic != DHCPMagic {
		return errors.New("Bad DHCP magic")
	}
	dhcp.Op = data[0]
	dhcp.Htype = data[1]
	dhcp.Hlen = data[2]
	if dhcp.Hlen > 16 {
		return errors.New("Bad Hardware length")
	}
	dhcp.Hops = data[3]
	dhcp.Xid = binary.BigEndian.Uint32(data[4:8])
	dhcp.Secs = binary.BigEndian.Uint16(data[8:10])
	dhcp.Flags = binary.BigEndian.Uint16(data[10:12])
	dhcp.ClientIp = net.IP(data[12:16])
	dhcp.YourIp = net.IP(data[16:20])
	dhcp.ServerIp = net.IP(data[20:24])
	dhcp.GatewayIp = net.IP(data[24:28])
	dhcp.HardwareAddress = data[28 : 28+dhcp.Hlen]
	dhcp.ServerName = data[44:108]
	dhcp.File = data[108:236]
	start := 240
	for start < length {
		option_type := data[start]
		// Skip padding
		if option_type == 0 {
			start++
			continue
		}
		// Stop at End Option
		if option_type == 255 {
			break
		}
		option := DHCPv4Option{}
		err := DHCPv4OptionFromBytes(&option, data[start:])
		if err != nil {
			return err
		}
		dhcp.Options = append(dhcp.Options, option)
		start += 2 + int(option.Length)
	}
	return nil
}

type DHCPv4Handler struct {
	fccommon.GenericHandler
	DHCPDiscoverCache *cache.Cache
}

// The handler for DHCP packets
func (h *DHCPv4Handler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	dhcp := DHCPv4{}
	err := DHCPv4FromBytes(&dhcp, l4Payload)
	if err != nil {
		return
	}
	var messageType uint8
	for _, o := range dhcp.Options {
		if o.Type == MessageType {
			messageType = o.Data[0]
			break
		}
	}

	switch messageType {
	case DHCPDISCOVER:
		h.handleDHCPDiscover(&dhcp)
	case DHCPOFFER:
		h.handleDHCPOffer(&dhcp)
	case DHCPREQUEST:
		h.handleDHCPRequest(&dhcp)
	case DHCPDECLINE:
		h.handleDHCPDecline(&dhcp)
	case DHCPACK:
		h.handleDHCPAck(&dhcp)
	case DHCPNAK:
		h.handleDHCPNak(&dhcp)
	case DHCPRELEASE:
		h.handleDHCPRelease(&dhcp)
	case DHCPINFORM:
		h.handleDHCPInform(&dhcp)
	}
}

func (h *DHCPv4Handler) handleDHCPDiscover(dhcp *DHCPv4) {
	mac, err := fccommon.MacFromBytes(dhcp.HardwareAddress)
	if err != nil {
		return
	}

	h.DHCPDiscoverCache.Set(mac, time.Now(), cache.DefaultExpiration)
}

func (h *DHCPv4Handler) handleDHCPOffer(dhcp *DHCPv4) {
	h.startBootSequenceWhenMatchingNetwork(dhcp)
}

func (h *DHCPv4Handler) startBootSequenceWhenMatchingNetwork(dhcp *DHCPv4) {
	mac, err := fccommon.MacFromBytes(dhcp.HardwareAddress)
	if err != nil {
		return
	}

	if _, found := h.DHCPDiscoverCache.Get(mac); !found {
		return
	}

	netmaskOption, found := dhcp.getOption(NetMask)
	if !found {
		return
	}

	netmask := net.IPv4Mask(
		netmaskOption.Data[0],
		netmaskOption.Data[1],
		netmaskOption.Data[2],
		netmaskOption.Data[3],
	)

	oldIpStr, err := h.Collector.MAC2IP(mac)
	if err != nil {
		return
	}

	oldIp := net.ParseIP(oldIpStr)
	if oldIp == nil {
		return
	}

	net := net.IPNet{
		IP:   dhcp.YourIp.Mask(netmask),
		Mask: netmask,
	}

	if net.Contains(oldIp) {
		ed, err := h.Collector.FindEndpointDataMAC(mac)
		if err != nil {
			return
		}

		h.DHCPDiscoverCache.Delete(mac)
		ed.Update(func(ed *fccommon.EndpointData) bool {
			ed.BootSequence.SoftStart()
			ed.BootSequence.Add(strconv.FormatInt(time.Now().Unix()*1000, 10) + ":" + "dhcpv4_startBootSequenceWhenMatchingNetwork")
			return false
		})
	}
}

func (d *DHCPv4) getOption(t uint8) (DHCPv4Option, bool) {
	for _, o := range d.Options {
		if o.Type == t {
			return o, true
		}
	}

	return DHCPv4Option{}, false
}

type endPointData struct {
	vendor      string
	fingerprint string
	host        string
	requestedIP net.IP
}

func (h *DHCPv4Handler) handleDHCPRequest(dhcp *DHCPv4) {
	mac, err := fccommon.MacFromBytes(dhcp.HardwareAddress)
	if err != nil {
		return
	}
	if mac == "000000000000" {
		return
	}
	ep := endPointData{}
	extractEndPointData(dhcp.Options, &ep)
	if ep.fingerprint == "" && ep.vendor == "" && ep.host == "" {
		return
	}

	h.startBootSequenceOnRequest(mac, ep.requestedIP)
	ed, err := h.Collector.FindEndpointDataMAC(mac)
	if err != nil {
		return
	}

	dhcpVendorBefore := ed.DhcpVendor
	ed.Update(func(ed *fccommon.EndpointData) bool {
		changed := false
		if ed.DhcpFingerprint != ep.fingerprint || ed.DhcpVendor != ep.vendor || ed.Hostname != ep.host {
			changed = true
		}

		ed.DhcpFingerprint = ep.fingerprint
		ed.DhcpVendor = ep.vendor
		ed.Hostname = ep.host
		return changed
	})
	h.Collector.CheckDhcpVendorChange(ed, dhcpVendorBefore)

	if *updateOnRequest {
		h.Collector.UpdateIPMAC(ep.requestedIP.String(), mac)
	}
}

func (h *DHCPv4Handler) startBootSequenceOnRequest(mac string, ip net.IP) bool {
	if _, found := h.DHCPDiscoverCache.Get(mac); !found {
		return false
	}

	oldMac, err := h.Collector.IP2MAC(ip.String())
	if err != nil {
		return false
	}

	ed, err := h.Collector.FindEndpointDataMAC(mac)
	if err != nil {
		return false
	}

	if oldMac == mac {
		h.DHCPDiscoverCache.Delete(mac)
		ed.Update(func(ed *fccommon.EndpointData) bool {
			ed.BootSequence.SoftStart()
			ed.BootSequence.Add(strconv.FormatInt(time.Now().Unix()*1000, 10) + ":" + "dhcpv4_startBootSequenceOnRequest")
			return false
		})
		return true
	}

	return false
}

func (h *DHCPv4Handler) handleDHCPDecline(dhcp *DHCPv4) {

}

func (h *DHCPv4Handler) handleDHCPAck(dhcp *DHCPv4) {
	h.startBootSequenceWhenMatchingNetwork(dhcp)
	mac, err := fccommon.MacFromBytes(dhcp.HardwareAddress)
	if err != nil {
		return
	}
	if mac == "000000000000" {
		return
	}
	var ip string
	if !dhcp.YourIp.Equal(net.IPv4zero) {
		ip = dhcp.YourIp.String()
	} else if !dhcp.ClientIp.Equal(net.IPv4zero) {
		ip = dhcp.ClientIp.String()
	} else {
		return
	}

	if !*updateOnRequest {
		h.Collector.UpdateIPMAC(ip, mac)
	}
}

func (h *DHCPv4Handler) handleDHCPNak(dhcp *DHCPv4) {

}

func (h *DHCPv4Handler) handleDHCPRelease(dhcp *DHCPv4) {

}

func (h *DHCPv4Handler) handleDHCPInform(dhcp *DHCPv4) {

}

func extractEndPointData(options []DHCPv4Option, ep *endPointData) {
	for _, o := range options {
		switch o.Type {
		case ParameterRequest:
			var tmp []string
			for _, b := range o.Data {
				tmp = append(tmp, strconv.FormatUint(uint64(b), 10))
			}
			ep.fingerprint = strings.Join(tmp, ",")
		case VendorClass:
			ep.vendor = string(o.Data[:])
		case HostName:
			ep.host = string(o.Data[:])
		case RequestedIPAddress:
			ep.requestedIP = net.IP(o.Data[:])
		}
	}
}
