package fcdnshandler

import (
	"net"
	"strings"

	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"

	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"
	"github.com/inverse-inc/packetfence-dhcp-forwarder/forwarder"
)

const (
	Name       = "dns"
	DefaultBPF = "port 53 or port 753"
)

type DNSHandler struct {
	fccommon.GenericHandler
}

// Create a new handler
func NewDNSHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	eh := &DNSHandler{}
	eh.Collector = collector
	return eh
}

// Handle function to process a packet
// info is the PacketInfo associated to this connection
// l4Payload is the bytes of the TCP or UDP packet
// rawPacket is the bytes of the packet as received on the wire
func (h *DNSHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {

	//defer fmt.Println("DNS has handled the packet for SRC IP", info.SrcIP, "DST IP", info.DstIP)

	var dns layers.DNS
	parser := gopacket.NewDecodingLayerParser(layers.LayerTypeDNS, &dns)

	var decoded = make([]gopacket.LayerType, 1)
	if err := parser.DecodeLayers(l4Payload, &decoded); err != nil {
		if strings.HasPrefix(err.Error(), "qname '0x40'") || strings.HasPrefix(err.Error(), "qname '0x80'") {
			return // unsupported yet
		}
		return
	}

	src, dst := info.SrcIP, info.DstIP

additionals:
	for _, rr := range dns.Additionals {
		if rr.Type != layers.DNSTypeOPT {
			continue
		}

		for _, opt := range rr.OPT {
			if opt.Code == forwarder.PFForwardSrcDst {
				src, dst = net.IP(opt.Data[:len(opt.Data)/2]).String(), net.IP(opt.Data[len(opt.Data)/2:]).String()
				break additionals
			}
		}
	}

	hosts := []string{}
	for _, rr := range dns.Questions {
		if rr.Type == layers.DNSTypeA || rr.Type == layers.DNSTypeAAAA {
			hosts = append(hosts, string(rr.Name[:]))
		} else {
			continue
		}

		ed, err := h.Collector.FindEndpointDataIP(src)
		if err != nil {
			continue
		}
		h.Collector.AddDestinationHost(ed, string(rr.Name[:]), 0, "udp")

	}

	for _, rr := range dns.Answers {
		mac, err := h.Collector.IP2MAC(dst)
		if err != nil {
			continue
		}

		if rr.Type == layers.DNSTypeA {
			for _, host := range hosts {
				h.Collector.AddIpHostMapping(mac, rr.IP.String(), host)
			}
		}
	}
}

// Returns the BPF of the packets the protocol handler is interested in
func (h *DNSHandler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h *DNSHandler) String() string {
	return "DNS Handler"
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h *DNSHandler) Name() string {
	return Name
}
