package fchttphandler

import (
	"bufio"
	"bytes"
	"net/http"
	"strings"

	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"
)

const (
	Name       = "http"
	DefaultBPF = "(tcp dst port 80)"
)

type HTTPHandler struct {
	fccommon.GenericHandler
}

// Create a new handler
func NewHTTPHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	eh := &HTTPHandler{}
	eh.Collector = collector
	return eh
}

// We need to build a request from the l4Payload to allow net/http to extract UserAgent and Host
func (h HTTPHandler) buildHTTPRequest(l4Payload []byte) (*http.Request, error) {
	buf := bytes.NewBuffer(l4Payload)
	read := bufio.NewReader(buf)
	return http.ReadRequest(read)
}

// Extract the UserAgent of the request
func (h HTTPHandler) extractUserAgent(request *http.Request) string {
	return request.UserAgent()
}

// Extract the Host of the request
func (h HTTPHandler) extractHost(request *http.Request) string {
	return request.Host
}

// Extract the correct destination host if there is a X-Forward
func (h HTTPHandler) extractXForward(request *http.Request) string {
	xf := request.Header.Get("X-Forwarded-For")

	return strings.Split(xf, ",")[0]
}

// Handle function to process a packet
// info is the PacketInfo associated to this connection
// l4Payload is the bytes of the TCP or UDP packet
// rawPacket is the bytes of the packet as received on the wire
// Update the User Agent for the defined matching IP, do nothing if IP is unknown
// Update the destination Host if valid for the matching IP, else do nothing
func (h HTTPHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	request, err := h.buildHTTPRequest(l4Payload)

	if err != nil {
		return
	}

	ip := info.SrcIP

	if xf := h.extractXForward(request); xf != "" {
		ip = xf
	}

	ed, err := h.Collector.FindEndpointDataIP(ip)

	if err != nil {
		return
	}

	if ua := h.extractUserAgent(request); ua != "" {
		ed.AddUserAgent(ua)
	}

	if host := h.extractHost(request); host != "" {
		h.Collector.AddDestinationHost(ed, host, 80, "tcp")
	}

	//fmt.Println("The HTTP handler has handled the packet")
}

// Returns the BPF of the packets the protocol handler is interested in
func (h HTTPHandler) GetBPF() string {
	return "(tcp port 80)"
}

// Friendly string representation of this protocol handler
func (h HTTPHandler) String() string {
	return "HTTP Handler"
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h HTTPHandler) Name() string {
	return "http"
}
