package fcmdnshandler

import (
	"fmt"
	"strings"
	"sync"
	"time"

	"git.inverse.ca/inverse/fingerbank-collector/common"

	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"

	"github.com/patrickmn/go-cache"
)

const (
	Name = "mdns"
	//		(
	//		  (
	//			((ether host 01:00:5E:00:00:FB) and (host 224.0.0.251))
	//			or
	//		    ((ether host 33:33:00:00:00:FB) and (host FF02::FB))
	//		  )
	//			and (port 5353)
	//		)
	DefaultBPF = "((((ether host 01:00:5E:00:00:FB) and (host 224.0.0.251)) or ((ether host 33:33:00:00:00:FB) and (host FF02::FB))) and (port 5353))"
)

type MDNSHandler struct {
	fccommon.GenericHandler
	Cache *cache.Cache // cache for known DNS records
	lock  *sync.Mutex
}

// Create a new handler
func NewMDNSHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	h := &MDNSHandler{}
	h.Collector = collector
	// Default expiration of 60 minutes, purge the cache every 2 minutes
	h.Cache = cache.New(60*time.Minute, 2*time.Minute)
	h.lock = &sync.Mutex{}
	return h
}

// Returns the BPF of the packets the protocol handler is interested in
func (h MDNSHandler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h MDNSHandler) String() string {
	return "MDNS Handler"
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h MDNSHandler) Name() string {
	return Name
}

// Handle function to process a packet
// info is the PacketInfo associated to this connection
// l4Payload is the bytes of the TCP or UDP packet
// rawPacket is the bytes of the packet as received on the wire
func (h MDNSHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {

	var dns layers.DNS
	parser := gopacket.NewDecodingLayerParser(layers.LayerTypeDNS, &dns)

	var decoded = make([]gopacket.LayerType, 1)
	if err := parser.DecodeLayers(l4Payload, &decoded); err != nil {
		if strings.HasPrefix(err.Error(), "qname '0x40'") || strings.HasPrefix(err.Error(), "qname '0x80'") {
			return // unsupported yet
		}
		fmt.Sprintln("Error while decoding mDNS packet: ", err)
	}
	if dns.OpCode != layers.DNSOpCodeQuery {
		return // not a query/reply
	}

	if dns.QR {
		// answer
		if dns.AA != true {
			return // answer is not authoritative
		}
		for _, rr := range dns.Answers {
			h.handleAnswer(rr)
		}

	} else {
		// query
		ed, err := h.Collector.FindEndpointDataIP(info.SrcIP)
		if err != nil {
			return
		}

		for _, rr := range dns.Questions {
			if rr.Type == layers.DNSTypeA || rr.Type == layers.DNSTypeAAAA {
				ed.AddMDNSRequest(string(rr.Name[:]))
			}
		}
	}

}

func (h MDNSHandler) handleAnswer(rr layers.DNSResourceRecord) error {
	switch rr.Type {
	case layers.DNSTypeA:
		fallthrough
	case layers.DNSTypeAAAA:
		ed, err := h.Collector.FindEndpointDataIP(rr.IP.String())
		if err != nil {
			return err
		}

		if services := h.FindServices(rr.Name); services == nil {
			return nil
		} else {
			ed.AddMDNSServices(services)
		}
	case layers.DNSTypePTR:
		h.AddToSrv(string(rr.PTR[:]), string(rr.Name[:]))

	case layers.DNSTypeSRV:
		h.AddToHosts(string(rr.SRV.Name[:]), string(rr.Name[:]), rr.TTL)
	}
	return nil
}

func (h MDNSHandler) FindServices(name []byte) []string {
	host := string(name[:])
	services := make([]string, 0)
	srv := h.GetKeys(h.GetSrvByHost(host))
	if srv == nil {
		return nil
	}

	for _, s := range srv {
		ptr := h.GetKeys(h.GetPtrBySrv(s))
		for _, p := range ptr {
			services = append(services, p)
		}
	}
	return services
}

// Called to perform an update on a cached value making sure no-one updates it in the meanwhile
// Helps prevent race conditions when changing data
// TODO: Add ability to timeout getting the lock and timeout on updating
func (h *MDNSHandler) Update(todo func(*MDNSHandler)) {
	h.lock.Lock()
	defer h.lock.Unlock()
	todo(h)
}

// Add to known SRV records
func (h MDNSHandler) AddToSrv(srv string, ptr string) {

	h.Update(func(h *MDNSHandler) {
		var srv2ptr map[string]bool
		if srv2ptr = h.GetPtrBySrv(srv); srv2ptr == nil {
			srv2ptr = make(map[string]bool)
		}
		srv2ptr[ptr] = true
		key := "Srv2Ptr:" + srv
		h.Cache.Set(key, &srv2ptr, cache.DefaultExpiration)
	})
}

// Add to known A (and AAAA) records
func (h MDNSHandler) AddToHosts(host string, srv string, ttl uint32) {
	h.Update(func(h *MDNSHandler) {
		var host2srv map[string]bool
		if host2srv = h.GetSrvByHost(host); host2srv == nil {
			host2srv = make(map[string]bool)
		}
		host2srv[srv] = true
		key := "Host2Srv:" + host
		h.Cache.Set(key, &host2srv, time.Duration(ttl)*time.Second)
	})
}

func (h MDNSHandler) GetSrvByHost(host string) map[string]bool {
	key := "Host2Srv:" + host
	return h.GetFromCache(key)
}

func (h MDNSHandler) GetPtrBySrv(srv string) map[string]bool {
	key := "Srv2Ptr:" + srv
	return h.GetFromCache(key)
}

func (h *MDNSHandler) GetFromCache(k string) map[string]bool {
	value, found := h.Cache.Get(k)
	if !found {
		return nil
	}
	kmap := *(value.(*map[string]bool)) // type assertion to force to a map
	return kmap
}

func (h MDNSHandler) GetKeys(m map[string]bool) []string {
	h.lock.Lock()
	defer h.lock.Unlock()
	i := 0
	keys := make([]string, len(m))
	for k := range m {
		keys[i] = k
		i++
	}
	return keys
}
