package fcradiushandler

import (
	"bytes"
	"encoding/binary"
	"fmt"
	"net"

	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"
	"github.com/bronze1man/radius"
)

const (
	DefaultBPF = "udp port 1813"
	Name       = "radius"
	PrettyName = "RADIUS Handler"
)

type RadiusHandler struct {
	fccommon.GenericHandler
}

func NewRadiusHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	eh := &RadiusHandler{}
	eh.Collector = collector
	return eh
}

func (h RadiusHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	pac, err := h.Parse(l4Payload)
	if err != nil {
		return
	}
	if pac.HasAVP(radius.FramedIPAddress) && pac.HasAVP(radius.CallingStationId) {
		err = h.Collector.UpdateIPMAC(pac.GetAVP(radius.FramedIPAddress).Decode(pac).(net.IP).String(), pac.GetAVP(radius.CallingStationId).Decode(pac).(string))
		if err != nil {
			fmt.Println(err.Error())
			return
		}
	}
	vsa := h.DecodeAVP(pac)
	var ed *fccommon.EndpointData
	if pac.HasAVP(radius.CallingStationId) {
		ed, err = h.Collector.FindEndpointDataMAC(pac.GetAVP(radius.CallingStationId).Decode(pac).(string))
		if err != nil {
			return
		}
	} else {
		return
	}
	if v1 := vsa["parameter-request-list"]; v1 != nil {
		ed.Update(func(ed *fccommon.EndpointData) bool {
			v := vsa["parameter-request-list"].(string)
			if ed.DhcpFingerprint != v {
				ed.DhcpFingerprint = v
				return true
			} else {
				return false
			}
		})
	}
	if v1 := vsa["class-identifier"]; v1 != nil {
		ed.Update(func(ed *fccommon.EndpointData) bool {
			v := vsa["class-identifier"].(string)
			if ed.DhcpVendor != v {
				ed.DhcpVendor = v
				return true
			} else {
				return false
			}
		})
	}
	if v1 := vsa["host-name"]; v1 != nil {
		ed.Update(func(ed *fccommon.EndpointData) bool {
			v := vsa["host-name"].(string)
			if ed.Hostname != v {
				ed.Hostname = v
				return true
			} else {
				return false
			}
		})
	}

	if v1 := vsa["cdp-capabilities-type"]; v1 != nil {
		for v, _ := range vsa["cdp-capabilities-type"].(map[string]bool) {
			ed.AddCDPCapability(v)
		}
	}

	if v1 := vsa["system-capabilities"]; v1 != nil {
		for v, _ := range vsa["system-capabilities"].(map[string]bool) {
			ed.AddLLDPCapability(v)
		}
	}
}

func (h RadiusHandler) Parse(l4Payload []byte) (p *radius.Packet, err error) {
	pac, err := radius.DecodePacket("s3cr3t", l4Payload)
	return pac, err
}

// Returns the BPF of the packets the protocol handler is interested in
func (h RadiusHandler) GetBPF() string {
	return "(udp port 1813)"
}

// Friendly string representation of this protocol handler
func (h RadiusHandler) String() string {
	return "Radius Accounting Handler"
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h RadiusHandler) Name() string {
	return "Radius Accounting"
}

// DecodeAVP to
func (h RadiusHandler) DecodeAVP(pac *radius.Packet) (vsa map[string]interface{}) {
	vsa = make(map[string]interface{})
	for j := range pac.AVPs {
		if pac.AVPs[j].Type == radius.VendorSpecific {
			bob := bytes.SplitN(pac.AVPs[j].Value[6:], []byte("="), 2)
			desc := string(bob[0])
			switch desc {
			case "dhcp-option", "http-tlv", "lldp-tlv", "cdp-tlv":
				id := int(binary.BigEndian.Uint16(bob[1][0:2]))
				vsa[Tlv.Tlvlist[desc].TlvType[id].Desc] = Tlv.Tlvlist[desc].TlvType[id].Extra.extra(Tlv.Tlvlist[desc].TlvType[id].Decode, bob[1][4:])
			default:
				// value = string(bob[1])
			}
		}
	}
	return vsa
}
