package fctcphandler

import (
	"encoding/binary"
	"fmt"
	"github.com/google/gopacket/layers"
	"strings"
)

/* IP-level quirks: */

type QuirkFlags uint32

const (
	QuirkDontFragment      QuirkFlags = 1 << iota /* DF used (probably PMTUD)      */
	QuirkNonZeroId                                /* Non-zero IDs when DF set      */
	QuirkZeroId                                   /* Zero IDs when DF not set      */
	QuirkECN                                      /* ECN supported                 */
	QuirkNotZeroMustBeZero                        /* IP "must be zero" field isn't */
	QuirkFlow                                     /* IPv6 flows used               */

	/* Core TCP quirks: */

	QuirkZeroSeq       /* SEQ is zero                        */
	QuirkNonZeroAck    /* ACK non-zero when ACK flag not set */
	QuirkZeroAck       /* ACK is zero when ACK flag set      */
	QuirkNonZeroUrgent /* URG non-zero when URG flag not set */
	QuirkUrgent        /* URG flag set                       */
	QuirkPush          /* PUSH flag on a control packet      */

	/* TCP option quirks: */

	QuirkOptionZeroTS1          /* Own timestamp set to zero      */
	QuirkOptionNonZeroTS2       /* Peer timestamp non-zero on SYN */
	QuirkOptionNonZeroEol       /* Non-zero padding past EOL      */
	QuirkOptionLargeWindowScale /* Excessive window scaling       */
	QuirkOptionBad              /* Problem parsing TCP options    */

	/* First and Last flags */

	quirk_last, quirk_allowed_flags QuirkFlags = 1 << iota, 0xFFFFFFFF >> (32 - iota) /* The last flags and allowed*/
	quirk_first                                = QuirkDontFragment
)

const TCP_IP4_MIN_SIZE = 40 // Minimum length of TCP Header 20 + Ipv4 header 20
const TCP_IP6_MIN_SIZE = 60 // Minimum length of TCP Header 20 + Ipv6 header 40

func ROL32(x, r uint32) uint32 {
	return (x << r) | (x >> (32 - r))
}

func hash32(data []byte, seed uint32) uint32 {

	var a, b, c uint32
	k := 0
	length := len(data)
	a = 0xdeadbeef + uint32(length) + seed
	b = a
	c = a

	for length > 12 {

		a += binary.BigEndian.Uint32(data[k : k+4])
		k += 4
		b += binary.BigEndian.Uint32(data[k : k+4])
		k += 4
		c += binary.BigEndian.Uint32(data[k : k+4])
		k += 4

		a -= c
		a ^= ROL32(c, 4)
		c += b
		b -= a
		b ^= ROL32(a, 6)
		a += c
		c -= b
		c ^= ROL32(b, 8)
		b += a
		a -= c
		a ^= ROL32(c, 16)
		c += b
		b -= a
		b ^= ROL32(a, 19)
		a += c
		c -= b
		c ^= ROL32(b, 4)
		b += a

		length -= 12
	}

	switch length {

	case 12:
		c += binary.BigEndian.Uint32(data[k+8 : k+12])
		b += binary.BigEndian.Uint32(data[k+4 : k+8])
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 11:
		c += (uint32(binary.BigEndian.Uint16(data[k+8:k+10])) << 8) | uint32(data[k+10])
		b += binary.BigEndian.Uint32(data[k+4 : k+8])
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 10:
		c += uint32(binary.BigEndian.Uint16(data[k+8 : k+10]))
		b += binary.BigEndian.Uint32(data[k+4 : k+8])
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 9:
		c += uint32(data[k+8])
		b += binary.BigEndian.Uint32(data[k+4 : k+8])
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 8:
		b += binary.BigEndian.Uint32(data[k+4 : k+8])
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 7:
		b += (uint32(binary.BigEndian.Uint16(data[k+4:k+6])) << 8) | uint32(data[k+6])
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 6:
		c += uint32(binary.BigEndian.Uint16(data[k+8 : k+10]))
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 5:
		b += uint32(data[k+4])
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 4:
		a += binary.BigEndian.Uint32(data[k : k+4])

	case 3:
		a += uint32(binary.BigEndian.Uint16(data[k:k+2])<<8) | uint32(data[k+2])

	case 2:
		a += uint32(binary.BigEndian.Uint16(data[k : k+2]))

	case 1:
		a += uint32(data[k])

	case 0:
		return c

	}

	c ^= b
	c -= ROL32(b, 14)
	a ^= c
	a -= ROL32(c, 11)
	b ^= a
	b -= ROL32(a, 25)
	c ^= b
	c -= ROL32(b, 16)
	a ^= c
	a -= ROL32(c, 4)
	b ^= a
	b -= ROL32(a, 14)
	c ^= b
	c -= ROL32(b, 24)

	return c

}

func (q QuirkFlags) String() string {
	// quirks
	quirks := []string{}
	for quirk_flag := quirk_first; quirk_flag != quirk_last; quirk_flag <<= 1 {
		if (quirk_flag & q) != quirk_flag {
			continue
		}
		if quirk, ok := QuirkFlagsStrings[quirk_flag]; ok {
			quirks = append(quirks, quirk)
		}
	}
	return strings.Join(quirks, ",")
}

var QuirkFlagsStrings = map[QuirkFlags]string{
	QuirkDontFragment:      "df",
	QuirkNonZeroId:         "id+",
	QuirkZeroId:            "id-",
	QuirkECN:               "ecn",
	QuirkNotZeroMustBeZero: "0+",
	QuirkFlow:              "flow",

	QuirkZeroSeq:       "seq-",
	QuirkNonZeroAck:    "ack+",
	QuirkZeroAck:       "ack-",
	QuirkNonZeroUrgent: "uptr+",
	QuirkUrgent:        "urgf+",
	QuirkPush:          "pushf+",

	QuirkOptionZeroTS1:          "ts1-",
	QuirkOptionNonZeroTS2:       "ts2+",
	QuirkOptionNonZeroEol:       "opt+",
	QuirkOptionLargeWindowScale: "exws",
	QuirkOptionBad:              "bad",
}

/*
// TCPOptionKind represents a TCP option code.
type TCPOptionKind uint8

const (
	TCPOptionKindEndList                         = 0
	TCPOptionKindNop                             = 1
	TCPOptionKindMSS                             = 2  // len = 4
	TCPOptionKindWindowScale                     = 3  // len = 3
	TCPOptionKindSACKPermitted                   = 4  // len = 2
	TCPOptionKindSACK                            = 5  // len = n
	TCPOptionKindEcho                            = 6  // len = 6, obsolete
	TCPOptionKindEchoReply                       = 7  // len = 6, obsolete
	TCPOptionKindTimestamps                      = 8  // len = 10
	TCPOptionKindPartialOrderConnectionPermitted = 9  // len = 2, obsolete
	TCPOptionKindPartialOrderServiceProfile      = 10 // len = 3, obsolete
	TCPOptionKindCC                              = 11 // obsolete
	TCPOptionKindCCNew                           = 12 // obsolete
	TCPOptionKindCCEcho                          = 13 // obsolete
	TCPOptionKindAltChecksum                     = 14 // len = 3, obsolete
	TCPOptionKindAltChecksumData                 = 15 // len = n, obsolete
)
*/

var TCPOptionLayoutString = map[layers.TCPOptionKind]string{
	layers.TCPOptionKindNop:           "nop",
	layers.TCPOptionKindMSS:           "mss",
	layers.TCPOptionKindWindowScale:   "ws",
	layers.TCPOptionKindSACKPermitted: "sok",
	layers.TCPOptionKindSACK:          "sack",
	layers.TCPOptionKindTimestamps:    "ts",
}

type TCPSignature struct {
	Ver           int8
	TcpType       uint8
	Ittl          uint8
	OptionLength  uint8
	MSS           int32 // Maximum Segment Size
	SynMSS        int32 // Maximum Segment Size of SYN
	WindowSize    int32
	PayloadLength uint16
	Scale         int16
	Layout        uint32
	Quirks        QuirkFlags
	PClass        uint8
	OptionPad     uint8
	Timestamp1    uint32
	Timestamp2    uint32
	Options       []layers.TCPOption
	HeaderLength  uint16
}

const MAX_DIST uint8 = 35
const SPECIAL_MSS int32 = 1331

func guessDist(ttl uint8) uint8 {
	switch {
	case ttl <= 32:
		return 32 - ttl
	case ttl <= 64:
		return 64 - ttl
	case ttl <= 128:
		return 128 - ttl
	}
	return 255 - ttl
}

// String
// sig = ver:ittl:olen:mss:wsize,scale:olayout:quirks:pclass
func (s *TCPSignature) String() string {
	var sig string = ""
	dist := guessDist(s.Ittl)
	// ver:ittl:olen
	if dist > MAX_DIST {
		sig = fmt.Sprintf("%d:%d+?:%d:", s.Ver, s.Ittl, s.OptionLength)
	} else {
		sig = fmt.Sprintf("%d:%d+%d:%d:", s.Ver, s.Ittl, dist, s.OptionLength)
	}

	// mss
	if s.MSS == SPECIAL_MSS && s.TcpType == TCP_SYN|TCP_ACK {
		sig += "*:"
	} else {
		sig += fmt.Sprintf("%d:", s.MSS)
	}

	// wsize
	multiWindow, isMtu := s.IsMultiWindow()
	if multiWindow > 0 {
		if isMtu {
			sig += fmt.Sprintf("mtu*%d", multiWindow)
		} else {
			sig += fmt.Sprintf("mss*%d", multiWindow)
		}
	} else {
		sig += fmt.Sprintf("%d", s.WindowSize)
	}

	sig += fmt.Sprintf(",%d:", s.Scale)

	// layout
	layouts := []string{}
	for _, option := range s.Options {
		if layout, ok := TCPOptionLayoutString[option.OptionType]; ok {
			layouts = append(layouts, layout)
		} else if option.OptionType == layers.TCPOptionKindEndList {
			layouts = append(layouts, fmt.Sprintf("eol+%d", s.OptionPad))
		} else {
			layouts = append(layouts, fmt.Sprintf("%d", uint(option.OptionType)))
		}
	}
	sig += strings.Join(layouts, ",") + ":" + s.Quirks.String()

	// pclass
	if s.PayloadLength > 0 {
		sig += ":+"
	} else {
		sig += ":0"
	}

	return sig
}

func (s *TCPSignature) IsMultiWindow() (int32, bool) {
	window := int32(s.WindowSize)
	mss := s.MSS
	if window == 0 || mss < 100 {
		return -1, false
	}
	isMtu := false

	if (window % mss) == 0 {
		return window / mss, isMtu
	}

	if s.Timestamp1 != 0 {
		mss12 := s.MSS - 12
		if (window % mss12) == 0 {
			return window / mss12, isMtu
		}
	}

	mss_alt := int32(1500 - TCP_IP4_MIN_SIZE)

	if (window % mss_alt) == 0 {
		return window / mss_alt, isMtu
	}

	mss_alt -= 12

	if (window % mss_alt) == 0 {
		return window / mss_alt, isMtu
	}

	if s.Ver == 6 {
		mss_alt = 1500 - TCP_IP6_MIN_SIZE

		if (window % mss_alt) == 0 {
			return window / mss_alt, isMtu
		}

		mss_alt -= 12

		if (window % mss_alt) == 0 {
			return window / mss_alt, isMtu
		}
	}

	// Checks for MTU
	isMtu = true

	mtu_alt := mss + TCP_IP4_MIN_SIZE
	if (window % mtu_alt) == 0 {
		return window / mtu_alt, isMtu
	}

	mtu_alt = mss + int32(s.HeaderLength)
	if (window % mtu_alt) == 0 {
		return window / mtu_alt, isMtu
	}

	if s.Ver == 6 {
		mtu_alt := mss + TCP_IP6_MIN_SIZE
		if (window % mtu_alt) == 0 {
			return window / mtu_alt, isMtu
		}
	}

	if (window % 1500) == 0 {
		return window / 1500, isMtu
	}
	synMSS := s.SynMSS
	if synMSS > 0 {
		isMtu = false
		if (window % synMSS) == 0 {
			return window / synMSS, isMtu
		}
		synMSS -= 12
		if (window % synMSS) == 0 {
			return window / synMSS, isMtu
		}
	}

	return -1, false
}
