package fctlshandler

import (
	"flag"
	"fmt"

	"github.com/dreadl0ck/ja3"
	"github.com/dreadl0ck/tlsx"
	"github.com/inverse-inc/go-utils/sharedutils"

	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"
)

const (
	Name       = "tls"
	DefaultBPF = "tcp port 443 and (tcp[((tcp[12:1] & 0xf0) >> 2):1] = 0x16) and ( (tcp[((tcp[12:1] & 0xf0) >> 2)+5:1] = 0x01) or (tcp[((tcp[12:1] & 0xf0) >> 2)+5:1] = 0x02) )"
	TLS_PORT   = 443
)

type TLSHandler struct {
	fccommon.GenericHandler
	IPDomainTracker *IPDomainTracker
}

var maxReverseIPDomainTrackSize = flag.Int(
	`tls-max-reverse-ip-domain-track-size`,
	sharedutils.EnvOrDefaultInt("COLLECTOR_TLS_MAX_REVERSE_IP_DOMAIN_TRACK_SIZE", 100),
	`The maximum amount of IP addresses to track for mapping IP addresses to domain names in the TLS handler to perform JA3s profiling of Server Hello packets. A value of 0 will disable profiling of JA3s signatures (COLLECTOR_TLS_MAX_REVERSE_IP_DOMAIN_TRACK_SIZE environment variable).`,
)

// Create a new handler
func NewTLSHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	eh := &TLSHandler{}
	eh.Collector = collector
	if *maxReverseIPDomainTrackSize != 0 {
		eh.IPDomainTracker = NewIPDomainTracker(*maxReverseIPDomainTrackSize)
	}
	return eh
}

func (h *TLSHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	contentType := l4Payload[5]

	switch contentType {
	case 0x1:
		h.handleClientHello(info, l4Payload, rawPacket)
	case 0x2:
		if h.IPDomainTracker != nil {
			h.handleServerHello(info, l4Payload, rawPacket)
		}
	}

}

func (h *TLSHandler) handleClientHello(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	ed, err := h.Collector.FindEndpointDataIP(info.SrcIP)
	if err != nil {
		return
	}
	ch := tlsx.ClientHelloBasic{}
	ch.Unmarshal(l4Payload)
	ja3Hash := fmt.Sprintf("%x", ja3.Digest(&ch))

	ed.AddJA3Fingerprint(ja3Hash)
	if h.IPDomainTracker != nil {
		h.IPDomainTracker.Add(fmt.Sprintf("%s:%d", info.SrcIP, info.SrcPort), fmt.Sprintf("%s:%d", info.DstIP, info.DstPort), ch.SNI, ja3Hash)
	}
	h.Collector.AddDestinationHost(ed, ch.SNI, info.DstPort, "tcp")
}

func (h *TLSHandler) handleServerHello(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	ed, err := h.Collector.FindEndpointDataIP(info.DstIP)
	if err != nil {
		return
	}

	sh := tlsx.ServerHelloBasic{}
	sh.Unmarshal(l4Payload)
	ja3sHash := fmt.Sprintf("%x", ja3.DigestJa3s(&sh))

	lockid := h.IPDomainTracker.RLock()
	defer h.IPDomainTracker.RUnlock(lockid)

	ipDomain, _ := h.IPDomainTracker.Get(fmt.Sprintf("%s:%d", info.DstIP, info.DstPort))
	if domains, found := ipDomain[fmt.Sprintf("%s:%d", info.SrcIP, info.SrcPort)]; found {
		for domain, clientSig := range domains {
			ed.AddJA3SigCouple(domain, info.SrcPort, clientSig, ja3sHash)
		}
	}
}

// Returns the BPF of the packets the protocol handler is interested in
func (h *TLSHandler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h *TLSHandler) String() string {
	return "TLS Handler"
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h *TLSHandler) Name() string {
	return Name
}
