package fcupnphandler

import (
	"bufio"
	"bytes"
	"net/http"
	"strings"

	"git.inverse.ca/inverse/fingerbank-collector/common"
)

const (
	Name       = "upnp"
	DefaultBPF = "(udp port 1900)"
)

type UPnPHandler struct {
	fccommon.GenericHandler
}

// Create a new handler
func NewUPnPHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	eh := &UPnPHandler{}
	eh.Collector = collector
	return eh
}

// We need to build a request from the l4Payload to allow net/http to extract UserAgent and Host
func (h *UPnPHandler) buildUPnPRequest(l4Payload []byte) (*http.Request, error) {
	buf := bytes.NewBuffer(l4Payload)
	read := bufio.NewReader(buf)
	return http.ReadRequest(read)
}

// Extract the UPnP Server string from an UPnP packet
func (h *UPnPHandler) extractServerString(request *http.Request) string {
	if len(request.Header["Server"]) > 0 {
		return request.Header["Server"][0]
	} else {
		return ""
	}
}

// Extract the UPnP User Agent from an UPnP packet
func (h *UPnPHandler) extractUserAgent(request *http.Request) string {
	if len(request.Header["User-Agent"]) > 0 {
		return request.Header["User-Agent"][0]
	} else {
		return ""
	}
}

// Verify SSDP type, discover, announcement (alive or byebye), then extract the matching UPnP value
func (h *UPnPHandler) extractUPnP(request *http.Request) string {

	if ssdp := request.Header.Get("Man"); ssdp == "\"ssdp:discover\"" {
		return request.Header.Get("St")
	} else if ssdp := request.Header.Get("NTS"); ssdp == "ssdp:alive" || ssdp == "ssdp:byebye" {
		return request.Header.Get("NT")
	} else {
		return ""
	}
}

// Split the content of the UPnP
func (h *UPnPHandler) splitUPnP(request *http.Request) (string, string, string, string) {
	c := h.extractUPnP(request)

	ty := strings.Split(c, ":")

	if ty[0] == "upnp" {
		return ty[0], ty[1], "", ""
	} else if ty[0] == "urn" {
		add := ty[1]
		t := ty[2]
		d := ty[3]
		return ty[0], add, t, d
	} else {
		return "", "", "", ""
	}
}

// Handle function to process a packet
// info is the PacketInfo associated to this connection
// l4Payload is the bytes of the TCP or UDP packet
// rawPacket is the bytes of the packet as received on the wire
// Update the UPnP information if the protocol was SSDP, else do nothing
func (h *UPnPHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	request, err := h.buildUPnPRequest(l4Payload)

	if err != nil {
		return
	}

	ed, err := h.Collector.FindEndpointDataIP(info.SrcIP)

	if err != nil {
		return
	}

	if ss := h.extractServerString(request); ss != "" {
		ed.AddUPnPServerString(ss)
	}

	if ua := h.extractUserAgent(request); ua != "" {
		ed.AddUPnPUserAgent(ua)
	}

	if is, upnp, t, d := h.splitUPnP(request); is != "" {
		if is == "upnp" {
			ed.Update(func(ed *fccommon.EndpointData) bool {
				changed := upnp != ed.UPnP
				ed.UPnP = upnp
				return changed
			})
		} else if is == "urn" {
			ed.AddUPnPURN(upnp)
			if t == "service" {
				ed.AddUPnPService(d)
			} else if t == "device" {
				ed.AddUPnPDevice(d)
			}
		}
	}

	//fmt.Println("The UPnP handler has handled the packet")
}

// Returns the BPF of the packets the protocol handler is interested in
func (h *UPnPHandler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h *UPnPHandler) String() string {
	return "UPnP Handler"
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h *UPnPHandler) Name() string {
	return Name
}
