package main

import (
	"bytes"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"io/ioutil"
	"math/rand"
	"net/http"
	"sync"
	"time"

	"git.inverse.ca/inverse/fingerbank-collector/common"

	"github.com/inverse-inc/go-utils/sharedutils"

	statsd "gopkg.in/alexcesaro/statsd.v2"
)

var statsdPrefix = flag.String("statsd-prefix", "collector-tester", "The statsd prefix to use")

var statsdProto = flag.String("statsd-proto", "udp", "The statsd network protocol to use (tcp or udp)")

var statsdAddress = flag.String("statsd-address", "127.0.0.1:8125", "The statsd network address")

var key = flag.String("key", "", "The key to use for communication with the collector")

var collectorUrl = flag.String("collector-url", "http://localhost:4723", "The URL to reach the collector")

var minLoopTime = flag.Duration("min-loop-time", 2*time.Second, "The minimal amount of time each loop has to take")

var maxConcurrency = flag.Int("max-concurrency", 10, "The maximum amount of concurrent requests to perform on the collector")

var statsdClient *statsd.Client

func init() {
	flag.Parse()
	var err error
	statsdClient, err = statsd.New(statsd.Prefix(*statsdPrefix), statsd.Network(*statsdProto), statsd.Address(*statsdAddress))
	sharedutils.CheckError(err)
}

func checkError(err error) {
	if err != nil {
		fmt.Println(err)
	}
}

func doRequest(r *http.Request) (*http.Response, error) {
	r.Header.Set("Authorization", "Token "+*key)

	resp, err := http.DefaultClient.Do(r)

	if err != nil {
		return nil, err
	}

	if resp.StatusCode >= 400 {
		body, _ := ioutil.ReadAll(resp.Body)
		fmt.Println(string(body))
		return resp, errors.New(fmt.Sprintf("Wrong status code coming out of the collector %d", resp.StatusCode))
	} else {
		return resp, nil
	}
}

func getAllEndpoints() []fccommon.ExpandedEndpointData {
	defer statsdClient.NewTiming().Send("getAllEndpoints")

	req, _ := http.NewRequest("GET", *collectorUrl+"/endpoint_data", bytes.NewBuffer([]byte{}))
	resp, err := doRequest(req)

	if err != nil {
		checkError(err)
		return []fccommon.ExpandedEndpointData{}
	}

	endpoints := []fccommon.ExpandedEndpointData{}
	defer resp.Body.Close()
	dec := json.NewDecoder(resp.Body)
	dec.Decode(&endpoints)

	return endpoints
}

func patchEndpoint(mac string) {
	defer statsdClient.NewTiming().Send("patchEndpoint")

	req, _ := http.NewRequest("PATCH", *collectorUrl+"/endpoint_data/"+mac, bytes.NewBuffer([]byte(`{"dhcp6_fingerprint":"192,168,5,3"}`)))
	resp, err := doRequest(req)
	resp.Body.Close()
	checkError(err)
}

func getEndpoint(mac string) {
	defer statsdClient.NewTiming().Send("getEndpoint")

	req, _ := http.NewRequest("GET", *collectorUrl+"/endpoint_data/"+mac+"", bytes.NewBuffer([]byte{}))
	resp, err := doRequest(req)
	resp.Body.Close()
	checkError(err)
}

func detailsEndpoint(mac string) {
	defer statsdClient.NewTiming().Send("detailsEndpoint")

	req, _ := http.NewRequest("GET", *collectorUrl+"/endpoint_data/"+mac+"/details", bytes.NewBuffer([]byte{}))
	resp, err := doRequest(req)
	resp.Body.Close()
	checkError(err)
}

func main() {

	var endpoints []fccommon.ExpandedEndpointData
	i := 0
	for {
		minTime := time.After(*minLoopTime)
		conc := make(chan int, *maxConcurrency)

		if i%10 == 0 || len(endpoints) == 0 {
			fmt.Println("Refreshing endpoints")
			endpoints = getAllEndpoints()
		}

		wg := &sync.WaitGroup{}
		for _, endpoint := range endpoints {
			wg.Add(1)
			go func(endpoint fccommon.ExpandedEndpointData) {
				conc <- 1
				// Only update the endpoint one out of 5 times
				if rand.Int()%5 == 0 {
					patchEndpoint(endpoint.MAC)
				}
				detailsEndpoint(endpoint.MAC)
				getEndpoint(endpoint.MAC)
				wg.Done()
				<-conc
			}(endpoint)
		}
		wg.Wait()
		i++

		<-minTime
	}

}
